/*========================================================================
 * AMSR_E_L2A_togs - convert AMSR-E Tb files from HDF to generic swath
 *
 * AMSR-E L2A Tb files in HDF are described in the 
 * Data Management Plan amsr_dmp_aug2001.doc found here:
 * http://nsidc.org/intranet/GROUPS/amsr/#amsre_documents
 * under the link: 'AMSR-E Data Management Plan (final)'
 * the swath structure is described in struct_amsr.h
 * and the generic swath interface is described in gswath.c
 *
 * 16-Jul-2002 K.Knowles knowlesk@kryos.colorado.edu 303-492-0644
 * National Snow & Ice Data Center, University of Colorado, Boulder
 * Copyright (C) 2002-2004 University of Colorado
 *========================================================================*/
#include "define.h"
#include "matrix.h"
#include "hdf.h"
#include "mfhdf.h"
#include "gswath.h"
#include "AMSR_E_L2A_togs.h"

static const char AMSR_E_L2A_togs_c_rcsid[] = "$Header: /NSIDC_CVS/FILES/generic_swath/src/AMSR_E_L2A_togs/AMSR_E_L2A_togs.c,v 1.8 2004/01/28 18:46:56 knowlesk Exp $";

#define usage									\
" usage: AMSR_E_L2A_togs [-v -r res -i] [filename...]\n"				\
"  input : AMSR-E L2A Tb files in HDF\n"					\
"  output: AMSR half orbit files (*.res1, *-bscan.res5, and *-ascan.res5)\n"	\
"\n"										\
"  option: i - get file list from stdin as opposed to command line\n"		\
"          r res - output Level 2A data at res 1-5 (default = Level 1B)\n"	\
"          v - verbose\n"

/*
 * .res1 == lowres == 10km
 * .res5 == hires  ==  5km
 */
static char *res_num_description[] = {
    "0 - .res1 (10km) & .res5 (5km) Level 1B (not resampled)",
    "1 - .res1 (10km) L2A resampled to 56km antenna pattern",
    "2 - .res1 (10km) L2A resampled to 38km antenna pattern",
    "3 - .res1 (10km) L2A resampled to 21km antenna pattern",
    "4 - .res1 (10km) L2A resampled to 12km antenna pattern",
    "5 - .res5 (5km) Level 1B upper channels (not resampled)"
};

/*
 * global verbosity setting
 */
static int verbose;

/*
 * check HDF routine return value
 */
#define check_hdf_error(io_status, description)	\
if (FAIL == io_status)				\
{ fprintf(stderr,"error: %s\n", description);	\
  HEprint(stderr, 0);				\
}

/*------------------------------------------------------------------------
 * replace_hdfeos_suffix - generate output file names by replacing 
 *                         the .hdf-eos suffix
 *
 *	input : ael2_filename - filename.HDF-EOS
 *		a_b_flag - "-ascan", "-bscan", or ""
 *		output_suffix - .res1 or .res5
 *
 *	output: output_filename - filename.res1
 *				  filename-ascan.res5
 *				  filename-bscan.res5
 *
 *	result: TRUE == successful
 *
 *------------------------------------------------------------------------*/
static bool replace_hdfeos_suffix(char *ael2_filename, char *output_filename, 
				  char *a_b_flag, const char *output_suffix)
{
  char *basename, *suffix;

  strcpy(output_filename, ael2_filename);

  basename = strrchr(output_filename, '/');

  if (basename != NULL) {
    basename += 1; 
  } else {
    basename = output_filename;
  }

  
  suffix = strstr(basename, ".hdf");
  if (!suffix) suffix = strstr(basename, ".HDF");
  if (!suffix) suffix = strstr(basename, ".hdfeos");
  if (!suffix) suffix = strstr(basename, ".HDFEOS");
  if (!suffix) suffix = strstr(basename, ".hdf-eos");
  if (!suffix) suffix = strstr(basename, ".HDF-EOS");

  if (!suffix) {
    fprintf(stderr, "%s: can't find .hdf suffix\n", basename);
    return FALSE;
  }

  strcpy(suffix, a_b_flag);

  strcpy(suffix+strlen(a_b_flag), output_suffix);

  return TRUE;
}

/*------------------------------------------------------------------------
 * get_Tb_sds - read a Tb SDS into a buffer
 *
 *	input : file_id - HDF file
 *		sds_name - name of SDS
 *		dims_parm - expected dimensions, rows by columns
 *			    or (0,0) if not known
 *
 *	output: dims_parm - SDS dimensions, rows by columns
 *
 *	result: pointer to matrix of scaled Tb data (Kelvins)
 *
 *------------------------------------------------------------------------*/
float **get_Tb_sds(int32 file_id, char *sds_name, int32 dims_parm[2])
{
  int ios, i,j;
  int32 dims[2], start[2]={0,0};
  int32 sds_index, sds_id, rank, ntype, nattrs, attr_index;
  int16 **sds_data=NULL;
  float **Tbs=NULL;
  float32 Tb_scale=0.01, Tb_offset=327.68;
  char name[MAX_NC_NAME];
  char error_msg[2*MAX_NC_NAME];

  strcpy(error_msg, sds_name);

/*
 * convert name to ID, get dimensions, etc.
 */
  sds_index = SDnametoindex(file_id, sds_name);
  check_hdf_error(sds_index, strcat(error_msg, ": index from name"));

  sds_id = SDselect(file_id, sds_index);
  check_hdf_error(sds_id, strcat(error_msg, ": ID from index"));
      
  ios = SDgetinfo(sds_id, name, &rank, dims, &ntype, &nattrs);
  check_hdf_error(ios, strcat(error_msg, ": dimensions"));

  if (verbose >= 3)
    fprintf(stderr,">>> found SDS #%1d \"%s\"(%1dx%1d)\n", 
	    sds_index, sds_name, dims[0], dims[1]); 

/*
 * set dimensions if not already set
 */
  if (0 == dims_parm[0]) dims_parm[0] = dims[0];
  if (0 == dims_parm[1]) dims_parm[1] = dims[1];

/*
 * make sure data type and dimensions match expectations
 */
  if (rank != 2 || ntype != DFNT_INT16) {
    fprintf(stderr, "%s: rank or data type mismatch\n", sds_name);
    return NULL;
  }

  if (dims_parm[0] != dims[0] || dims_parm[1] != dims[1]) {
    fprintf(stderr, "%s: dimensions %1dx%1d != expected %1dx%1d\n",
	    sds_name, dims[0], dims[1], dims_parm[0], dims_parm[1]);
    return NULL;
  }

/* 
 * allocate buffer and read data
 */
  sds_data = (int16 **)matrix(dims[0], dims[1], sizeof(int16), 0);
  if (NULL == sds_data) { perror(sds_name); return NULL; }


  ios = SDreaddata(sds_id, start, NULL, dims, (VOIDP)(sds_data[0]));
  check_hdf_error(ios, strcat(error_msg,": reading Tbs"));

/*
 * read scale and offset attributes
 */
#ifdef DONT_FORGET_TO_FIX_THIS
  if (nattrs >= 2) { 
    attr_index = SDfindattr(sds_id, "Tb_scale");
    ios = SDreadattr(file_id, attr_index, (VOIDP)&Tb_scale);
    check_hdf_error(ios, "reading Tb_scale attribute");

    attr_index = SDfindattr(sds_id, "Tb_offset");
    ios = SDreadattr(file_id, attr_index, (VOIDP)&Tb_offset);
    check_hdf_error(ios, "reading Tb_offset attribute");
  }
#endif

/*
 * convert stored data to brightness temp
 */
  Tbs = (float **)matrix(dims[0], dims[1], sizeof(float), 0);
  if (NULL == Tbs) { perror(sds_name); free(sds_data); return NULL; }

  for (i = 0; i < dims[0]; i++) {
    for (j = 0; j < dims[1]; j++) {
      Tbs[i][j] = (float)sds_data[i][j] * Tb_scale + Tb_offset;
    }
  }

  free(sds_data);

  SDendaccess(sds_id);

  return Tbs;
}

/*------------------------------------------------------------------------
 * get_geolocation_sds - read a "Latitude" or "Longitude" SDS into a buffer
 *
 *	input : file_id - HDF file
 *		sds_index - index of SDS
 *		dims_parm - expected dimensions, rows by columns
 *			    or (0,0) if not known
 *
 *	output: dims_parm - SDS dimensions, rows by columns
 *
 *	result: pointer to matrix of lat or lon data (decimal degrees)
 *
 *------------------------------------------------------------------------*/
float **get_geolocation_sds(int32 file_id, int32 sds_index, int32 dims_parm[2])
{
  int ios;
  int32 dims[2], start[2]={0,0};
  int32 sds_id, rank, ntype, nattrs;
  float **loc_data=NULL;
  char sds_name[MAX_NC_NAME];
  char error_msg[2*MAX_NC_NAME];

  sprintf(error_msg, "sds index=%1d", sds_index);

/*
 * convert index to ID, get dimensions, etc.
 */
  sds_id = SDselect(file_id, sds_index);
  check_hdf_error(sds_id, strcat(error_msg, ": ID from index"));
      
  ios = SDgetinfo(sds_id, sds_name, &rank, dims, &ntype, &nattrs);
  check_hdf_error(ios, strcat(error_msg, ": dimensions"));

  if (verbose >= 3)
    fprintf(stderr,">>> found SDS #%1d \"%s\"(%1dx%1d)\n", 
	    sds_index, sds_name, dims[0], dims[1]); 

/*
 * set dimensions if not already set
 */
  if (0 == dims_parm[0]) dims_parm[0] = dims[0];
  if (0 == dims_parm[1]) dims_parm[1] = dims[1];

/*
 * make sure data type and dimensions match expectations
 */
  if (rank != 2 || ntype != DFNT_FLOAT32) {
    fprintf(stderr, "%s: rank or data type mismatch\n", sds_name);
    return NULL;
  }

  if (dims_parm[0] != dims[0] || dims_parm[1] != dims[1]) {
    fprintf(stderr, "%s: dimensions %1dx%1d != expected %1dx%1d\n",
	    sds_name, dims[0], dims[1], dims_parm[0], dims_parm[1]);
    return NULL;
  }

/* 
 * allocate buffer and read data
 */
  loc_data = (float32 **)matrix(dims[0], dims[1], sizeof(float32), 0);
  if (NULL == loc_data) { perror(sds_name); return NULL; }

  ios = SDreaddata(sds_id, start, NULL, dims, (VOIDP)(loc_data[0]));
  check_hdf_error(ios, strcat(error_msg,": reading sds data"));

  SDendaccess(sds_id);

  return loc_data;
}

/*------------------------------------------------------------------------
 * get_timecode_vdata - read the "Time" vdata into a buffer
 *
 *	input : ael2_filename - name of AE_L2A...HDF-EOS file
 *		nrecs - number of records expected or 0 if not known
 *
 *	output: nrecs - number of records read
 *
 *	result: pointer to vector of timecode data
 *
 *------------------------------------------------------------------------*/
byte4 *get_timecode_vdata(char *ael2_filename, int32 *nrecs)
{
  int32 imode, ios, irec, nrecs_read, rec_size;
  int32 file_id, vdata_id, vdata_ref;
  char error_msg[2*MAX_NC_NAME], fields[MAX_NC_NAME], vdata_name[MAX_NC_NAME];
  double vd_time[AEL2_MAX_SCANS];
  byte4 *timecode;

  strcpy(error_msg, ael2_filename);

/*
 * vdata connection is completely different than SD connection
 * so, file must be opened, interface initialized, etc.
 */
  file_id = Hopen(ael2_filename, DFACC_RDONLY, 0);
  check_hdf_error(file_id, "opening AE_L2 hdf file");
  if (FAIL == file_id) { perror(ael2_filename); return NULL; }

  ios = Vstart(file_id);
  check_hdf_error(ios, "starting vdata interface");

  vdata_ref = VSfind(file_id, "Time");
  vdata_id = VSattach(file_id, vdata_ref, "r");
  ios = VSinquire(vdata_id, &nrecs_read, &imode, fields, &rec_size, vdata_name);
  check_hdf_error(ios, "querying Time vdata");

  if (verbose >= 3)
    fprintf(stderr,">>> found Vdata #%1d \"%s\"(%1dx1)\n", 
	    vdata_ref, vdata_name, nrecs_read); 

/*
 * set nrecs if not already set
 */
  if (0 == *nrecs) *nrecs = nrecs_read;

/*
 * make sure data type and dimensions match expectations
 */
  if (rec_size != 8) {
    fprintf(stderr, "%s: %1d bytes can't be float64\n", vdata_name, rec_size);
    return NULL;
  }

  if (nrecs_read != *nrecs) {
    fprintf(stderr, "%s: %1d records != %1d expected\n", 
	    vdata_name, nrecs_read, *nrecs);
    return NULL;
  }

/*
 * read "Time" field of "Time" vdata
 */
  ios = VSsetfields(vdata_id, "Time");
  check_hdf_error(ios, "can't find the Time field");

  ios = VSread(vdata_id, (VOIDP)vd_time, nrecs_read, imode);
  check_hdf_error(ios, "reading Time vdata");

  ios = VSdetach(vdata_id);
  ios = Vend(file_id);
  ios = Hclose(file_id);

/*
 * allocate buffer and convert Time to timecode
 */
  timecode = calloc(nrecs_read, sizeof(byte4));
  if (NULL == timecode) { perror("amsrstogs: timecode"); return NULL; }

  for (irec = 0; irec < nrecs_read; irec++) {
    timecode[irec] = (byte4)nint(vd_time[irec]);
  }

  *nrecs = nrecs_read;

  if (verbose >= 2)
    fprintf(stderr,">> start time: %d  | end time: %d\n", 
	    timecode[0], timecode[nrecs_read-1]);

  return timecode;

}

/*
 * main
 */
int main (int argc, char *argv[]) {

  bool do_stdin, do_lowres, do_hires;
  bool processing_completed;
  int res_number, chan, sample, scan, newline;
  char *option;

  char ael2_filename[FILENAME_MAX];
  char res1_filename[FILENAME_MAX];
  char res5a_filename[FILENAME_MAX];
  char res5b_filename[FILENAME_MAX];

  int32 ael2_file=0;

  FILE *res1_file=NULL, *res5a_file=NULL, *res5b_file=NULL;

  GSwath *scan_res1=NULL, *scan_res5=NULL;

  byte4 *timecode=NULL;

  float **lat_lowres=NULL;
  float **lon_lowres=NULL;
  float **lat_hiresa=NULL;
  float **lon_hiresa=NULL;
  float **lat_hiresb=NULL;
  float **lon_hiresb=NULL;

  char **sds_names_lowres;
  char **sds_names_hires;

  float **Tb_lowres[NUM_SDS_LOWRES];
  float **Tb_hires[NUM_SDS_HIRES];

  int32 dims_lowres[2]={0,0}, dims_hires[2]={0,0};

  /*
   * set defaults
   */
  verbose = 0;
  do_stdin = FALSE;
  res_number = 0;

  for (chan = 0; chan < NUM_SDS_LOWRES; chan++) Tb_lowres[chan] = NULL;
  for (chan = 0; chan < NUM_SDS_HIRES; chan++) Tb_hires[chan] = NULL;

  /*
   * get command line options
   */
  while (--argc > 0 && (*++argv)[0] == '-') {
    for (option = argv[0]+1; *option != '\0'; option++) {
      switch (*option) {
	case 'v':
	  ++verbose;
	  break;
	case 'V':
	  fprintf(stderr, "%s\n", AMSR_E_L2A_togs_c_rcsid);
	  break;
	case 'i':
	  do_stdin = TRUE;
	  break;
	case 'r':
	  ++argv; --argc;
	  if (argc <= 0 || sscanf(*argv, "%i", &res_number) != 1) {
	    fprintf(stderr, "AMSR_E_L2A_togs: -r option expects res number\n");
	    error_exit(usage);
	  }
	  break;
	  default:
	    fprintf(stderr, "AMSR_E_L2A_togs: invalid option %c\n", *option);
	    error_exit(usage);
      }
    }
  }

  do_lowres = (res_number <= 4);
  do_hires = (res_number == 0 || res_number == 5);

  /*
   * check command line args
   */
  if (!do_stdin && argc < 1) error_exit(usage);

  /*
   * echo options
   */
  if (verbose) {
    fprintf(stderr, "> AMSR_E_L2A_togs: $Revision: 1.8 $\n");
    fprintf(stderr, "> getting filenames from %s\n", 
	    do_stdin ? "stdin" : "command line");
    fprintf(stderr, "> res = %s\n", res_num_description[res_number]);
  }

  /*
   * allocate generic swath structures
   */
  if (do_lowres) {
    scan_res1 = new_gswath(GS_TYPE_AMSR_RES1);
    if (NULL == scan_res1) error_exit("AMSR_E_L2A_togs: FATAL ERROR");
  }

  if (do_hires) {
    scan_res5 = new_gswath(GS_TYPE_AMSR_RES5);
    if (NULL == scan_res5) error_exit("AMSR_E_L2A_togs: FATAL ERROR");
  }

  /*
   * set SDS names for selected res number
   */
  switch (res_number) {
      case 0: 
	  sds_names_lowres = sds_L1B_lowres;
	  sds_names_hires = sds_L1B_hires;
	  break;
      case 1:
	  sds_names_lowres = sds_L2A_res1;
	  break;
      case 2:
	  sds_names_lowres = sds_L2A_res2;
	  break;
      case 3:
	  sds_names_lowres = sds_L2A_res3;
	  break;
      case 4:
	  sds_names_lowres = sds_L2A_res4;
	  break;
      case 5:
	  sds_names_hires = sds_L1B_hires;
	  break;
      default:
	  fprintf(stderr,"AMSR_E_L2A_togs: unsupported resolution %1d\n", res_number);
	  error_exit(usage);
  }

  /*
   * process files
   */
  for (;;) {

    if (ael2_file && !processing_completed) {
      fprintf(stderr, "AMSR_E_L2A_togs: abandoning %s\n", ael2_filename);
    }

    processing_completed = FALSE;

    /*
     * clean up
     */
    if (ael2_file) { SDend(ael2_file); ael2_file = 0; }
    if (res1_file) { fclose(res1_file); res1_file = NULL; }
    if (res5a_file) { fclose(res5a_file); res5a_file = NULL; }
    if (res5b_file) { fclose(res5b_file); res5b_file = NULL; }

    if (timecode) { free(timecode); timecode = NULL; }

    if (lat_lowres) { free(lat_lowres); lat_lowres = NULL; }
    if (lon_lowres) { free(lon_lowres); lon_lowres = NULL; }
    if (lat_hiresa) { free(lat_hiresa); lat_hiresa = NULL; }
    if (lon_hiresa) { free(lon_hiresa); lon_hiresa = NULL; }
    if (lat_hiresb) { free(lat_hiresb); lat_hiresb = NULL; }
    if (lon_hiresb) { free(lon_hiresb); lon_hiresb = NULL; }

    for (chan = 0; chan < NUM_SDS_LOWRES; chan++) {
	if (Tb_lowres[chan]) { free(Tb_lowres[chan]); Tb_lowres[chan] = NULL; }
    }

    for (chan = 0; chan < NUM_SDS_HIRES; chan++) {
	if (Tb_hires[chan]) { free(Tb_hires[chan]); Tb_hires[chan] = NULL; }
    }

    dims_lowres[0] = dims_lowres[1] = 0;

    dims_hires[0] = dims_hires[1] = 0;

    /*
     * open AE_L2 input file
     */
    if (do_stdin) {
      if (NULL == fgets(ael2_filename, FILENAME_MAX, stdin)) break;
      if (feof(stdin)) break;
      newline = strlen(ael2_filename) - 1;
      if (newline > 0 && newline < FILENAME_MAX) {
	ael2_filename[newline] = '\0';
      } else {
	continue;
      }
    }
    else {
      if (argc < 1) break;
      strcpy(ael2_filename, *argv);
      --argc; ++argv;
    }

    ael2_file = SDstart(ael2_filename, DFACC_RDONLY);
    if (FAIL == ael2_file) {
      fprintf(stderr, "error starting SD access\n"); 
      perror(ael2_filename);
      continue;
    }

    /*
     * start a new res1 output file
     */
    if (do_lowres) {

      if (!replace_hdfeos_suffix(ael2_filename, res1_filename, "",
				 GS_file_extension[GS_TYPE_AMSR_RES1])) {
	continue;
      }

      res1_file = fopen(res1_filename, "w");
      if (!res1_file) { perror(res1_filename); continue; }
    }

    /*
     * start new res5 output files for A scans and B scans
     * The A and B feed horns are 1 1/2 scans apart. So,
     * if you try to interleave them both in a single .res5
     * file then either the geolocation or the timecode
     * will be out of sequence
     */
    if (do_hires) {

      if (!replace_hdfeos_suffix(ael2_filename, res5a_filename, "-ascan",
				 GS_file_extension[GS_TYPE_AMSR_RES5])) {
	continue;
      }

      res5a_file = fopen(res5a_filename, "w");
      if (!res5a_file) { perror(res5a_filename); continue; }

      if (!replace_hdfeos_suffix(ael2_filename, res5b_filename, "-bscan",
				 GS_file_extension[GS_TYPE_AMSR_RES5])) {
	continue;
      }

      res5b_file = fopen(res5b_filename, "w");
      if (!res5b_file) { perror(res5b_filename); continue; }
    }

    if (verbose >= 1) {
      fprintf(stderr, "> reading: %s\n", ael2_filename);
    }
    if (verbose >= 2) {
      if (res1_file) fprintf(stderr, ">> writing: %s\n", res1_filename);
      if (res5a_file) fprintf(stderr, ">> writing: %s\n", res5a_filename);
      if (res5b_file) fprintf(stderr, ">> writing: %s\n", res5b_filename);
    }

    /*
     * get timecode
     */
    timecode = get_timecode_vdata(ael2_filename, dims_lowres);
    if (NULL == timecode) continue;

    /*
     * get geolocation data
     */
    if (do_lowres) {
      lat_lowres = get_geolocation_sds(ael2_file, SDS_INDEX_LAT_LOWRES, dims_lowres);
      lon_lowres = get_geolocation_sds(ael2_file, SDS_INDEX_LON_LOWRES, dims_lowres);
      if (NULL == lat_lowres || NULL == lon_lowres) continue;
    }

    if (do_hires) {
      lat_hiresa = get_geolocation_sds(ael2_file, SDS_INDEX_LAT_HIRESA, dims_hires);
      lon_hiresa = get_geolocation_sds(ael2_file, SDS_INDEX_LON_HIRESA, dims_hires);
      lat_hiresb = get_geolocation_sds(ael2_file, SDS_INDEX_LAT_HIRESB, dims_hires);
      lon_hiresb = get_geolocation_sds(ael2_file, SDS_INDEX_LON_HIRESB, dims_hires);
      if (NULL == lat_hiresa || NULL == lon_hiresa 
          || NULL == lat_hiresb || NULL == lon_hiresb) continue;
    }

    /*
     * read brightness temps
     */
    if (do_lowres) {
      for (chan=0; chan < NUM_SDS_LOWRES; chan++) {
        Tb_lowres[chan] = get_Tb_sds(ael2_file, sds_names_lowres[chan], dims_lowres);
	if (NULL == Tb_lowres[chan]) continue;
      }

    /*
     * write res1 records
     */
      for (scan = 0; scan < dims_lowres[0]; scan++) {

	scan_res1->timecode = timecode[scan];

	for (sample = 0; sample < dims_lowres[1]
		 && sample < scan_res1->info->num_samples; sample++) {

	  scan_res1->pos[sample].lat = lat_lowres[scan][sample];
	  scan_res1->pos[sample].lon = lon_lowres[scan][sample];


	  for (chan = 0; chan < NUM_SDS_LOWRES; chan++) {
	    scan_res1->pos[sample].temp[chan] = Tb_lowres[chan][scan][sample];
	  }
	}

	write_gswath(scan_res1, scan, res1_file, NUM_SDS_LOWRES, lowres_channels);
      }
      if (verbose >= 2) fprintf(stderr,">> %d res1 scans\n", dims_lowres[0]);

    }

    /*
     * write res5 records
     */
    if (do_hires) {

      for (chan=0; chan < NUM_SDS_HIRES; chan++) {
        Tb_hires[chan] = get_Tb_sds(ael2_file, sds_names_hires[chan], dims_hires);
	if (NULL == Tb_hires[chan]) continue;
      }

      for (scan = 0; scan < dims_hires[0]; scan++) {

      /*
       * A and B scan are recorded simultaneously
       * so they share the same timecode
       */
	scan_res5->timecode = timecode[scan];

      /*
       * A scan
       */
	for (sample = 0; sample < dims_hires[1]
		 && sample < scan_res5->info->num_samples; sample++) {

	  scan_res5->pos[sample].lat = lat_hiresa[scan][sample];
	  scan_res5->pos[sample].lon = lon_hiresa[scan][sample];

	  scan_res5->pos[sample].temp[0] = Tb_hires[0][scan][sample];
	  scan_res5->pos[sample].temp[1] = Tb_hires[1][scan][sample];

	}

	write_gswath(scan_res5, scan, res5a_file, 
		     NUM_SDS_HIRES/2, hires_channels);

      /*
       * B scan
       */
	for (sample = 0; sample < dims_hires[1]
		 && sample < scan_res5->info->num_samples; sample++) {

	  scan_res5->pos[sample].lat = lat_hiresb[scan][sample];
	  scan_res5->pos[sample].lon = lon_hiresb[scan][sample];

	  scan_res5->pos[sample].temp[0] = Tb_hires[2][scan][sample];
	  scan_res5->pos[sample].temp[1] = Tb_hires[3][scan][sample];

	}

	write_gswath(scan_res5, scan, res5b_file, 
		     NUM_SDS_HIRES/2, hires_channels);

      }

      if (verbose >= 2) fprintf(stderr,">> %d res5 scans\n", 2*dims_hires[0]);

    }

    processing_completed = TRUE;
  }

  exit(EXIT_SUCCESS);
}
