/*========================================================================
 * get_smmr_tbs (modified from extract.c)
 *
 * Convert SMMR Tb files in HDF to generic swath format.
 *
 * See README.smmr_pfndr for SMMR Tb file format
 *
 * 02-Sep-1998 K.Knowles knowlesk@kryos.colorado.edu 303-492-0644
 * National Snow & Ice Data Center, University of Colorado, Boulder
 * Copyright (C) 1998-2004 University of Colorado
 * ========================================================================*/
#include "define.h"
#include "dfsd.h"
#include "df.h"
#include "struct_smmr.h"

static const char get_smmr_tbs_c_rcsid[] = "$Header: /NSIDC_CVS/FILES/generic_swath/src/get_smmr_tbs/get_smmr_tbs.c,v 1.15 2004/01/23 01:06:10 knowlesk Exp $";

#define usage								\
"\n"									\
"usage: get_smmr_tbs [-q bitnums -s -v] input_file\n"			\
"\n"									\
"  input : input_file - SMMR Tbs in HDF\n"				\
"\n"									\
"  output: input_file.smmr - SMMR Tbs in generic swath format\n"	\
"\n"									\
"  option: q bitnums - reject scans with these status bits set\n"	\
"                      possible values are [012345], default = none\n"	\
"          s - print summary stats\n"					\
"          v - verbose (may be repeated to increase verbosity)\n"	\
"\n"									\
"  input_file should include the following HDF objects:  \n"		\
"\tCODE\tFILE\t\t\tTAG\tTYPE\n"						\
"\t----\t----\t\t\t---\t----\n"						\
"\tSP1\tSpacecraft Position-1\tSDS\tINT16\n"				\
"\tSP2\tSpacecraft Position-2\tSDS\tFLOAT32\n"				\
"\tSAL\tScan Angle and Location\tSDS\tINT16\n"				\
"\tCAL\tCalibration\t\tSDS\tINT16\n"					\
"\t06V\t6.6GHz-V\t\tSDS\tINT16\n"					\
"\t06H\t6.6GHz-H\t\tSDS\tINT16\n"					\
"\t10V\t10.69GHz-V\t\tSDS\tINT16\n"					\
"\t10H\t10.69GHz-H\t\tSDS\tINT16\n"					\
"\t18V\t18GHz-V\t\t\tSDS\tINT16\n"					\
"\t18H\t18GHz-H\t\t\tSDS\tINT16\n"					\
"\t21V\t21GHz-V\t\t\tSDS\tINT16\n"					\
"\t21H\t21GHz-H\t\t\tSDS\tINT16\n"					\
"\t37V\t37GHz-V\t\t\tSDS\tINT16\n"					\
"\t37H\t37GHz-H\t\t\tSDS\tINT16\n"					\
"\tSFT\tSurface Type\t\tSDS\tINT8\n"					\
"\n"


static int verbose = 0;

/* Legitimate ROVs for input data */
#define NO_DATA 0
#define STATUS_OK 0
#define CHANNEL_OFF -32000

#define MIN_SCAN_START 0
#define MAX_SCAN_START 86400

#define MIN_GMT_YEAR 1978
#define MAX_GMT_YEAR 1987

#define MIN_GMT_DOY 1
#define MAX_GMT_DOY 365

#define MIN_LAT -9000
#define MAX_LAT 9000

#define MIN_LON -18000
#define MAX_LON 18000

#define MAXRANK 2
#define MAXLEN 10
#define LISTSIZE 17
#define NUMOBJ 13
#define SECS_PER_YEAR 31536000
#define SECS_PER_DAY 86400
#define MINUTES_PER_DAY 1440

/* 
Interpretation of Scan Status Word
-----------------------------------------------------------------------------
    (16-bit word --- all bits will be zero except as follows:)

    Bit 0 = 1: Possible loss of data quality in level 1A processing 
    (flag passed from input TAT data).


    Bit 1 = 1: Scan is in period of initialization of calibration and
     engineering data running averages. Brightness temperature calibrations may be
     less reliable during this period.


    Bit 2 = 1: One or more of the instrument component temperatures used 
    in the radiometric calibrations has deviated by more than +/- 3K from its
     running average value in the previous scan.  The accuracy of the brightness 
    temperature calibrations in this scan may be reduced.


    Bit 3 = 1: Pitch, roll, or yaw error value is greater than +/- 0.95
    degrees.  This indicates an anomalous spacecraft attitude variation.


    Bit 4 = 1: The average value of the footprint incidence angles in the
     scan has deviated by more than +/- 0.1 degrees from the average value for in 
    previous scan.  This indicates where caution should be exercised in using 
    incidence angles in subsequent geophysical processing.


    Bit 5 = 1: Scan is in the "sun-in-the-cold-horn" period.  Cold 
    calibration counts have been interpolated and brightness temperature
    calibration accuracy may be reduced.


    (Bits 6-15 not used)
*/

#define MAX_STATUS_BITNUM 5

static unsigned short scan_status_mask = 0x0000;

/* debug control */
#ifdef DEBUG
static int trap_lo=-1, trap_hi=-1, tripped=0;
#endif

/* Arrays to be dynamically allocated when needed. */

float32 *f32;

int16 *i16;

int8 *i8;

/* Global variables */
int32 dims[MAXRANK];					/* dimensions of array */
uint16 ref;					/* SDS reference number in file */
enum types {FLOAT32,INT16,INT8} DFNT_TYPE;

static bool sync_scan_time(struct SMMR_REC *smmr, int scan_start_time);

int main(int argc, char *argv[]) {

  int bitnum, chan, i, j, k, scan, sample;	/* lcv */
  int err;			/* error flag */

  unsigned int f_scp,f_wr;

  intn rank;			/* rank of array */

  char *option, *scan_status_select=NULL;
  char infilename[FILENAME_MAX];
  char outfilename[FILENAME_MAX];
  FILE *outfile;

  /* The following variables are used in the HDF utility DFANlablist. */
  uint16 reflist[LISTSIZE];	/* array of SDS reference numbers */
  char lablst[MAXLEN*LISTSIZE+1]; /* array of SDS labels */
  int entries;			/* number of SDS's in file */
  int start = 1;		/* position to begin reading file */

  /* 
   * The object codes for the SMMR data we want,
   * and the expected number of columns, rank, and type
   * (they should all have SMMR_SCANS_PER_ORBIT rows)
   * NOTE: the order is important: SP1 must be read 
   * prior to others for detection of bad scan status to work.
   * Also, order of temperatures must match order of
   * temperatures in the data structure SMMR_REC_NT array.
   */
  char *obj_codes[NUMOBJ]={ "SP1","SP2","SAL","06H","06V",
			  "10H", "10V", "18H", "18V",
			  "21H", "21V", "37H", "37V" };
  const int expected_cols[NUMOBJ]={ 13, 9, 278, 94, 94, 94,
				  94, 94, 94, 94, 94, 94, 94 };
  const int expected_rank = 2;
  const enum types expected_types[NUMOBJ] =
  { INT16, FLOAT32, INT16, INT16, INT16, INT16, 
    INT16, INT16, INT16, INT16, INT16, INT16, INT16 };

  /* 
   * A pointer to an array of smmr records for output
   */
  struct SMMR_REC *smmr;
  size_t smmr_rec_size = sizeof(struct SMMR_REC);
  size_t written_bytes = (size_t) 0;
  size_t nbytes;

  /* Local variables and constants */
  int ascending;
  bool do_summary=FALSE;
  int scan_start_time;
  int16 scan_status;
  int16 gmt_year, gmt_doy, gmt_min;
  int bad_status_scans;
  int no_time_scans;
  int16 lat, lon;
  int16 tb;
  int16 minimum_tb[SMMR_NUM_CHAN];
  int16 maximum_tb[SMMR_NUM_CHAN];
  long int num_min_violators[SMMR_NUM_CHAN];
  long int num_max_violators[SMMR_NUM_CHAN];
  long int num_samples[SMMR_NUM_CHAN];
  long int num_channel_off[SMMR_NUM_CHAN];
  long int num_no_data[SMMR_NUM_CHAN];

/*
 *	get command line options
 */
  while (--argc > 0 && (*++argv)[0] == '-') {
    for (option = argv[0]+1; *option != '\0'; option++) {
      switch (*option) {
      case 'q':
	++argv; --argc;
	if (argc <= 0) { error_exit(usage); }
	scan_status_select = strdup(*argv);
	if (!scan_status_select) {
	  perror("get_smmr_tbs: scan_status_select"); exit(EXIT_FAILURE); }
	break;
      case 's':
	do_summary = TRUE;
	break;
      case 'v':
	++verbose;
	break;
      case 'V':
	fprintf(stderr, "%s\n", get_smmr_tbs_c_rcsid);
	break;
      default:
	fprintf(stderr, "gsgrid: invalid option %c\n", *option);
	error_exit(usage);
      }
    }
  }

  /*
   * get command line args
   */
  if (argc < 1) { error_exit(usage); }

  if (strlen(*argv)+5 > FILENAME_MAX) {
    fprintf(stderr,"get_smmr_tbs: file name is too long, max=%d chars\n"
	    "%s\n", FILENAME_MAX, *argv);
    exit(EXIT_FAILURE);
  }
  strcpy(infilename, *argv);
  sprintf(outfilename, "%s.smmr", infilename);

  /*
   * set scan_status_mask based on bitnums selected
   */
  if (scan_status_select) {
    for (option = scan_status_select; *option != '\0'; option++) {
      bitnum = *option - '0';
      if (0 <= bitnum && bitnum <= MAX_STATUS_BITNUM) {
	scan_status_mask |= 1 << bitnum;
      } else {
	fprintf(stderr,"get_smmr_tbs: bad bit flag selector %c\n", *option);
      }
    }
  }


  /* 
   * Output a message echoing inputs.
   */
  if (verbose) {
    fprintf(stderr,"> get_smmr_tbs $Revision: 1.15 $\n");
    fprintf(stderr,">   input_file : %s\n", infilename);
    fprintf(stderr,">   output_file: %s\n", outfilename);
    fprintf(stderr,">   scan status mask = 0x%4.4x\n", scan_status_mask);
  }

  /* Initialize processing stats counters */
  bad_status_scans = 0;
  no_time_scans = 0;
  for (i = TB06H; i <= TB37V; i++) {
    minimum_tb[i]=SHRT_MAX;
    maximum_tb[i]=SHRT_MIN;
    num_min_violators[i]=0L;
    num_max_violators[i]=0L;
    num_samples[i]=0L;
    num_channel_off[i]=0L;
    num_no_data[i]=0L;
  }
    
  /* 
   * Allocate an array of structures initialized to zeroes
   * to be filled with the desired data
   */
  smmr = (struct SMMR_REC *) 
    calloc (SMMR_SCANS_PER_ORBIT, smmr_rec_size);
  if (!smmr) { perror("get_smmr_tbs: struct SMMR_REC"); exit(EXIT_FAILURE); }

  /* Return a list of all reference numbers and labels for the SDS tag. */
  entries = DFANlablist(infilename, DFTAG_SD, reflist, 
			lablst, LISTSIZE, MAXLEN, start);
  if (verbose >= 2)
    fprintf(stderr,">> DFANlablist found %d SDSs in this file.\n",entries);

  if (-1 == entries) { 
    error_exit("get_smmr_tbs: DFANlablist: error reading reference numbers");
  }

  /* 
   * Loop for each SDS needed to populate a piece of
   * the output structure
   */
  for (i = 0; i < NUMOBJ; i++) {

    /* 
     * Initialize the ref and data type
     * for the current object being read
     */
    SetFileDefs(obj_codes[i],reflist);

    /* 
     * Specify reference number of SDS to be read 
     * when DFSDgetdata is called. 
     */
    err = DFSDreadref(infilename,ref);
    if (err != 0) {
      error_exit("get_smmr_tbs: DFSDreadref: error setting references");
    }

    /* Get the rank and dimension of the next SDS in the file. */
    err = DFSDgetdims(infilename, &rank, dims, MAXRANK);
    if (err != 0) {
      error_exit("get_smmr_tbs: DFSDgetdims: error reading dimensions");
    }

    if (verbose >= 3)
      fprintf(stderr,">>> Object = %s, rank=%u, dims=%u x %u\n",
	      obj_codes[i],rank,dims[0],dims[1]);

    /* Check for unexpected dimensions*/
    if (rank != expected_rank || 
	dims[0] != SMMR_SCANS_PER_ORBIT ||
	dims[1] != expected_cols[i]) {

      fprintf(stderr,"get_smmr_tbs: DFSDgetdims: bad rank dims %u %u %u\n"
	      "  should have been %u %u %u)\n",
	      rank, dims[0], dims[1],
	      expected_rank, SMMR_SCANS_PER_ORBIT, expected_cols[i]);
      exit(EXIT_FAILURE);
    }

    switch (DFNT_TYPE) {

    case INT16:

      if (expected_types[i] != INT16) {
	error_exit("get_smmr_tbs: expected INT16 data not found");
      }
      else
      { if (verbose >= 2)
	  fprintf(stderr,">> SDS %2d (%s) is INT16 data, as expected\n", 
		  i, obj_codes[i]);
      }

      /* Allocate i16 array (contiguous block) */
      i16 = (int16 *)malloc(dims[0] * dims[1] * sizeof(int16));
      if (!i16) { perror("get_smmr_tbs: int16"); exit(EXIT_FAILURE); }

      /* Sets number type to be used for next DFSDgetdata. */
      DFSDsetNT(DFNT_INT16);

      /* Get dataset of the next SDS in the file and store in array. */
      err = DFSDgetdata(infilename, rank, dims, i16);
      if (err!=0) {
	error_exit("get_smmr_tbs: read from file to array failed");
      }

      /* Check for unexpected dimensions (again!) */
      if (rank != expected_rank ||
	  dims[0] != SMMR_SCANS_PER_ORBIT ||
	  dims[1] != expected_cols[i]) {
	fprintf(stderr,"get_smmr_tbs: DFSDgetdata: bad rank dims %u %u %u\n"
		"  should have been %u %u %u)\n",
		rank, dims[0], dims[1],
		expected_rank, SMMR_SCANS_PER_ORBIT, expected_cols[i]);
	exit(EXIT_FAILURE);
      }

      /*
       * Retrieve all data from this SDS
       * to the output data structure
       */
      if (i == 0) {

	/* 
	 * This is the SP1 data.  Check the scan status
	 * word ("column" 1) first, to see if the rest of the data 
	 * should be retrieved.
	 */
	for (scan = 0; scan < SMMR_SCANS_PER_ORBIT; scan++) {

	  scan_status = *(i16+(scan*expected_cols[i])+1);

	  if (scan_status & scan_status_mask != 0) {
	    bad_status_scans++;
	    continue;
	  }

	  smmr[scan].status = scan_status;
	    
	  gmt_year = *(i16+(scan*expected_cols[i])+2);

	  if (0 == gmt_year) {
	    no_time_scans++;
	    continue;
	  }

	  gmt_doy = *(i16+(scan*expected_cols[i])+3);
	  gmt_min = *(i16+(scan*expected_cols[i])+4);

	  smmr[scan].ascending = *(i16+(scan*expected_cols[i])+0);

	  smmr[scan].gmt_year = gmt_year + 1900;
	  smmr[scan].gmt_doy = gmt_doy;
	  smmr[scan].gmt_seconds = gmt_min*60;
	      
#ifdef DEBUG
	    if (trap_lo <= scan && scan <= trap_hi)
	      ++tripped;
#endif
	}
      }
      else if (i == 2) {

	/* 
	 * This is the SAL data.  
	 * Get latitude/longitude pairs for each sample across the scan.
	 * Skip the scan if there has been no time data saved.
	 */
	for (scan=0; scan<SMMR_SCANS_PER_ORBIT; scan++) {

	  if (smmr[scan].smmr_time == (byte4)NO_DATA) continue;

#ifdef DEBUG
	  if (trap_lo <= scan && scan <= trap_hi)
	    ++tripped;
#endif
	  for (sample=0; sample<SMMR_RECS_PER_SCAN; sample++) {

	    lat = *(i16+(scan*expected_cols[i])+(2*sample));
	    lon = *(i16+(scan*expected_cols[i])+(2*sample+1));
#ifdef DEBUG
	    if (verbose >= 4) {
	      if (fmod(scan,100) == 0 && fmod(sample,20) == 0)
		fprintf(stderr,">>>> Scan=%d, sample=%d, lat=%d, lon=%d\n",
			scan,sample,lat,lon);
	  }

#endif
	    if (lat < MIN_LAT || lat > MAX_LAT ||
		lon < MIN_LON || lon > MAX_LON ) {
	      fprintf(stderr,"get_smmr_tbs: invalid geographic coordinates\n");
	      fprintf(stderr,"  SDS=%s, scan=%d, sample=%d, lat=%d, lon=%d\n",
		     obj_codes[i], scan, sample, lat, lon);
	      exit(EXIT_FAILURE);
	    }
	    else
	    {
	      smmr[scan].full[sample].lat = 
		(byte2)(lat+(SMMR_LAT_BIAS/SMMR_LAT_SCALE));
	      smmr[scan].full[sample].lon = 
		(byte2)(lon+(SMMR_LON_BIAS/SMMR_LON_SCALE));
	    }
	      
	  }
	}
      }
      else if (i >= 3 && i <= 12) {

	chan = i-3;
	/* 
	 * This is the brightness temperature data.  
	 * Get the temperature for each sample across the scan.
	 * Skip the scan if there has been no time data saved.
	 */
	for (scan=0; scan<SMMR_SCANS_PER_ORBIT; scan++) {

	  if (smmr[scan].smmr_time == (byte4)NO_DATA) continue;

	  for (sample=0; sample<SMMR_RECS_PER_SCAN; sample++) {

	    tb = *(i16+(scan*expected_cols[i])+sample);

#ifdef DEBUG
	    if (verbose >= 4) {
	      if (fmod(scan,100) == 0 && fmod(sample,20) == 0)
		fprintf(stderr,">>>> Object=%s, scan=%d, sample=%d, tb=%d\n",
			obj_codes[i],scan,sample,tb);
	    }
	    if (trap_lo <= scan && scan <= trap_hi)
	      ++tripped;
#endif

	    /* 
	     * Do range checking and statistics first 
	     * Do not save values that violate legitimate ROV
	     */
	    if (tb < minimum_tb[chan]) minimum_tb[chan]=tb;
	    if (tb > maximum_tb[chan]) maximum_tb[chan]=tb;

	    if (tb == CHANNEL_OFF) {
	      num_channel_off[chan]++;
	    }
	    else if (tb == NO_DATA) {
	      num_no_data[chan]++;
	    }
	    else if (tb < smmr_Tb_min[chan]/SMMR_TEMP_SCALE) {
	      num_min_violators[chan]++;
#ifdef DEBUG
	      if (verbose >= 4)
		fprintf(stderr,">>>> ROV error: object=%s, scan=%d, sample=%d, tb=%d\n",
			obj_codes[i], scan, sample, tb);
#endif
	    }
	    else if (tb > smmr_Tb_max[chan]/SMMR_TEMP_SCALE) {
	      num_max_violators[chan]++;
#ifdef VIOLATORS
	      fprintf(stderr,"ROV: object=%s, scan=%d, sample=%d, tb=%d.\n",
		      obj_codes[i],scan,sample,tb);
#endif
	    }
	    else {
	      smmr[scan].full[sample].temp[chan] = (byte2) tb;
	      num_samples[chan]++;
	    }

	  }
	}
      }

      free(i16);
      i16 = NULL;
      break;

    case FLOAT32:

      if (expected_types[i] != FLOAT32) {
	error_exit("get_smmr_tbs: expected FLOAT32 data not found");
      }
      else {
	if (verbose >= 2)
	  fprintf(stderr,">> SDS %2d (%s) is FLOAT32 data, as expected.\n", 
		  i, obj_codes[i]);
      }

      /* Allocate f32 array (contiguous block) */
      f32 = (float32 *)malloc(dims[0] * dims[1] * sizeof(float32));
      if (!f32) { perror("get_smmr_tbs: float32"); exit(EXIT_FAILURE); }

      /* Sets number type to be used for next DFSDgetdata. */
      DFSDsetNT(DFNT_FLOAT32);

      /* Get dataset of the next SDS in the file and store in array. */
      err = DFSDgetdata(infilename, rank, dims, f32);
      if (err != 0) {
	error_exit("get_smmr_tbs: read from file to array failed");
      }

      /* Check for unexpected dimensions (again!) */
      if (rank != expected_rank ||
	  dims[0] != SMMR_SCANS_PER_ORBIT ||
	  dims[1] != expected_cols[i]) {
	fprintf(stderr,"get_smmr_tbs: DFSDgetdata: bad rank dims %u %u %u\n"
		"  should have been %u %u %u)\n",
		rank, dims[0], dims[1],
		expected_rank, SMMR_SCANS_PER_ORBIT, expected_cols[i]);
	exit(EXIT_FAILURE);
      }

      /*
       * Retrieve all data from this SDS
       * to the output data structure
       */
      if (1 == i) {

	/* 
	 * This is the SP2 data.  We only need "column" 0, the
	 * Scan start time in GMT seconds/day
	 */
	for (scan=0; scan<SMMR_SCANS_PER_ORBIT; scan++) {

	  scan_start_time = nint(*(f32+(scan*expected_cols[i])));

#ifdef DEBUG
	  if (verbose >= 3) {
	    if (fmod(scan,100) == 0)
	      fprintf(stderr,">>> Scan=%d, scan_start_time=%d (secs/day)\n",
		      scan,scan_start_time);
	  }
	  if (trap_lo <= scan && scan <= trap_hi)
	    ++tripped;
#endif

	  if (!sync_scan_time(&(smmr[scan]), scan_start_time)) {
	    fprintf(stderr,"get_smmr_tbs: SDS=%s, scan=%d\n",
		    obj_codes[i], scan);
	    exit(EXIT_FAILURE);
	  }

#ifdef DEBUG
	  if (trap_lo <= scan && scan <= trap_hi)
	    ++tripped;
#endif


	}
      }

      free(f32);
      f32 = NULL;
      break;

    case INT8:

      /* 
       * If we eventually need this type of data
       * put this case back in.  But for now,
       * exit in error
       */
      error_exit("get_smmr_tbs: unexpected int8 data");

    } /* switch */

  } /* for */

  /* 
   * Write the smmr data to output file
   */
  outfile = fopen(outfilename, "wb");
  if (!outfile) { perror(outfilename); exit(EXIT_FAILURE); }

  for (scan=0; scan<SMMR_SCANS_PER_ORBIT; scan++) {

    nbytes = fwrite (smmr+scan, 1, smmr_rec_size, outfile);
    if (nbytes != smmr_rec_size) {
      perror("outfilename"); exit(EXIT_FAILURE); }

    written_bytes += nbytes;
  }

  fclose(outfile);

  if (verbose)
    fprintf(stderr,"> %ld bytes written to %s\n", written_bytes, outfilename);

    /* And output status */
  if (do_summary) {

    fprintf(stderr,"\n\n>> ==================Status Summary=============\n\n");
    fprintf(stderr,"Number of scans skipped due to bad status   =%d\n",bad_status_scans);
    fprintf(stderr,"Number of scans skipped due to no time data =%d\n",no_time_scans);
    fprintf(stderr,"Channel Statistics:\n");
    fprintf(stderr,"        Channel Min      Max      Min_Viol     Max_Viol     Chan_Off No_Data Num_Legit_Saved Total\n");
    for (i=TB06H; i<=TB37V; i++) {

      fprintf(stderr,"STATS %8d %8d %8d %12ld %12ld %12ld %12ld %12ld %12ld\n",i,
	      minimum_tb[i],
	      maximum_tb[i],
	      num_min_violators[i],
	      num_max_violators[i],
	      num_channel_off[i],
	      num_no_data[i],
	      num_samples[i],
	      num_channel_off[i]+num_no_data[i]+num_samples[i]);
    }

    fprintf(stderr,"\n\n>> =============End of Status Summary==========\n\n");
  }

  exit(EXIT_SUCCESS);

} /* main */

/**************************************************************************/
/* Looks for object name and sets type and reference */

int SetFileDefs(argv,reflist)
char *argv;
uint16 *reflist;
{
  if (strcmp(argv,"SP1") == 0)
    {
      DFNT_TYPE = INT16;
      ref = reflist[0];
    }
  else if (strcmp(argv,"SP2") == 0)
    {
      DFNT_TYPE = FLOAT32;
      ref = reflist[1];
    } 
  else if (strcmp(argv,"SAL") == 0)
    {
      DFNT_TYPE = INT16;
      ref = reflist[2];
    }
  else if (strcmp(argv,"CAL") == 0)
    {
      DFNT_TYPE = INT16;
      ref = reflist[3];
    }
  else if (strcmp(argv,"06H") == 0)
    {
      DFNT_TYPE = INT16;
      ref = reflist[4];
    }
  else if (strcmp(argv,"06V") == 0)
    {
      DFNT_TYPE = INT16;
      ref = reflist[5];
    }
  else if (strcmp(argv,"10H") == 0)
    {
      DFNT_TYPE = INT16;
      ref = reflist[6];
    }
  else if (strcmp(argv,"10V") == 0)
    {
      DFNT_TYPE = INT16;
      ref = reflist[7];
    }
  else if (strcmp(argv,"18H") == 0)
    {
      DFNT_TYPE = INT16;
      ref = reflist[8];
    }
  else if (strcmp(argv,"18V") == 0)
    {
      DFNT_TYPE = INT16;
      ref = reflist[9];
    }
  else if (strcmp(argv,"21H") == 0)
    {
      DFNT_TYPE = INT16;
      ref = reflist[10];
    }
  else if (strcmp(argv,"21V") == 0)
    {
      DFNT_TYPE = INT16;
      ref = reflist[11];
    }
  else if (strcmp(argv,"37H") == 0)
    {
      DFNT_TYPE = INT16;
      ref = reflist[12];
    }
  else if (strcmp(argv,"37V") == 0)
    {
      DFNT_TYPE = INT16;
      ref = reflist[13];
    }
  else if (strcmp(argv,"SFT") == 0)
    {
      DFNT_TYPE = INT8;
      ref = reflist[14];
    }
  else
    {
      error_exit(usage);
    }
}

/*------------------------------------------------------------------------
 * sync_scan_time - set smmr_time and synchronize with GMT date & time
 *------------------------------------------------------------------------*/
static bool sync_scan_time(struct SMMR_REC *scan, int scan_start_time) {

  int leap_day;
  byte4 smmr_time=0;

  /* 
   * number of leap days prior to the year starting in 1978
   * So, for 1981, 
   * prior_leap_days[1981-1978]=prior_leap_days[3]=1 (the leap day in 1980)
   */
  static const int prior_leap_days[] = { 0, 0, 0, 1, 1, 1, 1, 2, 2, 2 };

  /*
   * check for missing data indicator
   */
  if (0 == scan->gmt_year) return TRUE;

  leap_day = (scan->gmt_year % 4) ? 0 : 1;

  if (scan->gmt_year < MIN_GMT_YEAR ||
      scan->gmt_year > MAX_GMT_YEAR) {
    fprintf(stderr,"sync_scan_time: invalid year %d\n", scan->gmt_year);
    goto error_return;
  }
  else if (scan->gmt_doy < MIN_GMT_DOY ||
	   scan->gmt_doy > MAX_GMT_DOY+leap_day) {
    fprintf(stderr,"sync_scan_time: invalid day of year %d\n", scan->gmt_doy);
    goto error_return;
  }
  else if (scan_start_time < MIN_SCAN_START ||
	   scan_start_time > MAX_SCAN_START) {
    fprintf(stderr,"sync_scan_time: invalid scan start time %d\n",
	    scan_start_time);
    goto error_return;
  }
  else {

    /*
     * normalize date-time, again!
     */
    if (scan->gmt_seconds >= SECS_PER_DAY) {
      scan->gmt_seconds -= SECS_PER_DAY;
      scan->gmt_doy += 1;
      if (scan->gmt_doy > 365 + leap_day) {
	scan->gmt_doy -= 365 + leap_day;
	scan->gmt_year += 1;
      }
    }

    /*
     * check for bogus doy rollover and correct
     */
    if (0 == scan->gmt_seconds && scan_start_time > 60) {
      scan->gmt_doy -= 1;
      if (scan->gmt_doy < 1) {
	scan->gmt_year -= 1;
	scan->gmt_doy = scan->gmt_year%4 ? 365 : 366;
      }
    }

    smmr_time = (scan->gmt_year - MIN_GMT_YEAR)*SECS_PER_YEAR;

    smmr_time += (scan->gmt_doy - 1)*SECS_PER_DAY;

    /*
     * Add in leap days from years prior to the 
     * current one. Note that the current year leap day is
     * already accounted for in gmt_doy
     */
    smmr_time += prior_leap_days[scan->gmt_year - MIN_GMT_YEAR]*SECS_PER_DAY;

    smmr_time += scan_start_time;

    scan->smmr_time = smmr_time;

    scan->gmt_seconds = scan_start_time;

  }

  /*
   * normalize date-time, again!
   */
  if (scan->gmt_seconds >= SECS_PER_DAY) {
    scan->gmt_seconds -= SECS_PER_DAY;
    scan->gmt_doy += 1;
    if (scan->gmt_doy > 365 + leap_day) {
      scan->gmt_doy -= 365 + leap_day;
      scan->gmt_year += 1;
    }
  }

  return TRUE;

error_return:
  scan->gmt_year = scan->gmt_doy = scan->gmt_seconds = 0;
  scan->smmr_time = 0;
  return FALSE;
}
