/*======================================================================
 * gsconcat - concatenate generic swath orbit files into files defined
 *            by time.  Input gswath records are not modified in this 
 *            application.
 *
 * 02/16/1998 brodzik brodzik@kryos.colorado.edu 303-492-8263
 * National Snow & Ice Data Center, University of Colorado, Boulder
 * Copyright (C) 1998-2004 University of Colorado
 *======================================================================*/
static const char gsconcat_c_rcsid[] = "$Header: /NSIDC_CVS/FILES/generic_swath/src/gsconcat/gsconcat.c,v 1.23 2004/01/29 23:19:07 brodzik Exp $";

#include "define.h"
#include "gswath.h"
#include "gslicer.h"
#include "pmtime.h"

#define usage \
"usage: gsconcat [-dtv] [-b begin_date] [-e end_date]\n"\
"                [-p period] [-s hhmmss] [-o output_directory] \n"\
"                [-i] [orbit_file...]\n"\
"\n"\
"gsconcat $Revision: 1.23 $ \n"\
" input : orbit_file - orbit files with generic swath data records, sorted in time order\n"\
"\n"\
" output : orbit files subsetted by time\n"\
"          output filenames will be written to stdout\n"\
"\n"\
" options :\n"\
"         b yyyymmdd - record date to begin (default is beginning record)\n"\
"         d - delete input files after processing\n"\
"         e yyyymmdd - record date to end (default is ending record)\n"\
"         i - read input file names from stdin\n"\
"         o output_directory - directory for output files (default is .)\n"\
"         p period - subsetting time period (in seconds)\n"\
"         s hhmmss - time of day to begin subset\n"\
"                    (defaults are in gswath.h)\n"\
"         t - subset by orbit (this option is not yet implemented)\n"\
"             this option overrides -p\n"\
"         v verbose mode (may be repeated)\n"\
"\n"\
" notes: Records are processed sequentially.  Any out of sequence records\n"\
"        will be warned about and skipped.\n"\
"\n"

static void starting_params(byte4 *begin_yyyymmdd, byte4 *begin_hhmmss,
			    GSwath *sensor_data, byte4 *start_timecode,
			    int *subset_length);

main ( int argc,  char *argv[])
{
  int scan;
  int i;
  int channel[GS_MAX_CHANNELS];
  char infilename[FILENAME_MAX];
  FILE *infile;
  char *outfiledir;
  char *option;
  byte4 begin_yyyymmdd, end_yyyymmdd;
  bool do_delete;
  bool verbose;
  bool very_verbose;
  bool do_stdin;
  bool subset_by_orbit;
  bool status;
  byte4 start_timecode;
  int subset_length; 
  byte4 begin_hhmmss;
  byte4 prev_rec_timecode;
  GS_type swath_type, prev_swath_type;
  GSwath *sensor_data;
  byte4 total_scans;
  gslicer_class *gslicer;
  gslicer_time_comparison record_place;
  byte4 num_invalid_gswaths;
  byte4 num_outofsequence_gswaths;
  byte4 num_duplicate_gswaths;

  /*
   *      set defaults
   */
  do_delete = FALSE;
  verbose = FALSE;
  very_verbose = FALSE;
  do_stdin = FALSE;
  subset_by_orbit = FALSE;
  subset_length = GSLICER_DEFAULT_SUBSET_LENGTH_SEC;
  begin_yyyymmdd = default_begin_yyyymmdd_gslicer();
  end_yyyymmdd = default_end_yyyymmdd_gslicer();
  begin_hhmmss = GSLICER_DEFAULT_BEGIN_HHMMSS;
  prev_rec_timecode=0;
  prev_swath_type=GS_TYPE_UNKNOWN;
  swath_type=GS_TYPE_UNKNOWN;
  total_scans=0;
  gslicer=NULL;
  sensor_data=NULL;
  outfiledir=NULL;
  num_invalid_gswaths=0;
  num_outofsequence_gswaths=0;
  num_duplicate_gswaths=0;

  /*
   * 	get command line options (if any)
   */

  while (--argc > 0 && (*++argv)[0] == '-') {
    for (option = argv[0]+1; *option != '\0'; option++) {
      switch (*option) {
      case 'b':
	++argv; --argc;
	if (0 >= argc) error_exit(usage);
	if (1 != sscanf(*argv, "%8lu", &begin_yyyymmdd))
	  error_exit(usage);
	break;
      case 'd':
	do_delete = TRUE;
	break;
      case 'e':
	++argv; --argc;
	if (0 >= argc) error_exit(usage);
	if (1 != sscanf(*argv, "%8lu", &end_yyyymmdd))
	  error_exit(usage);
	break;
      case 'i':
	do_stdin = TRUE;
	break;
      case 'o':
	++argv; --argc;
	if (0 >= argc) error_exit(usage);
	outfiledir = strdup(*argv);
	break;
      case 'p':
	++argv; --argc;
	if (0 >= argc) error_exit(usage);
	if (1 != sscanf(*argv, "%d", &subset_length))
	  error_exit(usage);
	break;
      case 's':
	++argv; --argc;
	if (0 >= argc) error_exit(usage);
	if (1 != sscanf(*argv, "%6lu", &begin_hhmmss))
	  error_exit(usage);
	break;
      case 't':
	subset_by_orbit = TRUE;
	fprintf (stderr,
		 "gsconcat: WARNING ignoring option -t: not yet implemented\n");
	subset_by_orbit = FALSE;
	break;
      case 'v':
	if (verbose) very_verbose = TRUE;
	verbose=TRUE;
	break;
      case 'V':
	fprintf(stderr,"%s\n", gsconcat_c_rcsid);
	break;
      default:
	fprintf (stderr, "gsconcat: invalid option %c\n", *option);
	error_exit(usage);
      } /* end switch on *option */
    } /* end loop for each option found */
  } /* end loop while options are on input line */

  if (!do_stdin && argc < 1)
    {
      fprintf(stderr,"No input files entered.  At least one input file is required.\n");
      error_exit(usage);
    }

  if (NULL == outfiledir) outfiledir = strdup(".");

  if (verbose) {
    fprintf(stderr,  "> Beginning gsconcat processing for:\n");
    fprintf(stderr,  ">  -b begin_date=%lu (defaults to first input record)\n",begin_yyyymmdd);
    fprintf(stderr,  ">  -d do_delete =%d\n",do_delete);
    fprintf(stderr,  ">  -e end_date=%lu (defaults to last input record)\n",end_yyyymmdd);
    fprintf(stderr,  ">  -i use stdin =%d\n",do_stdin);
    fprintf(stderr,  ">  -o outfiledir   =%s\n",outfiledir);
    if (GSLICER_DEFAULT_SUBSET_LENGTH_SEC == subset_length) 
      fprintf(stderr,">  subset_length will be determined by input swath type.\n");
    else
      fprintf(stderr,">  -p subset_length =%d seconds\n",subset_length);
    if (GSLICER_DEFAULT_BEGIN_HHMMSS == begin_hhmmss) 
      fprintf(stderr,">  subset_time_of_day will be determined by input swath type.\n");
    else
      fprintf(stderr,">  -s subset_time_of_day =%06lu seconds\n",
	      begin_hhmmss);
    fprintf(stderr,  ">  -t subset_by_orbit =%d\n",subset_by_orbit);
    fprintf(stderr,  "> $Id: gsconcat.c,v 1.23 2004/01/29 23:19:07 brodzik Exp $\n");
  } /* end if verbose */

  /* Loop for each input file */
  for (;;) {

    /* Get the next input file name */
    if (do_stdin) {
      if (NULL == fgets(infilename,FILENAME_MAX,stdin)) 
	goto cleanup_exit;
      /* fgets returns a final newline, so strip it off */
      /* (this is different from gets which does this for you) */
      *(infilename+strlen(infilename)-1)='\0';
    } else {
      if (1 > argc) goto cleanup_exit;
      assert(FILENAME_MAX > strlen(*argv));
      strncpy(infilename, *argv,FILENAME_MAX-1);
      --argc; ++argv;
    }

    /* 
     * Find out GS_type of the records in this file,
     * and verify that all input files are the same type
     * Skip and warn about any files that are a different type
     * than the starting type
     */
    swath_type = get_type_from_name_gswath(infilename,verbose);
    if (GS_TYPE_UNKNOWN == swath_type) {
      fprintf(stderr,
     	      "gsconcat warning: skipping %s, unknown swath type\n",
	      infilename);
      continue;
    }
    else if (GS_TYPE_UNKNOWN == prev_swath_type) {
      prev_swath_type = swath_type;
    } 
    else if (prev_swath_type != swath_type) {
      fprintf(stderr,    
     	      "gsconcat warning: skipping %s, swath type %s different from %s\n",
	      infilename,GS_type_string[swath_type],
	      GS_type_string[prev_swath_type]);
      continue;
    }

    /* 
     * Allocate a sensor data record, if necessary 
     * and initialize the channel array to all bands for this swath_type
     */
    if (NULL == sensor_data) {
      sensor_data = new_gswath(swath_type);
      if (NULL == sensor_data) goto cleanup_error_exit;
      if (very_verbose) fprintf(stderr,">> allocated new gswath object\n");
      for (i=0; i<sensor_data->info->num_channels; i++) {
     	channel[i]=i;
      }
    }

    /* Open this input file */
    infile = fopen(infilename,"r");
    if (NULL == infile) {
      perror(infilename);
      goto cleanup_error_exit;
    } else if (verbose) 
      fprintf(stderr,"> opening input file %s\n",infilename);
    
    /* Loop for each scan in this file */
    for (scan=0; !feof(infile); scan++) {

      /* Read the next gswath record */
      status = read_gswath(sensor_data, scan, infile,
			   sensor_data->info->num_channels, channel);
      if (feof(infile)) break;
      if (!status) {
	fprintf(stderr,"gsconcat: error reading next gswath from %s\n",
		infilename);
	goto cleanup_error_exit;
      }
      if (very_verbose && 0 == scan) 
	fprintf(stderr,">> first scan yyyydoy=%04d%03d, sec=%f\n",
		sensor_data->gmt_year,sensor_data->gmt_doy,
		sensor_data->gmt_seconds);

#ifdef DEBUG
      if (very_verbose) {
	fprintf(stderr,"DEBUG>> next scan=%d, yyyydoy=%04d%03d, sec=%f, timecode=%ld, status=%d\n",
		scan,sensor_data->gmt_year,sensor_data->gmt_doy,
		sensor_data->gmt_seconds,sensor_data->timecode,
		sensor_data->status);
      }
#endif

      /*
       * ignore records with zero timecodes or gmt_seconds=-1
       */
      if (0 == sensor_data->timecode || -1 == sensor_data->gmt_seconds) {

#ifdef DEBUG
	if (very_verbose) fprintf(stderr,"DEBUG>> Ignoring this gswath due to invalid time information\n");
#endif

	num_invalid_gswaths++;
	continue;
      }

      /* If there is no gslicer yet, initialize a new one */
      if (NULL == gslicer) {

	starting_params(&begin_yyyymmdd,&begin_hhmmss,sensor_data,
			&start_timecode,&subset_length);

       	gslicer = init_gslicer(outfiledir,start_timecode,subset_length,
			       infilename,sensor_data->type,verbose);
	if (verbose) {
	  fprintf(stderr,"> gslicer init'd for yyyymmdd=%lu, hhmmss=%lu, subset_length=%d secs\n",
		  begin_yyyymmdd, begin_hhmmss, subset_length);
	}
      }

      /* Check for records out of sequence */
      if (sensor_data->timecode == prev_rec_timecode) {
	fprintf(stderr,
		"gsconcat warning: duplicate record times found in %s\n",
		infilename);
	fprintf(stderr,
		"...skipping second record (scan=%d)\n",scan);
	fprintf(stderr," duplicate timecode = %ld\n",prev_rec_timecode);
	num_duplicate_gswaths++;
	continue;
      } else if (sensor_data->timecode < prev_rec_timecode) {
	fprintf(stderr,
		"gsconcat warning: record times out of sequence in %s\n",
		infilename);
	fprintf(stderr," skipping record with timecode=%ld (at scan=%d): precedes last sequential timecode=%ld\n",
		sensor_data->timecode,scan,prev_rec_timecode);
	num_outofsequence_gswaths++;
	continue;
      }
      prev_rec_timecode = sensor_data->timecode;
      
      /* 
       * Loop to iterate on gslicer's gsfile
       * (Assumes input records are sequential.)
       * If record is after the current gsfile, opens new gsfiles in 
       * sequence until the enclosing one is found that this record belongs in
       */
      for (;;) {
	
	record_place = compare_time_to_gslicer(gslicer,sensor_data->timecode);
       	if (LATER_THAN_GSFILE == record_place) {

	  /* 
	   * if the end of gslicer's gsfile is after
	   * end_yyyymmdd, we're done
	   */
	  if (default_end_yyyymmdd_gslicer() != end_yyyymmdd) {
	    if (gslicer->end_timecode >
		gregorian_to_timecode_pmtime (end_yyyymmdd, (byte4)0, 
					      gslicer->swath_type)) {
	      total_scans = total_scans + scan + 1;
	      goto cleanup_exit;
	    }
	  }

	  /* Otherwise, have gslicer go to the next gsfile */
	  status = next_gslicer(gslicer,stdout,verbose);
	  if (!status) goto cleanup_error_exit;
       	} else {
       	  break;
       	}
      } /* end loop to iterate on catfiles */
      
      /* Skip any records prior to gslicer's current gsfile */
      if (PRIOR_TO_GSFILE != record_place) {
	status = append_gslicer(gslicer,sensor_data,channel);
	if (!status) {
	  fprintf(stderr,"gsconcat:  error appending to gsfile\n");
	  goto cleanup_error_exit;
	} 
      }

    } /* end loop for each gswath record */

    total_scans = total_scans + scan + 1;

    fclose(infile);
    infile = NULL;
    if (very_verbose) fprintf(stderr,">> closing input file %s\n",infilename);
    if (do_delete) {
      remove(infilename);
      if (verbose) fprintf(stderr,
			   "> deleting input file %s\n",infilename);
    }

  } /* end loop for each input file */ 

cleanup_exit:

  free(sensor_data);
  if (NULL != infile) {
    fclose(infile);
    if (very_verbose) fprintf(stderr,">> closing input file %s\n",infilename);
  }
  if (do_delete && NULL != infilename) {
    remove(infilename);
    if (verbose) fprintf(stderr,
			 "> deleting infile %s\n",infilename);
  }
  if (NULL != gslicer) {
    if (very_verbose) fprintf(stderr,"> closing gsfile %s after writing %lu records.\n",
			      gslicer->filename,gslicer->num_gswaths);
    close_gslicer(gslicer,stdout);
  }
  
  if (verbose) {
    fprintf(stderr,"> %ld total gswaths found.\n",total_scans);
    fprintf(stderr,"> %ld (%5.1f%%) invalid gswaths found and ignored.\n",
	    num_invalid_gswaths,
	    (float)num_invalid_gswaths/(float)total_scans*100.);
    fprintf(stderr,"> %ld (%5.1f%%) duplicate timecode gswaths found and ignored.\n",
	    num_duplicate_gswaths,
	    (float)num_duplicate_gswaths/(float)total_scans*100.);
  }

  if (0 < num_outofsequence_gswaths) {
    fprintf(stderr,"************************Warning:**************************\n");
  }
  if (0 < num_outofsequence_gswaths || verbose) {
    fprintf(stderr,"  %ld (%5.1f%%) out of sequence gswaths found and ignored.\n",
	    num_outofsequence_gswaths,
	    (float)num_outofsequence_gswaths/(float)total_scans*100.);
  }

  if (verbose) {
    fprintf(stderr,"> End gsconcat.\n");
  }

  exit(EXIT_SUCCESS);

cleanup_error_exit:

  if (NULL != infile) fclose(infile);
  if (NULL != gslicer) {
    if (very_verbose) fprintf(stderr,"> closing gslicer with gsfile %s.\n",gslicer->filename);
    close_gslicer(gslicer,NULL);
  }

  fprintf(stderr,"gsconcat: Exiting in error.\n");
  exit(EXIT_FAILURE);

} /* end main */


/*----------------------------------------------------------------------
 *  starting_params - Determines starting parameters for first gslicer
 *  object.  Any user-defined input values will override defaults
 * 
 *     input : begin_yyyymmdd - beginning date of records in this
 *                              gslicer (if default, uses date in sensor_data)
 *             begin_hhmmss - beginning time of day of records in
 *                            this gslicer (if set to default for 
 *                            GS_TYPE_UNKNOWN, it will use default for the 
 *                            type of data in sensor_data)
 *             sensor_data - GSwath record
 *             subset_length - period (in seconds) of each time slice
 *                            (if set to default for GS_TYPE_UNKNOWN,
 *                            it will use default for the type of data
 *                            in sensor_data)
 *     output : 
 *             begin_yyyymmdd - beginning date corresponding to timecode
 *             begin_hhmmss - begin time of day corresponding to timecode
 *             start_timecode - starting timecode
 *             subset_length - period (in seconds) of each time slice
 * 
 *    result: TRUE iff success
 *    note: 
 *          1) Throws assertion for NULL sensor_data, or if sensor_data->type
 *             is out of range
 *          2) If user inputs an illogical begin_hhmmss, prints a
 *             warning to stderr, but continues to use it
 *
 *----------------------------------------------------------------------*/
static void starting_params(byte4 *begin_yyyymmdd, byte4 *begin_hhmmss,
			    GSwath *sensor_data, byte4 *start_timecode,
			    int *subset_length) {
  int yyyy;
  byte2 month,day;
  byte2 hh,mm,ss;

  assert(NULL != sensor_data);
  assert(GS_TYPE_SMMR_PF <= sensor_data->type &&
 	 sensor_data->type <= NUM_GS_TYPE);

  /* 
   * Determine the starting timecode from whatever is input
   * if no begin_yyyymmdd, use the date of the sensor_data
   * if no begin_hhmmss, use default for the type of data in this swath
   * if no subset_length, use default for the type of data in this swath
   */
  if (default_begin_yyyymmdd_gslicer() == *begin_yyyymmdd) {
    
    /* Get a yyyymmdd from sensor_data's year and doy */
    yyyy = sensor_data->gmt_year;
    yyyy = (yyyy < 100 ? yyyy + 1900 : yyyy);
    ydoy_to_ymd_pmtime(yyyy,(byte2) sensor_data->gmt_doy,&month,&day);
    *begin_yyyymmdd = (yyyy * 10000) + (month * 100) + day;

  }
   
  if (GSLICER_DEFAULT_BEGIN_HHMMSS == *begin_hhmmss) {
    
    *begin_hhmmss = GS_daily_begin_hhmmss[sensor_data->type];

    /* 
     * Back the begin day up by 1 if this record time falls between
     * midnight and the default for SMMR records
     */
    if (GS_TYPE_SMMR_PF == sensor_data->type) {

      parse_value_pmtime(*begin_hhmmss,&hh,&mm,&ss);
      if (sensor_data->gmt_seconds < 
	  ((hh * PMTIME_SECS_PER_HOUR) + (mm * PMTIME_SECS_PER_MINUTE) + ss)) {
	day_offset_pmtime(begin_yyyymmdd,-1);
      }
    }
  } else {

    /* Print a warning if input begin_hhmmss doesn't seem logical */
    parse_value_pmtime(*begin_hhmmss,&hh,&mm,&ss);
    if (0 > *begin_hhmmss ||
	PMTIME_HOURS_PER_DAY <= hh ||
	PMTIME_MINUTES_PER_HOUR <= mm ||
	PMTIME_SECS_PER_MINUTE <= ss )
      fprintf(stderr,"gsconcat WARNING: using suspicious input value -s hhmmss=%lu\n",
	      *begin_hhmmss);
  }

  *start_timecode = gregorian_to_timecode_pmtime (*begin_yyyymmdd, 
						  *begin_hhmmss,
						  sensor_data->type);

  if (GSLICER_DEFAULT_SUBSET_LENGTH_SEC == *subset_length) {
    *subset_length = GS_daily_subset_length_sec[sensor_data->type];
  }

}

