/*======================================================================
 * gslicer.c - manages sliced subsets of generic swath files
 *
 * 02/22/1998 brodzik brodzik@kryos.colorado.edu 303-492-8263
 * National Snow & Ice Data Center, University of Colorado, Boulder
 * Copyright (C) 1998 University of Colorado
 *======================================================================*/
#include "define.h"
#include "pmtime.h"
#include "days.h"
#define gslicer_c_
#include "gslicer.h"

static const char gslicer_c_rcsid[] = "$Header: /NSIDC_CVS/FILES/generic_swath/src/gsconcat/gslicer.c,v 1.11 2004/01/29 23:18:46 brodzik Exp $";

const char *id_gslicer(void) {
  return gslicer_c_rcsid;
}

#define FILEID_LENGTH 22 /* fileid format is yyyymmdd_hhmmss with optional -ascan or -bscan for res5 files */


#define DEFAULT_BEGIN_DATE 0
#define DEFAULT_END_DATE BYTE4_MAX

static const char *scan_id_string[] = 
{ "",
  "-ascan",
  "-bscan"
};

static void build_fileid(const byte4 timecode, const char *infilename,
			 const GS_type swath_type, char *fileid);
static bool open_gsfile(gslicer_class *this, bool verbose);

/*----------------------------------------------------------------------
 *  init_gslicer - initialize a gslicer manager
 *
 *    input: pathname - directory to locate the generic
 *                      swath (gs) file to be managed
 *           start_timecode - beginning swath time (seconds) for 
 *                            records in the gsfile
 *           subset_length - duration (in seconds) following
 *                       start_timecode for records to be included in the
 *                       gsfile
 *           swath_type - generic swath type of these records.
 *           verbose - boolean verbose flag
 *
 *    result: pointer to new instance of gslicer
 *            or NULL in case of error
 *
 *----------------------------------------------------------------------*/
gslicer_class *init_gslicer(const char *pathname, 
			    const byte4 start_timecode,
			    const byte4 subset_length,
			    const char *infilename,
			    const GS_type swath_type,
			    bool verbose) {

  gslicer_class *this;
  char *ptr;

  if (verbose) fprintf(stderr,"> %s\n", id_gslicer);

  /*
   * Allocate storage for and initialize new gslicer
   */
  this = (gslicer_class *) calloc(1, sizeof(gslicer_class));
  if (NULL == this) {
    perror("init_gslicer: gslicer");
    return NULL;
  }

  this->pathname = strdup(pathname);
  if (NULL == this->pathname) {
    perror("init_gslicer: pathname string");
    close_gslicer(this,NULL);
    return NULL;
  }

  this->fileid = (char *) calloc(FILEID_LENGTH, sizeof(char));
  if (NULL == this->fileid) {
    perror("init_gslicer: fileid string");
    close_gslicer(this,NULL);
    return NULL;
  }

  this->start_timecode = start_timecode;
  this->swath_type = swath_type;

  /* res5 files can be ascan or bscan, parse the filename for this info */
  this->scan_id = strdup(scan_id_string[0]);
  if (GS_TYPE_AMSR_RES5 == swath_type) {
      ptr = strrchr(infilename, '-');
      if (0 == strncmp(ptr,scan_id_string[1],strlen(scan_id_string[1]))) {
	  this->scan_id = strdup(scan_id_string[1]);
      } else if (0 == strncmp(ptr,scan_id_string[2],strlen(scan_id_string[2]))) {
	  this->scan_id = strdup(scan_id_string[2]);
      } else {
	  fprintf(stderr,"gslicer warning: cannot parse res5 file for scan type\n");
      }
  }
  build_fileid(this->start_timecode,this->scan_id,this->swath_type,this->fileid);

  this->filename = (char *) calloc(FILENAME_MAX, sizeof(char));
  if (NULL == this->filename) {
    perror("init_gslicer: filename string");
    close_gslicer(this,NULL);
    return NULL;
  }
  sprintf(this->filename,"%s/gs_%s%s",this->pathname,this->fileid,
	  GS_file_extension[swath_type]);


  this->end_timecode = this->start_timecode + subset_length;
  this->num_gswaths = 0;

  if (0 >= subset_length) {
    fprintf(stderr,"init_gslicer error: timecodes out of sequence, start=%ld, end=%ld\n",
	    this->start_timecode, this->end_timecode);
    close_gslicer(this,NULL);
    return NULL;
  }

  /* 
   * Open this file for writing
   */
  if (!open_gsfile(this,verbose)) {
    close_gslicer(this,NULL);
    return NULL;
  }

  /*
   * Return pointer to new object
   */
  return this;

}

/*----------------------------------------------------------------------
 *  compare_time_to_gslicer - determine whether the input time is
 *                          before, during or after the data that
 *                          are being managed in this gslicer
 * 
 *    input : this - gslicer
 *            timecode - timecode value
 * 
 *    result: indicator of timecode, relative to timecodes that
 *            belong in this gslicer (i.e. prior, during, after)
 *----------------------------------------------------------------------*/
gslicer_time_comparison compare_time_to_gslicer(gslicer_class *this, 
						const byte4 timecode) {

  assert(NULL != this);

  if (timecode < this->start_timecode) {
    return PRIOR_TO_GSFILE;
  } else if (timecode < this->end_timecode) {
    return DURING_GSFILE;
  } else {
    return LATER_THAN_GSFILE;
  }

}

/*----------------------------------------------------------------------
 *  append_gslicer - append a gswath record to the gsfile managed by
 *                 this gslicer
 * 
 *    input : this - gslicer
 *            gs - generic swath record to append to gslicer's gsfile
 *            channel - list of channels
 * 
 *    result: TRUE iff successful completion
 *            (checks that input gs time belongs in gslicer's current gsfile)
 * 
 *    effect: swath data in gs is appended to end of gslicer's gsfile
 *            an assertion will be thrown for:
 *               a) NULL this
 *----------------------------------------------------------------------*/
bool append_gslicer(gslicer_class *this, GSwath *gs, int channel[]) {

  assert(NULL != this);

  if (DURING_GSFILE != compare_time_to_gslicer(this,gs->timecode)) {
    fprintf(stderr,"append_gslicer error: input gs time=%ld does not belong in current gsfile [start=%ld, end=%ld)\n",
	    gs->timecode,this->start_timecode, this->end_timecode);
    close_gslicer(this,NULL);
    return FALSE;
  }

  /* Write this gswath record to this gslicer's generic swath file */
  if (write_gswath(gs,this->num_gswaths,this->fp,
		   gs->info->num_channels,channel)) {
    (this->num_gswaths)++;
    return TRUE;
  } else { 
    return FALSE; 
  }

}


/*----------------------------------------------------------------------
 *  next_gslicer - closes the current gsfile, and opens another in sequence
 * 
 *    input : this - current gslicer
 *            verbose - verbosity flag
 * 
 *    result: TRUE iff successful completion
 *    effect: current gslicer is closed and re-initialized to another,
 *            in sequence with the current gslicer.
 *    notes:
 *      1) If no records have been written to current gslicer's gsfile,
 *         no gsfile will be saved.
 * 
 *    assertions will be thrown for:
 *      a) NULL this
 *      b) this->start_timecode doesn't precede this->end_timecode
 *
 *----------------------------------------------------------------------*/
bool next_gslicer (gslicer_class *this, FILE *notify_fp, bool verbose) {

  byte4 timecode_diff;
  assert(NULL != this);
  assert(this->end_timecode > this->start_timecode);

  /* 
   * Close the old file stream, 
   * and delete it if there were no swaths written 
   * Reset num_gswaths.
   * If non-NULL notify stream, write the name of this file to it
   */
  fclose(this->fp);
  if (0 == this->num_gswaths) {
    remove(this->filename);
    if (verbose) fprintf(stderr,"next_gslicer> deleting empty gsfile %s\n",
			 this->filename);
  } else {
    if (verbose) 
      fprintf(stderr,
	      "next_gslicer> closing gsfile %s after writing %lu records.\n",
	      this->filename,this->num_gswaths);
    if (NULL != notify_fp) {
      if (isatty(fileno(notify_fp))) {
	fprintf(notify_fp,"Writing output gsfile %s\n",this->filename);
      } else {
	fprintf(notify_fp,"%s\n",this->filename);
      }
    }
  }
  this->num_gswaths = 0;

  /* Determine the new timecodes from the old ones */
  timecode_diff = this->end_timecode - this->start_timecode;
  this->start_timecode = this->end_timecode;
  this->end_timecode = this->start_timecode + timecode_diff;

  /* Determine the new fileid and filename from the new timecodes */
  build_fileid(this->start_timecode,this->scan_id,this->swath_type,this->fileid);
  sprintf(this->filename,"%s/gs_%s%s",this->pathname,this->fileid,
	  GS_file_extension[this->swath_type]);

  /* Open the new file for writing */
  if (!open_gsfile(this,verbose)) {
    close_gslicer(this,NULL);
    return FALSE;
  }

  return TRUE;
    
} 
 
/*----------------------------------------------------------------------
 *  close_gslicer - free storage associated with this object
 * 
 *    input: this - gslicer
 *           notify_fp - output file stream or NULL
 * 
 *    effect: deallocate all memory associated with this object
 *            if notify_fp is not NULL, the name of the closed file
 *               is output to it; if notify_fp is not a terminal, 
 *               output message is full name of external file just closed,
 *               otherwise, it is slightly more verbose.
 *
 *----------------------------------------------------------------------*/
void close_gslicer(gslicer_class *this, FILE *notify_fp) {

  if (NULL == this) return;
  if (NULL != this->fp) fclose(this->fp);
  if (NULL != this->filename) {
    if (0 == this->num_gswaths) {
      remove(this->filename);
    } else {
      if (NULL != notify_fp) {
	if (isatty(fileno(notify_fp))) {
	  fprintf(notify_fp,"Writing output gsfile %s\n",this->filename);
	} else {
	  fprintf(notify_fp,"%s\n",this->filename);
	}
      }
    }
  }
  if (NULL != this->filename) free(this->filename);
  if (NULL != this->fileid) free(this->fileid);
  if (NULL != this->pathname) free(this->pathname);
  free(this);

}

/*----------------------------------------------------------------------
 *  default_begin_yyyymmdd_gslicer - returns the nominal begin_yyyymmdd
 * 
 *    input : void
 * 
 *    result: placeholder for begin_yyyymmdd (will trigger gsfiles
 *            that begin with first input record)
 *----------------------------------------------------------------------*/
byte4 default_begin_yyyymmdd_gslicer(void) {

  return DEFAULT_BEGIN_DATE;

}

/*----------------------------------------------------------------------
 *  default_end_yyyymmdd_gslicer - returns the nominal end_yyyymmdd
 * 
 *    input : void
 * 
 *    result: placeholder for end_yyyymmdd (will trigger gsfiles
 *            that end with last input record)
 *----------------------------------------------------------------------*/
byte4 default_end_yyyymmdd_gslicer(void) {

  return DEFAULT_END_DATE;

}

/*---------------------------Local function definitions-----------------*/

/*----------------------------------------------------------------------
 *  build_fileid - Determines a generic swath fileid string from the 
 *                 input timecode and filename
 * 
 *    input : timecode - swath timecode
 *            scan_id - scan id string
 *            swath_type - swath identifier
 * 
 *    output: fileid - new fileid
 *----------------------------------------------------------------------*/
static void build_fileid(const byte4 timecode, const char *scan_id,
			 const GS_type swath_type, char *fileid) {

  long year, jdn;
  unsigned int month,day;
  
  jdn = ymdg_to_jdn(GS_epoch_year[swath_type],0,0) 
    + timecode/PMTIME_SECS_PER_DAY;
  jdn_to_ymdg(jdn,&year,&month,&day,NULL,NULL);
  sprintf(fileid,"%04u%02u%02u_%06ld%s",
	  year,month+1,day+1,
	  hours_to_hhmmss_pmtime ((double)(timecode % PMTIME_SECS_PER_DAY) / 
				  PMTIME_SECS_PER_HOUR),
	  scan_id);

}

/*----------------------------------------------------------------------
 *  open_gsfile - opens the generic swath file associated with this gslicer
 * 
 *    input : this - gslicer
 *            verbose - verbosity flag
 * 
 *    result : TRUE iff success
 *    effect: opens gslicer's generic swath file for writing
 *----------------------------------------------------------------------*/
static bool open_gsfile(gslicer_class *this, bool verbose) {

  /* 
   * Open a new file for writing
   */
  this->fp = fopen(this->filename, "wb");
  if (NULL == this->fp) {
    perror (this->filename);
    return FALSE;
  } else if (verbose) {
    fprintf(stderr,
	    "open_gsfile> opening generic swath file %s for writing %s records.\n",
	    this->filename,GS_type_string[this->swath_type]);
#ifdef DEBUG
    fprintf(stderr,
	    "open_gsfile> timecode interval: [%lu,%lu)\n",
	    this->start_timecode,this->end_timecode);
#endif
  }
  return TRUE;

}
