/*========================================================================
 * gsgrid - grid generic swath data
 *
 * 30-Nov-1995 K.Knowles knowlesk@kryos.colorado.edu 303-492-0644
 * National Snow & Ice Data Center, University of Colorado, Boulder
 * Copyright (C) 1995-2004 University of Colorado
 *========================================================================*/
#include "define.h"
#include "matrix.h"
#include "grids.h"
#include "maps.h"
#include "gswath.h"

static const char gsgrid_c_rcsid[] = "$Header: /NSIDC_CVS/FILES/generic_swath/src/gsgrid/gsgrid.c,v 1.23 2004/01/23 02:17:03 knowlesk Exp $";

#define usage "$Revision: 1.23 $\n"					   \
"usage: gsgrid [-g gpd_file] [-v] [-T test_mode]\n"			   \
"              [-f] [-w] [-p power] [-k kernel] \n"			   \
"              [-m min_weight] [-r max_radius]\n"			   \
"              [-c channels] [-AD] [-t time] [-d doy]\n"		   \
"              [-o output_name] [-i] [orbit_file...]\n"			   \
"\n"									   \
" input : orbit_file - orbit files\n"					   \
"               automatically determines type from file extension\n"	   \
"               skips any file which doesn't match type of first file\n"   \
"               all orbits will be placed into a single grid\n"		   \
"\n"									   \
" output: either band sequential on stdout or one file for each channel\n" \
"         output_name.ChP, where Ch=channel, P=polarization\n"		   \
"\n"									   \
" option:\n"								   \
"         g gpd_file - specify grid parameters (default .gpd)\n"	   \
"         v - verbose (may be repeated)\n"				   \
"         T test_mode - [g,h,d,s,w]\n"					   \
"         f - forward resampling\n"					   \
"         w - weighted interpolation\n"					   \
"         p power - 0=smooth, 6=sharp, 2=default \n"			   \
"         k kernel - force kernel size (rowsxcols)\n"			   \
"         m min_weight - minimum weight sum for interpolation\n"	   \
"         r max_radius - maximum radius for interpolation (in pixels)\n"   \
"         c channels - list of channels (0123456789AB)\n"		   \
"         AD - ascending or descending only\n"				   \
"         t time - select orbit by local time (decimal hours)\n"	   \
"         d doy - encode time relative to this day of year\n"		   \
"         o output_name - base name for files (default stdout)\n"	   \
"         i - get list of orbit files from stdin\n"			   \
"\n"									   \
" note: -f and -w options select interpolation method as follows:\n"	   \
"       default = nearest-neighbor\n"					   \
"       -w      = bilinear interpolation\n"				   \
"       -ww     = cubic convolution\n"					   \
"       -f      = drop-in-the-bucket averaging\n"			   \
"       -fw     = inverse distance weighted\n"				   \
"	-fww	= sin(x)/x convolution\n"				   \
"\n"

#define MAX_CHANNELS 12
#define UNDEFINED INT_MAX
#define BETA_MAX 9e9

#define MIN_CHANNELS_VALID_DATA 1
#define VV_INTERVAL 100
#define DEGREES_PER_HOUR ((float)360/24) /* sun synchronous orbit */
#define SECONDS_PER_HOUR ((float)60*60)
#define SECONDS_PER_DAY ((float)24*SECONDS_PER_HOUR)
#define DAYS_PER_YEAR(year) ((int)(year)%4 ? 365 : 366) /* 1987 to 2086 */
#define TEMP_SCALE .1

#define TB_FILL_VALUE 0
#define TIME_FILL_VALUE SHRT_MIN
#define INVALID_TIME_sec (60*SHRT_MAX)
#define MAX_TIME_sec (60*(SHRT_MAX-1)) /* maximum valid time in seconds */
#define MIN_TIME_sec (60*(SHRT_MIN+1)) /* minimum valid time in seconds */

/*
 * these time limits are +/-22.75 days
 * it's possible you might want to process more data than that
 * if so you probably don't care about the time
 * so we let it overflow and you get a warning
 * unless you want orbits segregated by local time
 * in which case a valid time is required therefore processing is halted
 */

static int verbose;

#ifdef DEBUG
static bool test_hist, test_grad, test_data, test_stat, test_beta;
#endif

#define HISTO_SIZE 3200
#define GRADO_SCALE .0025

/*------------------------------------------------------------------------
 * gradient - estimate magnitude of gradient at a point
 *
 *	input : scan - current sensor data record
 *		sample - position number
 *		band - band number
 *
 *	result: |gradient| in Kelvins per kilometer or -1 on error
 *
 *------------------------------------------------------------------------*/
static float gradient(GSwath *scan, int sample, int band)
{ float grad = -1;
  float dTdx, dTdy, Tb, Tx, Ty, dx, dy;
  int ss, sx, sy;

  switch (scan->type)
  { case GS_TYPE_SMMR_PF:
      assert(0 <= band && band <= 9);

/*
 *	determine neighbors
 */
      assert(0 <= sample && sample <= 93);
      ss = sample;
      sx = sample + 1;
      sy = 93 - sample;

/*
 *	correct geometry for edge of scan
 */
      if (sample == 46) { sx = 45; sy = 47; }
      if (sample == 93) { sx = 92; sy = 0; }

/*
 *	get partials
 */
      Tb = scan->pos[ss].temp[band];
      Tx = scan->pos[sx].temp[band];
      Ty = scan->pos[sy].temp[band];

      dx = arc_length_km(scan->pos[ss].lat, scan->pos[ss].lon,
			 scan->pos[sx].lat, scan->pos[sx].lon);

      dy = arc_length_km(scan->pos[ss].lat, scan->pos[ss].lon,
			 scan->pos[sy].lat, scan->pos[sy].lon);

      if (dx > 35.35 || dy > 35.35) /* diagonal of 25x25 */
      { return -1;
      }

      if (Tb != 0 && Tx != 0 && Ty != 0 && dx != 0 && dy != 0)
      { dTdx = (Tb - Tx)/dx;
	dTdy = (Tb - Ty)/dy;

/*
 *	get magnitude
 */
	grad = sqrt(dTdx*dTdx + dTdy*dTdy);
      }
      break;

    default:
      return -1;
    }

  return grad;
}

/*------------------------------------------------------------------------
 * delta_t_local - absolute local time difference in seconds
 *
 *	input : ref_tod - reference local time of day (seconds)
 *		gmt_tod - GMT time of day at point (seconds)
 *		lon_pt - longitude of point (decimal degrees)
 *
 *	result: delta between reference and point local time (seconds)
 *
 *------------------------------------------------------------------------*/
static int delta_t_local(int ref_tod, int gmt_tod, float lon_pt)
{ int delta, local_tod;

  if (UNDEFINED == gmt_tod) { return UNDEFINED; }

  local_tod = gmt_tod + ((lon_pt/DEGREES_PER_HOUR) * SECONDS_PER_HOUR);

  delta = ref_tod - local_tod;

/*
 *	delta is independent of date
 */
  while (delta < 0) { delta += SECONDS_PER_DAY; }
  while (delta >= SECONDS_PER_DAY) { delta -= SECONDS_PER_DAY; }

/*
 *	take closest way round
 */
  if (delta > SECONDS_PER_DAY/2) 
  { delta = SECONDS_PER_DAY - delta;
  }

  return delta;
}

/*------------------------------------------------------------------------
 * weight_function - interpolation weight as a function of
 *		     distance from center
 *	input : d2 - distance squared
 *		p1 - tuning parameter
 *
 *	result: weight
 *
 *------------------------------------------------------------------------*/
static double (*weight_function)(double d2, double p1);
static double weight_param;

static double drop_in_the_bucket(double d2, double one)
{ return one;
}

static double inverse_distance(double d2, double power)
{ return 1/pow(d2, power/2);
}

static double linear_interpolation(double d2, double sqrt_d2_max)
{ return sqrt_d2_max - sqrt(d2);
}

static double sin_x_over_x(double d2, double two_pi)
{ double x = two_pi*sqrt(d2);
  return (x == 0 ? 1 : sin(x)/x);
}

static double cubic_convolution(double d2, double p1)
{ double x = sqrt(d2);
  return (x == 0 ? 1 : 1/(x*x*x));
}

/*
 * QC counters
 */
static float total_pts=0;
static int bad_time=0, bad_temp=0, time_reject=0;
static int outside_grid=0, outside_d2_max=0;
static int under_min_weight=0, outside_roi=0;
static int temp_low[MAX_CHANNELS]={0,0,0,0,0,0,0,0,0,0,0,0};
static int temp_high[MAX_CHANNELS]={0,0,0,0,0,0,0,0,0,0,0,0};

/*
 *	debug variables
 *	use rdpix in idl to get values
 *	use assign in dbx to set values
 */
#ifdef DEBUG
  int x_rdpix=UNDEFINED, y_rdpix=UNDEFINED, test_rdpix=0;
#endif

int main (int argc, char *argv[])
{ int band, sample, scan, line;
  bool do_stdin, forward, weighted, wide_weighted, valid;
  bool ascending_only, descending_only, select_by_time;
  bool warn_about_time_overflow;
  int channel[MAX_CHANNELS];
  int col, row, k_cols, k_rows, nbands, nbytes, nparams;
  double dr, ds, d2, d2_max, power, weight, min_weight;
  double r, s, r1, s1, r2, s2;
  float Tb, gmt, max_radius;
  int tcode, time_select, start_doy, diff_days;
  float **grid[MAX_CHANNELS], **beta[MAX_CHANNELS];
  int **time, data_time;
  GSwath *sensor_data;
  GS_type swath_type;
  int2 *row_data;
  grid_class *grid_def;
  char *option, *channel_select, *test_mode;
  int newline;
  char *output_name, *gpd_filename;
  char orbit_filename[FILENAME_MAX], output_filename[FILENAME_MAX];
  char readln[MAX_STRING];
  FILE *gpd_file, *orbit_file=NULL, *test_file, *output_file;
#ifdef DEBUG
  int histo[MAX_CHANNELS][HISTO_SIZE];
  float histo_bin;
  int grado[MAX_CHANNELS][HISTO_SIZE];
  float grado_bin;
#endif

/*
 *	set defaults
 */
  gpd_filename = strdup(".gpd");
  forward = FALSE;
  weighted = FALSE;
  k_cols = k_rows = 0;
  power = 2;
  min_weight = 0;
  max_radius = 0;
  wide_weighted = FALSE;
  verbose = 0;
  test_mode = NULL;
#ifdef DEBUG
  test_grad = test_hist = test_data = test_stat = test_beta = FALSE;
#endif
  do_stdin = FALSE;
  channel_select = NULL;
  ascending_only = FALSE;
  descending_only = FALSE;
  select_by_time = FALSE;
  start_doy = UNDEFINED;
  warn_about_time_overflow = TRUE;
  output_name = NULL;
  output_file = NULL;

/*
 *	get command line options
 */
  while (--argc > 0 && (*++argv)[0] == '-')
  { for (option = argv[0]+1; *option != '\0'; option++)
    { switch (*option)
      { case 'g':
	  ++argv; --argc;
	  if (argc <= 0) { error_exit(usage); }
	  free(gpd_filename);
	  gpd_filename = strdup(*argv);
	  break;
	case 'f':
	  forward = TRUE;
	  break;
	case 'w':
	  if (weighted) { wide_weighted = TRUE; }
	  weighted = TRUE;
	  break;
	case 'k':
	  ++argv; --argc;
	  if (argc <= 0) { error_exit(usage); }
	  nparams = sscanf(*argv, "%dx%d", &k_rows, &k_cols);
	  if (0 == nparams) { error_exit(usage); }
	  if (1 == nparams) { k_cols = k_rows; }
	  break;
	case 'p':
	  ++argv; --argc;
	  if (argc <= 0) { error_exit(usage); }
	  if (sscanf(*argv, "%lf", &power) != 1) { error_exit(usage); }
	  break;
	case 'm':
	  ++argv; --argc;
	  if (argc <= 0) { error_exit(usage); }
	  if (sscanf(*argv, "%lf", &min_weight) != 1) { error_exit(usage); }
	  break;
	case 'r':
	  ++argv; --argc;
	  if (argc <= 0) { error_exit(usage); }
	  if (sscanf(*argv, "%f", &max_radius) != 1) { error_exit(usage); }
	  break;
	case 'c':
	  ++argv; --argc;
	  if (argc <= 0) { error_exit(usage); }
	  channel_select = strdup(*argv);
	  break;
	case 'A':
	  ascending_only = TRUE;
	  break;
	case 'D':
	  descending_only = TRUE;
	  break;
	case 't':
	  ++argv; --argc;
	  if (argc <= 0) { error_exit(usage); }
	  if (sscanf(*argv, "%f", &gmt) != 1) { error_exit(usage); }
	  time_select = gmt*SECONDS_PER_HOUR;
	  select_by_time = TRUE;
	  break;
	case 'd':
	  ++argv; --argc;
	  if (argc <= 0) { error_exit(usage); }
	  if (sscanf(*argv, "%d", &start_doy) != 1) { error_exit(usage); }
	  break;
	case 'v':
	  ++verbose;
	  break;
	case 'V':
	  fprintf(stderr,"%s\n", gsgrid_c_rcsid);
	  break;
	case 'T':
	  ++argv; --argc;
	  if (argc <= 0) { error_exit(usage); }
	  test_mode = strdup(*argv);
	  break;
	case 'o':
	  ++argv; --argc;
	  if (argc <= 0) { error_exit(usage); }
	  output_name = strdup(*argv);
	  break;
	case 'i':
	  do_stdin = TRUE;
	  break;
	default:
	  fprintf(stderr, "gsgrid: invalid option %c\n", *option);
	  error_exit(usage);
	}
    }
  }

/*
 *	check command line args
 */
  if (!do_stdin && argc < 1) error_exit(usage);

/*
 *	check for test mode flags
 */
#ifdef DEBUG
  if (test_mode)
  { if (strchr(test_mode, 'g')) test_grad = TRUE;
    if (strchr(test_mode, 'h')) test_hist = TRUE;
    if (strchr(test_mode, 'd')) test_data = TRUE;
    if (strchr(test_mode, 's')) test_stat = TRUE;
    if (strchr(test_mode, 'w')) test_beta = TRUE;

    if (test_data)
    { strcat(strcat(strcpy(output_filename,output_name),"."), "data");
      test_file = fopen(output_filename, "w");
      if (NULL == test_file) { perror(output_filename); error_exit("gsgrid: FATAL ERROR"); }
      fprintf(stderr, "> writing test data to: %s\n", output_filename);
    }
  }
#else
  if (test_mode) fprintf(stderr,"gsgrid: test mode not available\n");
#endif

/*
 *	take a peak at the first orbit file to determine the data type
 */
  swath_type = GS_TYPE_UNKNOWN;
  while (GS_TYPE_UNKNOWN == swath_type)
  { 
    if (do_stdin)
    { if (NULL == fgets(orbit_filename, FILENAME_MAX, stdin)) break;
      if (feof(stdin)) error_exit("gsgrid: FATAL ERROR");
      newline = strlen(orbit_filename) - 1;
      if (newline > 0 && newline < FILENAME_MAX) {
	orbit_filename[newline] = '\0';
      } else {
	continue;
      }
    }
    else
    { if (argc < 1) error_exit("gsgrid: FATAL ERROR");
      strcpy(orbit_filename, *argv);
      --argc; ++argv;
    }

    swath_type = get_type_from_name_gswath(orbit_filename, verbose);
  }

  if (verbose >= 2) {
    fprintf(stderr,">> processing %s data\n", GS_type_string[swath_type]);
    fprintf(stderr,">> skipping all files not ending in %s file extension\n",
	    GS_file_extension[swath_type]);
  }

/*
 *	allocate sensor data record
 */
  sensor_data = new_gswath(swath_type);
  if (!sensor_data) error_exit("gsgrid: FATAL ERROR");

  assert(sensor_data->info->num_channels <= MAX_CHANNELS);

/*
 *	select channels
 */
  if (channel_select)
  { if (verbose) fprintf(stderr, "> selected channels:");
    nbands = 0;
    for (option = channel_select; *option != '\0'; option++)
    { band = *option - '0';
    if (band < 0 || band > 9) band = *option - 'A' + 10;
    if (band < 0 || band > 36) band = *option - 'a' + 10;
      if (0 <= band && band < sensor_data->info->num_channels)
      { channel[nbands] = band;
	if (verbose) fprintf(stderr, " %s", 
			     sensor_data->info->channel_name[band]);
	++nbands;
      }
      else
      { fprintf(stderr,"gsgrid: bad channel selector %c\n", *option);
      }
    }
    if (verbose) fprintf(stderr,"\n");
  }
  else
  { if (verbose) fprintf(stderr, "> all %d channels\n", 
			 sensor_data->info->num_channels);
    nbands = sensor_data->info->num_channels;
    for (band = 0; band < nbands; band++)
    { channel[band] = band;
    }
  }

/*
 *	check ascending vs. descending selection
 */
  if (ascending_only && descending_only)
  { ascending_only = descending_only = FALSE;
  }

  if (verbose && (ascending_only || descending_only))
    fprintf(stderr,"> processing %s data only\n",
	    ascending_only ? "ASCENDING" : "DESCENDING");

  if (verbose && select_by_time) 
    fprintf(stderr,"> selecting orbit closest to %5.2f local time\n",
	    time_select/SECONDS_PER_HOUR);

  if (verbose && UNDEFINED != start_doy)
    fprintf(stderr,"> times relative to start of day %3.3d\n", start_doy);

/*
 *	initialize grid
 */
  grid_def = init_grid(gpd_filename);
  if (NULL == grid_def) { error_exit("gsgrid: FATAL ERROR"); }
  if (verbose) 
    fprintf(stderr,"> using %s\n>       %s\n", 
	    grid_def->gpd_filename, grid_def->mapx->mpp_filename);

/*
 *	allocate storage buffers:
 *	computation: (grid dimensions)
 *	  grid = accumulator for interpolation results
 *	  beta = interpolation weighting factors
 *	  time = time of day (GMT in seconds)
 *	output data: (single grid row)
 *	  row_data = scaled Tbs
 */
  if (verbose >= 2)
    fprintf(stderr, ">> allocating grid buffers...\n");

  for (band = 0; band < nbands; band++)
  { grid[band] = (float **)matrix(grid_def->rows, grid_def->cols, 
				  sizeof(float), matrix_ZERO);
    if (NULL == grid[band]) { error_exit("gsgrid: FATAL ERROR"); }

    beta[band] = (float **)matrix(grid_def->rows, grid_def->cols, 
				  sizeof(float), matrix_ZERO);
    if (NULL == beta[band]) { error_exit("gsgrid: FATAL ERROR"); }
  }

  time = (int **)matrix(grid_def->rows, grid_def->cols, 
			sizeof(int), matrix_ZERO);
  if (NULL == time) { error_exit("gsgrid: FATAL ERROR"); }

  for (row = 0; row < grid_def->rows; row++)
  { for (col = 0; col < grid_def->cols; col++)
    { time[row][col] = UNDEFINED;
    }
  }

  row_data = (int2 *)calloc(grid_def->cols, sizeof(int2));
  if (NULL == row_data) { perror("row_data"); error_exit("gsgrid: FATAL ERROR"); }

/*
 *	reverse navigation is faked by always going forward
 *	and keeping track of the nearest-neighbor, in this case
 *	beta is used to keep track of the closest distance so far
 */
  if (!forward && !weighted)
  { for (band = 0; band < nbands; band++)
    { for (row = 0; row < grid_def->rows; row++)
      { for (col = 0; col < grid_def->cols; col++)
	{ beta[band][row][col] = BETA_MAX;
	}
      }
    }
  }

#ifdef DEBUG
  if (test_hist) memset(histo, 0, MAX_CHANNELS*HISTO_SIZE*sizeof(int));
  if (test_grad) memset(grado, 0, MAX_CHANNELS*HISTO_SIZE*sizeof(int));
#endif

/*
 *	determine region of interest
 */
  gpd_file = fopen(grid_def->gpd_filename, "r");
  if (NULL == gpd_file) { perror(grid_def->gpd_filename); error_exit("gsgrid: FATAL ERROR"); }
  for (line = 1; line <= 5; line++) 
  { fgets(readln, MAX_STRING, gpd_file);
  }
  if (feof(gpd_file) ||
      sscanf(readln, "%lf%lf%lf%lf", &r1, &s1, &r2, &s2) != 4)
  { r1 = s1 = -0.5;
    r2 = grid_def->cols - 0.5;
    s2 = grid_def->rows - 0.5;
  }
  fclose(gpd_file);

  if (verbose)
    fprintf(stderr, "> region of interest [(%2.2f,%2.2f):(%2.2f,%2.2f)]\n",
	    r1, s1, r2, s2);

/*
 *	determine extent of kernel
 */
  if (0 == k_cols || 0 == k_rows)
  { k_cols = nint(sensor_data->info->nominal_res_km
		  / grid_def->mapx->scale/grid_def->cols_per_map_unit);
    if (k_cols < 1) k_cols = 1;
    k_rows = nint(sensor_data->info->nominal_res_km
		  / grid_def->mapx->scale/grid_def->rows_per_map_unit);
    if (k_rows < 1) k_rows = 1;
  }

  if (verbose)
    fprintf(stderr, "> %dx%d kernel\n", k_rows, k_cols);

/*
 *	set interpolation distance tolerance
 */
  dr = k_cols/2.0;
  ds = k_rows/2.0;

  d2_max = (max_radius ? max_radius*max_radius : BETA_MAX);

  if (verbose && do_stdin)
    fprintf(stderr, "> reading input file names from stdin\n");

/*
 *	choose interpolation method
 *	set weight function and tuning parameter
 */
  if (verbose) fprintf(stderr, "> binning method is ");

  if (!forward && !weighted && !wide_weighted)
  { if (verbose) fprintf(stderr, "nearest neighbor\n");
  }

  else if (!forward && weighted && !wide_weighted)
  { weight_function = linear_interpolation;
    weight_param = sqrt(d2_max);
    if (verbose) fprintf(stderr, "linear interpolation\n");
  }

  else if (!forward && weighted && wide_weighted)
  { weight_function = cubic_convolution;
    weight_param = 0;
    if (verbose) fprintf(stderr, "cubic convolution\n");
  }

  else if (forward && !weighted && !wide_weighted)
  { weight_function = drop_in_the_bucket;
    weight_param = 1;
    if (verbose) fprintf(stderr, "drop in the bucket\n");
  }

  else if (forward && weighted && !wide_weighted)
  { weight_function = inverse_distance;
    weight_param = power;
    if (verbose) fprintf(stderr, "inverse distance\n");
  }

  else if (forward && weighted && wide_weighted)
  { weight_function = sin_x_over_x;
    weight_param = 2*PI;
    if (verbose) fprintf(stderr, "sin(x)/x\n");
  }

/*
 *	process orbit files
 */
  for (;;)
  {
    if (verbose)
      fprintf(stderr,"> %s\n", orbit_filename);

    orbit_file = fopen(orbit_filename, "r");
    if (NULL == orbit_file) { perror(orbit_filename); goto next_orbit; }

/*
 *	process scans
 */
    for (scan = 0; !feof(orbit_file); scan++)
    { 
      read_gswath(sensor_data, scan, orbit_file, nbands, channel);
      if (feof(orbit_file)) break;
      if (ferror(orbit_file)) { perror(orbit_filename); break; }

      if (verbose >= 3 && scan && (0 == scan%VV_INTERVAL))
	fprintf(stderr,">> processing scan %4d: time %5.2f hours UTC\n", 
		scan, sensor_data->gmt_seconds/60/60);

      if (-1 == sensor_data->gmt_seconds)
      { ++bad_time;
	continue;
      }

/*
 *	get data time
 */
      if (UNDEFINED == start_doy) 
      { start_doy = sensor_data->gmt_doy;
	if (verbose)
	  fprintf(stderr,"> picking day %3.3d for start times\n", start_doy);
      }

      /*
       * watch out for roll over from end of one year to beginning of next
       */
      diff_days = sensor_data->gmt_doy - start_doy;

      if (diff_days < MIN_TIME_sec/SECONDS_PER_DAY) 
      { diff_days += DAYS_PER_YEAR(sensor_data->gmt_year - 1);
      }
      else if (diff_days > MAX_TIME_sec/SECONDS_PER_DAY)
      { diff_days -= DAYS_PER_YEAR(sensor_data->gmt_year);
      }

      data_time = sensor_data->gmt_seconds + diff_days*SECONDS_PER_DAY;

      if (data_time > MAX_TIME_sec || data_time < MIN_TIME_sec)
      { 
/*
 *	let time overflow and warn only once
 */
	if (warn_about_time_overflow)
	{ fprintf(stderr,"gsgrid: %s: time %d exceeds min/max %d/%d at scan %d\n",
		  orbit_filename, data_time/60, MIN_TIME_sec/60, MAX_TIME_sec/60, scan);
	  warn_about_time_overflow = FALSE;
	}

        data_time = INVALID_TIME_sec;

        if (select_by_time) { 
	  fprintf(stderr,"gsgrid: -t option requires valid time data\n");
	  error_exit("gsgrid: FATAL ERROR");
	}

      }


/*
 *	select ascending or descending
 */
      if ((descending_only && DES != sensor_data->orbit_dir) || 
	  (ascending_only  && ASC != sensor_data->orbit_dir)) 
	continue;

      total_pts += sensor_data->info->num_samples;

#ifdef DEBUG
	if (test_data)
	{ fprintf(test_file," %10d %6.2f %7.2f %8.1f %8.1f", 
		  sensor_data->timecode,
		  sensor_data->pos[sensor_data->info->center_pos].lat,
		  sensor_data->pos[sensor_data->info->center_pos].lon, 
		  r, s);
	  for (band = 0; band < nbands; band++)
	  { fprintf(test_file, " %7.2f", sensor_data->pos[sensor_data->info->center_pos].temp[band]);
	  }
	  fprintf(test_file, "\n");
	}
#endif

/*
 *	process samples
 */
      for (sample = 0; sample < sensor_data->info->num_samples; sample++)
      {
/*
 *	find corresponding grid coordinates
 */
	valid = forward_grid(grid_def, (double)sensor_data->pos[sample].lat, 
			     (double)sensor_data->pos[sample].lon, &r, &s)
	  && within_mapx(grid_def->mapx, sensor_data->pos[sample].lat,
			 sensor_data->pos[sample].lon);
	if (!valid) { ++outside_grid; continue;	}

/*
 *	check for region of interest
 */
	if (r < r1 || r >= r2 || s < s1 || s >= s2) 
	{ ++outside_roi; continue; }

/*
 *	spread value over kernel
 */
	for (row=(int)(s-k_rows/2+.5); row <= (int)(s+k_rows/2+.5); row++)
	{ 
	  if (row < 0 || row >= grid_def->rows) continue;

	  ds = s - row;

	  for (col=(int)(r-k_cols/2+.5); col <= (int)(r+k_cols/2+.5); col++)
	  { 
	    if (col < 0 || col >= grid_def->cols) continue;

	    dr = r - col;

#ifdef DEBUG
	    if (col == x_rdpix && row == (grid_def->cols-1 - y_rdpix))
	      ++test_rdpix;
#endif

/*
 *	check interpolation distance criterium
 */
	    d2 = dr*dr + ds*ds;

	    if (d2 > d2_max) 
	    { ++outside_d2_max;
	      continue;
	    }
	      
/*
 *	if appropriate, select orbit based on time
 */
	    if (select_by_time && time[row][col] != UNDEFINED)
	    {
	      assert(INVALID_TIME_sec != data_time);

	      if ((data_time - time[row][col]) 
		  > sensor_data->info->orbit_period_sec/2)
	      { 
/*
 *	different orbits, reject one or the other
 */
		++time_reject;

/*
 *	if the local time of the current sample is farther
 *	from the select time than the local time of the
 *	sample in this grid cell then chuck it (i.e. continue
 *	processing with the next sample, otherwise clear the 
 *	old value from the grid cell and go with the new value
 */
		if (delta_t_local(time_select, data_time, 
				  sensor_data->pos[sample].lon)
		    > delta_t_local(time_select, time[row][col], 
				    sensor_data->pos[sample].lon))
		{ continue;
		}
		else
		{ for (band = 0; band < nbands; band++)
		  { grid[band][row][col] = 0;
		    beta[band][row][col] = (forward || weighted ? 0 : BETA_MAX);
		  }
		}
	      }
	    }

/*
 *	determine appropriate weight based on distance from center
 */
	    if (forward || weighted)
	    { weight = weight_function(d2, weight_param);
	    }

/*
 *	get sample data and validate
 */
	    for (band = 0; band < nbands; band++)
	    { 
	      Tb = sensor_data->pos[sample].temp[band];

	      if (Tb < sensor_data->info->Tb_min[channel[band]])
	      { ++temp_low[channel[band]];
		continue;
	      }
	      if (Tb > sensor_data->info->Tb_max[channel[band]])
	      { ++temp_high[channel[band]];
		continue;
	      }

/*
 *	update nearest-neighbor or weighted sum
 */
	      if (!forward && !weighted)
	      { if (d2 > beta[band][row][col]) continue;
		grid[band][row][col] = Tb;
		beta[band][row][col] = d2;
	      }
	      else
	      { grid[band][row][col] += weight*Tb;
		beta[band][row][col] += weight;
	      }
/*
 *	update time
 */
	      time[row][col] = data_time;

#ifdef DEBUG
	      if (test_hist)
	      { histo_bin = Tb/TEMP_SCALE;
		if (histo_bin < 0) histo_bin = 0;
		if (histo_bin >= HISTO_SIZE) histo_bin = HISTO_SIZE-1;
		++histo[band][nint(histo_bin)];
	      }
	      if (test_grad)
	      { grado_bin = gradient(sensor_data,sample,band)/GRADO_SCALE;
		if (grado_bin < 0)
		{ if (verbose >= 2)
		    fprintf(stderr,"grad: bad loc: sample=%d scan=%d\n",
			    sample, scan);
		}
		else
		{ if (grado_bin >= HISTO_SIZE) grado_bin = HISTO_SIZE-1;
		  ++grado[band][nint(grado_bin)];
		}
	      }
#endif

	    } /* band */
/*
 *	could actually get the wrong time in the time
 *	grid by having some bad data in some channel
 *	that gets filled in by a subsequent scan
 */

	  } /* col */
	} /* row */

      } /* sample */
    } /* scan */

  next_orbit:
    if (orbit_file) fclose(orbit_file);
    orbit_file = NULL;
    
    if (do_stdin)
    { fgets(orbit_filename, FILENAME_MAX, stdin);
      if (feof(stdin)) break;
    }
    else
    { if (argc < 1) break;
      strcpy(orbit_filename, *argv);
      --argc; ++argv;
    }

    swath_type = get_type_from_name_gswath(orbit_filename, FALSE);

    if (swath_type != sensor_data->type)
    { fprintf(stderr,"gsgrid: skipping %s, wrong type\n", orbit_filename);
      goto next_orbit;
    }

  } /* orbit */

#ifdef DEBUG
  if (test_data) fclose(test_file);
#endif

/*
 *	normalize and output results
 */
  for (band = 0; band < nbands; band++)
  { 
    if (output_name)
    { if (output_file) fclose(output_file);
      sprintf(output_filename, "%s.%s", output_name, 
	      sensor_data->info->channel_name[channel[band]]);
      output_file = fopen(output_filename, "w");
      if (NULL == output_file) { perror(output_filename); error_exit("gsgrid: FATAL ERROR"); }
      if (verbose)
	fprintf(stderr, "> writing %s\n", output_filename);
    }
    else
    { strcpy(output_filename, "stdout");
      output_file = stdout;
    }

    if (forward || weighted)
    { for (row = 0; row < grid_def->rows; row++)
      { for (col = 0; col < grid_def->cols; col++)
	{ 
	  weight = beta[band][row][col];
	  if (verbose && weighted && weight < min_weight) ++under_min_weight;
	  row_data[col] = (weight > min_weight
			   ? nint(grid[band][row][col]/weight/TEMP_SCALE)
			   : TB_FILL_VALUE);
	}
        nbytes = fwrite(row_data, sizeof(int2), grid_def->cols, output_file);
	if (grid_def->cols != nbytes) 
	{ perror(output_filename); error_exit("gsgrid: FATAL ERROR"); }
      }
    }
    else
    { for (row = 0; row < grid_def->rows; row++)
      { for (col = 0; col < grid_def->cols; col++)
	{ 
	  d2 = beta[band][row][col];
	  row_data[col] = (d2 <= d2_max
			   ? nint(grid[band][row][col]/TEMP_SCALE)
			   : TB_FILL_VALUE);
	}
        nbytes = fwrite(row_data, sizeof(int2), grid_def->cols, output_file);
	if (grid_def->cols != nbytes) 
        { perror(output_filename); error_exit("gsgrid: FATAL ERROR"); }
      }
    }
  }

/*
 *	output time of day file
 */
  if (output_name)
  { if (output_file) { fclose(output_file); }
    strcat(strcat(strcpy(output_filename,output_name),"."), "tim");
    output_file = fopen(output_filename, "w");
    if (NULL == output_file) { perror(output_filename); error_exit("gsgrid: FATAL ERROR"); }
    if (verbose)
    { fprintf(stderr, "> writing %s\n", output_filename); }
  }
  else
  { strcpy(output_filename, "stdout");
    output_file = stdout;
  }

  for (row = 0; row < grid_def->rows; row++)
  { for (col = 0; col < grid_def->cols; col++)
    { tcode = time[row][col];
      row_data[col] = (tcode != UNDEFINED
		       ? nint(tcode/60) /* minutes */
		       : TIME_FILL_VALUE);
    }
    nbytes = fwrite(row_data, sizeof(int2), grid_def->cols, output_file);
    if (grid_def->cols != nbytes) 
    { perror(output_filename); error_exit("gsgrid: FATAL ERROR"); }
  }
  fclose(output_file);
  output_file = NULL;

#ifdef DEBUG
/*
 *	check for output name
 */
  if (!output_name && (test_beta || test_hist || test_grad))
    fprintf(stderr,"gsgrid: must set output name to get test output\n");

  if (output_name)
  {
/*
 *	output weight file
 */
    if (test_beta)
    { fprintf(stderr,"weight file is floatx%dx%dx%d\n",
	      grid_def->cols, grid_def->rows, nbands);
      if (output_file) { fclose(output_file); }
      strcat(strcat(strcpy(output_filename,output_name),"."), "beta");
      output_file = fopen(output_filename, "w");
      if (NULL == output_file) { perror(output_filename); error_exit("gsgrid: FATAL ERROR"); }
      fprintf(stderr, "> writing %s\n", output_filename);
      for (band = 0; band < nbands; band++)
      { for (row = 0; row < grid_def->rows; row++)
	{ fwrite(beta[0][row], sizeof(float), grid_def->cols, output_file);
	}
      }
      fclose(output_file);
      output_file = NULL;
    }

/*
 *	output histogram file
 */
    if (test_hist)
    { fprintf(stderr, "histogram files are int4x%dx%d\n", HISTO_SIZE, nbands);

      if (output_file) fclose(output_file);
      strcat(strcat(strcpy(output_filename,output_name),"."), "hist");
      output_file = fopen(output_filename, "w");
      if (NULL == output_file) { perror(output_filename); error_exit("gsgrid: FATAL ERROR"); }
      fprintf(stderr, "> writing %s\n", output_filename);
      for (band = 0; band < nbands; band++)
      { fwrite(histo[band], sizeof(int), HISTO_SIZE, output_file);
      }
      fclose(output_file);
    }

/*
 *	output gradogram file
 */
    if (test_grad)
    { fprintf(stderr, "gradogram files are int4x%dx%d\n", HISTO_SIZE, nbands);

      strcat(strcat(strcpy(output_filename,output_name),"."), "grad");
      output_file = fopen(output_filename, "w");
      if (NULL == output_file) { perror(output_filename); error_exit("gsgrid: FATAL ERROR"); }
      fprintf(stderr, "> writing %s\n", output_filename);
      for (band = 0; band < nbands; band++)
      { fwrite(grado[band], sizeof(int), HISTO_SIZE, output_file);
      }
      fclose(output_file);
      output_file = NULL;
    }
  }
#endif

/*
 *	report some stats
 */
#ifdef DEBUG
  if (test_stat)
  { fprintf(stderr, ">%10d total points processed\n", (int)total_pts);
    fprintf(stderr, ">%5d scans (%2d%%) skipped because of bad time code\n",
	    bad_time, (int)(100*(bad_time/total_pts)));
    fprintf(stderr, ">temp too low  ");
    bad_temp = 0;
    for (band = 0; band < MAX_CHANNELS; band++)
    { fprintf(stderr,"%5d", temp_low[band]);
      bad_temp += temp_low[band];
    }
    fprintf(stderr, ">\n");
    fprintf(stderr, ">temp too high ");
    for (band = 0; band < MAX_CHANNELS; band++)
    { fprintf(stderr,"%5d", temp_high[band]);
      bad_temp += temp_high[band];
    }
    fprintf(stderr, ">\n");
    fprintf(stderr, ">%5d spots (%2d%%) skipped due to bad temperature\n",
	    bad_temp, (int)(100*(bad_temp/total_pts/nbands)));

    fprintf(stderr, ">%5d spots (%2d%%) were rejected by local time\n",
	    time_reject, (int)(100*(time_reject/(k_rows*k_cols)/total_pts)));
    fprintf(stderr, ">%5d spots (%2d%%) were off the grid\n",
	    outside_grid, (int)(100*(outside_grid/total_pts)));
    fprintf(stderr, ">%5d spots (%2d%%) outside the region of interest\n",
	    outside_roi, (int)(100*(outside_roi/total_pts)));
    fprintf(stderr, ">%5d spots (%2d%%) were outside the max interp. radius\n",
	    outside_d2_max, (int)(100*(outside_d2_max/(k_rows*k_cols)/total_pts)));
    fprintf(stderr, ">%5d spots (%2d%%) were under the min. weight\n",
	    under_min_weight, (int)(100*(under_min_weight/total_pts)));
  }
#endif

  return EXIT_SUCCESS;
}
