/*========================================================================
 * gswath - generic swath interface
 *
 * 30-Nov-1995 K.Knowles knowlesk@kryos.colorado.edu 303-492-0644
 * National Snow & Ice Data Center, University of Colorado, Boulder
 * Copyright (C) 1995-2004 University of Colorado
 *========================================================================*/
#include "define.h"
#define gswath_c_
#include "gswath.h"
#include "struct_smmr.h"
#include "struct_ssmi.h"
#include "struct_amsr.h"
#include "pmtime.h"

static const char gswath_c_rcsid[]="$Header: /NSIDC_CVS/FILES/generic_swath/src/gswath/gswath.c,v 1.17 2004/01/23 01:57:57 knowlesk Exp $";

const char *id_gswath(void)
{
  return gswath_c_rcsid;
}

extern bool verbose, very_verbose;

static GS_orbit_dir orbit_direction(GSwath *this);

static bool read_smmr_pf(GSwath *this, int scan, FILE *fp, 
			 int nbands, int channel[]);

static bool write_smmr_pf(GSwath *this, int scan, FILE *fp, 
			  int nbands, int channel[]);

static bool read_ssmi_full(GSwath *this, int scan, FILE *fp, 
			   int nbands, int channel[]);

static bool write_ssmi_full(GSwath *this, int scan, FILE *fp, 
			    int nbands, int channel[]);

static bool read_ssmi_long(GSwath *this, int scan, FILE *fp, 
			   int nbands, int channel[]);

static bool write_ssmi_long(GSwath *this, int scan, FILE *fp, 
			    int nbands, int channel[]);

static bool read_amsr_res1(GSwath *this, int scan, FILE *fp, 
			   int nbands, int channel[]);

static bool write_amsr_res1(GSwath *this, int scan, FILE *fp, 
			    int nbands, int channel[]);

static bool read_amsr_res5(GSwath *this, int scan, FILE *fp, 
			   int nbands, int channel[]);

static bool write_amsr_res5(GSwath *this, int scan, FILE *fp, 
			    int nbands, int channel[]);

static const float UNDEFINED_LAT = -9e9;

/*
 *	macro to align memory components
 */
#define m_a_f ((int)sizeof(double))
#define mem_align(nbytes) (size_t)((((int)(nbytes)-1)/m_a_f + 1)*m_a_f)

/*------------------------------------------------------------------------
 * new_gswath - allocate new GSwath structure
 *
 *	input : type - specific swath type to create
 *
 *	result: pointer to new GSwath or NULL on failure
 *		can use stdlib free to deallocate memory
 *
 *------------------------------------------------------------------------*/
GSwath *new_gswath(GS_type type) {
  register int sample;
  int num_samples, num_channels;
  GSwath *this=NULL;
  GS_info *info=NULL;
  static GS_info *smmr_info=NULL;
  static GS_info *ssmi_full_info=NULL, *ssmi_long_info=NULL;
  static GS_info *amsr_res1_info=NULL, *amsr_res5_info=NULL;
  byte1 *block_ptr;
  size_t nbytes;

/*
 * invariant information
 * once the info block for each type of swath 
 * is malloced and initialized it is saved (static)
 * and re-used on subsequent calls
 */
  switch (type)
  { case GS_TYPE_SMMR_PF:
      if (NULL == smmr_info)
      { smmr_info = (GS_info *)malloc(sizeof(GS_info));
	if (NULL == smmr_info) { perror("new_gswath"); return NULL; }

	smmr_info->num_samples = SMMR_RECS_PER_SCAN;
	smmr_info->num_channels = SMMR_NUM_CHAN;
	smmr_info->center_pos = SMMR_RECS_PER_SCAN*3/4;
	smmr_info->one_scan_sec = SMMR_SEC_PER_SCAN;
	smmr_info->one_scan_deg = 360./SMMR_SCANS_PER_ORBIT;
	smmr_info->orbit_period_sec = SMMR_SEC_PER_ORBIT;
	smmr_info->orbit_declination = SMMR_DECLINATION;
	smmr_info->nominal_res_km = SMMR_NOMINAL_RES_km;
	smmr_info->Tb_min = smmr_Tb_min;
	smmr_info->Tb_max = smmr_Tb_max;
	smmr_info->channel_name = smmr_channel_name;
	smmr_info->read_swath = read_smmr_pf;
	smmr_info->write_swath = write_smmr_pf;
      }
      num_samples = smmr_info->num_samples;
      num_channels = smmr_info->num_channels;
      info = smmr_info;
      break;
    case GS_TYPE_SSMI_FULL:
      if (NULL == ssmi_full_info)
      { ssmi_full_info = (GS_info *)malloc(sizeof(GS_info));
	if (NULL == ssmi_full_info) { perror("new_gswath"); return NULL; }
      
	ssmi_full_info->num_samples = SSMI_RECS_PER_FULL_SCAN;
	ssmi_full_info->num_channels = SSMI_FULL_CHANNELS;
	ssmi_full_info->center_pos = SSMI_RECS_PER_FULL_SCAN/2;
	ssmi_full_info->one_scan_sec = 2*SSMI_SEC_PER_SCAN;
	ssmi_full_info->orbit_period_sec = SSMI_SEC_PER_ORBIT;
	ssmi_full_info->one_scan_deg = 
	  360/(ssmi_full_info->orbit_period_sec/ssmi_full_info->one_scan_sec);
	ssmi_full_info->orbit_declination = SSMI_DECLINATION;
	ssmi_full_info->nominal_res_km = SSMI_FULL_RES_km;
	ssmi_full_info->Tb_min = ssmi_full_Tb_min_Kelvins;
	ssmi_full_info->Tb_max = ssmi_full_Tb_max_Kelvins;
	ssmi_full_info->channel_name = ssmi_full_channel_name;
	ssmi_full_info->read_swath = read_ssmi_full;
	ssmi_full_info->write_swath = write_ssmi_full;
      }
      num_samples = ssmi_full_info->num_samples;
      num_channels = ssmi_full_info->num_channels;
      info = ssmi_full_info;
      break;
    case GS_TYPE_SSMI_LONG:
      if (NULL == ssmi_long_info)
      { ssmi_long_info = (GS_info *)malloc(sizeof(GS_info));
	if (NULL == ssmi_long_info) { perror("new_gswath"); return NULL; }
      
	ssmi_long_info->num_samples = SSMI_RECS_PER_LONG_SCAN;
	ssmi_long_info->num_channels = SSMI_LONG_CHANNELS;
	ssmi_long_info->center_pos = SSMI_RECS_PER_LONG_SCAN/2;
	ssmi_long_info->one_scan_sec = SSMI_SEC_PER_SCAN;
	ssmi_long_info->orbit_period_sec = SSMI_SEC_PER_ORBIT;
	ssmi_long_info->one_scan_deg = 
	  360/(ssmi_long_info->orbit_period_sec/ssmi_long_info->one_scan_sec);
	ssmi_long_info->orbit_declination = SSMI_DECLINATION;
	ssmi_long_info->nominal_res_km = SSMI_LONG_RES_km;
	ssmi_long_info->Tb_min = ssmi_long_Tb_min_Kelvins;
	ssmi_long_info->Tb_max = ssmi_long_Tb_max_Kelvins;
	ssmi_long_info->channel_name = ssmi_long_channel_name;
	ssmi_long_info->read_swath = read_ssmi_long;
	ssmi_long_info->write_swath = write_ssmi_long;
      }
      num_samples = ssmi_long_info->num_samples;
      num_channels = ssmi_long_info->num_channels;
      info = ssmi_long_info;
      break;
    case GS_TYPE_AMSR_RES1:
      if (NULL == amsr_res1_info)
      { amsr_res1_info = (GS_info *)malloc(sizeof(GS_info));
	if (NULL == amsr_res1_info) { perror("new_gswath"); return NULL; }
      
	amsr_res1_info->num_samples = AMSR_RECS_PER_RES1_SCAN;
	amsr_res1_info->num_channels = AMSR_RES1_CHANNELS;
	amsr_res1_info->center_pos = AMSR_RECS_PER_RES1_SCAN/2;
	amsr_res1_info->one_scan_sec = AMSR_SEC_PER_SCAN;
	amsr_res1_info->orbit_period_sec = AMSR_SEC_PER_ORBIT;
	amsr_res1_info->one_scan_deg = 
	  360/(amsr_res1_info->orbit_period_sec/amsr_res1_info->one_scan_sec);
	amsr_res1_info->orbit_declination = AMSR_DECLINATION;
	amsr_res1_info->nominal_res_km = AMSR_RES1_km;
	amsr_res1_info->Tb_min = &(amsr_Tb_min_Kelvins[AMSR_06V]);
	amsr_res1_info->Tb_max = &(amsr_Tb_max_Kelvins[AMSR_06V]);
	amsr_res1_info->channel_name = &(amsr_channel_name[AMSR_06V]);
	amsr_res1_info->read_swath = read_amsr_res1;
	amsr_res1_info->write_swath = write_amsr_res1;
      }
      num_samples = amsr_res1_info->num_samples;
      num_channels = amsr_res1_info->num_channels;
      info = amsr_res1_info;
      break;
    case GS_TYPE_AMSR_RES5:
      if (NULL == amsr_res5_info)
      { amsr_res5_info = (GS_info *)malloc(sizeof(GS_info));
	if (NULL == amsr_res5_info) { perror("new_gswath"); return NULL; }
      
	amsr_res5_info->num_samples = AMSR_RECS_PER_RES5_SCAN;
	amsr_res5_info->num_channels = AMSR_RES5_CHANNELS;
	amsr_res5_info->center_pos = AMSR_RECS_PER_RES5_SCAN/2;
	amsr_res5_info->one_scan_sec = AMSR_SEC_PER_SCAN/2.;
	amsr_res5_info->orbit_period_sec = AMSR_SEC_PER_ORBIT;
	amsr_res5_info->one_scan_deg = 
	  360/(amsr_res5_info->orbit_period_sec/amsr_res5_info->one_scan_sec);
	amsr_res5_info->orbit_declination = AMSR_DECLINATION;
	amsr_res5_info->nominal_res_km = AMSR_RES5_km;
	amsr_res5_info->Tb_min = &(amsr_Tb_min_Kelvins[AMSR_89V]);
	amsr_res5_info->Tb_max = &(amsr_Tb_max_Kelvins[AMSR_89V]);
	amsr_res5_info->channel_name = &(amsr_channel_name[AMSR_89V]);
	amsr_res5_info->read_swath = read_amsr_res5;
	amsr_res5_info->write_swath = write_amsr_res5;
      }
      num_samples = amsr_res5_info->num_samples;
      num_channels = amsr_res5_info->num_channels;
      info = amsr_res5_info;
      break;
    default:
      fprintf(stderr,"new_gswath: unknown type %d\n", type);
      return NULL;
  }

/*
 *	nbytes = size of block
 */
  nbytes = mem_align(sizeof(GSwath));

  nbytes += mem_align(sizeof(GS_nav));

  nbytes += mem_align(num_samples*sizeof(GSample));

  nbytes += num_samples*mem_align(num_channels*sizeof(float));

  block_ptr = (byte1 *)calloc(1, nbytes);
  if (!block_ptr) { perror("new_gswath"); return NULL; }

  this = (GSwath *)block_ptr;
  block_ptr += mem_align(sizeof(GSwath));

  this->nav = (GS_nav *)block_ptr;
  block_ptr += mem_align(sizeof(GS_nav));

  this->pos = (GSample *)block_ptr;
  block_ptr += mem_align(num_samples*sizeof(GSample));

  for (sample=0; sample < num_samples; sample++)
  { this->pos[sample].temp = (float *)block_ptr;
    block_ptr += mem_align(num_channels*sizeof(float));
  }

  assert(block_ptr == ((byte1 *)this + nbytes));

  this->nav->dir = UNKNOWN_DIR;
  this->nav->lat = UNDEFINED_LAT;
  this->nav->timecode = 0;;

  this->type = type;
  this->info = info;

  return this;
}

/*------------------------------------------------------------------------
 * get_type_from_name_gswath - GS_type from file name
 *
 *	input : orbit_filename - name of file
 *		verbose - flag
 *
 *	result: swath type based on file extension 
 *		or GS_TYPE_UNKNOWN
 *
 *------------------------------------------------------------------------*/
GS_type get_type_from_name_gswath(char *orbit_filename, bool verbose)
{ char *suffix;
  GS_type swath_type;

  suffix = strrchr(orbit_filename, '.');
  if (!suffix) { goto error_return; }

  for (swath_type = GS_TYPE_SMMR_PF; swath_type < NUM_GS_TYPE; swath_type++)
  { if (streq(suffix, GS_file_extension[swath_type])) break;
  }
  assert(swath_type < NUM_GS_TYPE);
  if (swath_type >= NUM_GS_TYPE) { goto error_return; }

  if (verbose) 
  { fprintf(stderr,"> processing %s\n", GS_type_string[swath_type]); }

  return swath_type;

 error_return:
  if (verbose) 
  { fprintf(stderr, "gswath: %s: can't find valid file extension\n",
	    orbit_filename);
  }
  return GS_TYPE_UNKNOWN;

}

/*------------------------------------------------------------------------
 * close_gswath - deallocate GSwath structure
 *
 *	input : this - generic swath structure
 *
 *------------------------------------------------------------------------*/
void close_gswath(GSwath *this)
{
  if (NULL == this) return;
  if (NULL != this->info) free(this->info);
  /*
   * nav and pos are part of same malloc as *this
   */
  free(this);
}

/*------------------------------------------------------------------------
 * read_gswath - read a scan into a generic swath structure
 *
 *	input : this - generic swath structure
 *		scan - scan line number
 *		fp - file pointer
 *		nbands - number of channels
 *		channel - list of channels
 *
 *	result: TRUE == success
 *
 *------------------------------------------------------------------------*/
bool read_gswath(GSwath *this, int scan, FILE *fp, int nbands, int channel[])
{ 
  assert(this && fp && channel);
  assert(scan >= 0);
  assert(nbands >= 0);

  return this->info->read_swath(this, scan, fp, nbands, channel);
}

/*------------------------------------------------------------------------
 * write_gswath - write a generic swath structure into a scan
 *
 *	input : this - generic swath structure
 *		scan - scan line number
 *		fp - file pointer
 *		nbands - number of channels
 *		channel - list of channels
 *
 *	result: TRUE == success
 *
 *------------------------------------------------------------------------*/
bool write_gswath(GSwath *this, int scan, FILE *fp, int nbands, int channel[])
{ 
  assert(this && fp && channel);
  assert(scan >= 0);
  assert(nbands >= 0);

  return this->info->write_swath(this, scan, fp, nbands, channel);
}

/*------------------------------------------------------------------------
 * copy_gswath - copy a scan
 *
 *	input : from - source data
 *
 *	output: to - destination
 *
 *------------------------------------------------------------------------*/
bool copy_gswath(GSwath *to, GSwath *from)
{ int band, sample;

  assert(to->info->num_samples == from->info->num_samples);
  assert(to->info->num_channels == from->info->num_channels);

  if (to->type != from->type) return FALSE;

  to->gmt_year = from->gmt_year;
  to->gmt_doy = from->gmt_doy;
  to->gmt_seconds = from->gmt_seconds;
  to->timecode = from->timecode;
  to->status = from->status;
  to->orbit_dir = from->orbit_dir;

  for (sample = 0; sample < to->info->num_samples; sample++) {
    to->pos[sample].lat = from->pos[sample].lat;
    to->pos[sample].lon = from->pos[sample].lon;
    for (band = 0; band < to->info->num_channels; band++)
    { to->pos[sample].temp[band] = from->pos[sample].temp[band];
    }
  }

  to->nav->dir = from->nav->dir;
  to->nav->lat = from->nav->lat;
  to->nav->timecode = from->nav->timecode;

  return TRUE;
}

/*------------------------------------------------------------------------
 * orbit_direction - get best guess at orbit direction 
 *		     for current scan based on previous scan
 *
 *	input : this - current scan
 *
 *	result: orbit direction
 *
 *------------------------------------------------------------------------*/
static GS_orbit_dir orbit_direction(GSwath *this)
{ GS_orbit_dir dir=UNKNOWN_DIR;
  float lat_center=UNDEFINED_LAT;
  GS_nav *prev;
  static const int MAX_SCANS_GAP = 4;
  float scan_gap_seconds, scan_gap_degrees;

/*
 *	get navigation information
 */
  prev = (GS_nav *)this->nav;

/*
 *	this approximation assumes that the orbit
 *	declination isn't too big (i.e. it's a 
 *	polar orbiting satellite) and that the 
 *	center of scan coincides with the ground track
 */

  lat_center = this->pos[this->info->center_pos].lat;

  scan_gap_seconds = MAX_SCANS_GAP*this->info->one_scan_sec;
  scan_gap_degrees = MAX_SCANS_GAP*this->info->one_scan_deg;

/*
 *	make sure there isn't a big gap of missing scans
 */
  if (0 == this->timecode
      || this->timecode - prev->timecode > scan_gap_seconds)
  { dir = UNKNOWN_DIR;
  }
  else
  { 
    if (lat_center < prev->lat 
	&& lat_center > -(90-fabs(this->info->orbit_declination)))
    { dir = DES;
    }
    else if (lat_center > prev->lat 
	     && lat_center < (90-fabs(this->info->orbit_declination)))
    { dir = ASC;
    }
    else
    { dir = prev->dir;
    }

  }

/*
 *	make sure orbit doesn't change direction in mid-latitudes
 */ 
  if ((ASC == prev->dir && DES == dir 
       && lat_center < (90-(fabs(this->info->orbit_declination)+scan_gap_degrees)))
      ||
      (DES == prev->dir && ASC == dir 
       && lat_center > -(90-(fabs(this->info->orbit_declination)+scan_gap_degrees))))
  { dir = UNKNOWN_DIR;
    fprintf(stderr,"gswath: prohibited orbit dir change at timecode %d\n",
	    this->timecode);
  }

  prev->timecode = this->timecode;
  prev->lat = lat_center;
  prev->dir = dir;

  return dir;
}

/*------------------------------------------------------------------------
 * read_smmr_pf - read SMMR Pathfinder data into generic swath
 *
 * all read/write routines must have same interface as read/write_gswath
 *
 *------------------------------------------------------------------------*/
static bool read_smmr_pf(GSwath *this, int scan, FILE *fp, 
			 int nbands, int channel[])
{ register int band, sample;
  int nbytes;
  static struct SMMR_REC smmr;

/*
 *	read next smmr record
 */
  if (fseek(fp, scan*sizeof(smmr), SEEK_SET) != 0) { return FALSE; }
  nbytes = fread(&smmr, 1, sizeof(smmr), fp);
  if (feof(fp) || ferror(fp) || sizeof(smmr) != nbytes) { return FALSE; }

/*
 *	decode smmr record
 */
  this->status = smmr.status;

  this->gmt_year = smmr.gmt_year;
  this->gmt_doy = smmr.gmt_doy;
  this->timecode = smmr.smmr_time;

  if (smmr.smmr_time)
  { this->gmt_seconds = smmr.gmt_seconds;
  }
  else
  { this->gmt_seconds = -1;
  }

  for (sample = 0; sample < this->info->num_samples; sample++)
  {
    this->pos[sample].lat 
      = smmr.full[sample].lat * SMMR_LAT_SCALE - SMMR_LAT_BIAS;
    this->pos[sample].lon 
      = smmr.full[sample].lon * SMMR_LON_SCALE - SMMR_LON_BIAS;

    for (band = 0; band < nbands; band++)
    { this->pos[sample].temp[band] 
	= smmr.full[sample].temp[channel[band]] * SMMR_TEMP_SCALE;
    }
  }

  this->orbit_dir = orbit_direction(this);

  return TRUE;
}

/*------------------------------------------------------------------------
 * write_smmr_pf - write generic swath into SMMR Pathfinder .smmr file
 *
 *------------------------------------------------------------------------*/
static bool write_smmr_pf(GSwath *this, int scan, FILE *fp, 
			 int nbands, int channel[])
{ register int band, sample;
  int nbytes;
  static struct SMMR_REC smmr;

/*
 *	encode smmr record
 */
  smmr.status = this->status;

  smmr.gmt_year = this->gmt_year;
  smmr.gmt_doy = this->gmt_doy;
  smmr.smmr_time = this->timecode;

  if (-1 != this->gmt_seconds)
  { smmr.gmt_seconds = this->gmt_seconds;
  }
  else
  { smmr.smmr_time = 0;
  }

  for (sample = 0; sample < this->info->num_samples; sample++)
  {
     
    smmr.full[sample].lat 
      = nint((this->pos[sample].lat + SMMR_LAT_BIAS) / SMMR_LAT_SCALE);
    smmr.full[sample].lon
      = nint((this->pos[sample].lon + SMMR_LON_BIAS) / SMMR_LON_SCALE);

    for (band = 0; band < nbands; band++)
    { smmr.full[sample].temp[channel[band]] 
	= nint(this->pos[sample].temp[band] / SMMR_TEMP_SCALE);
    }
  }

/*	This is a slightly different definition of ascending from the
 *	original data. There it indicates what's happening at the
 *	satellite sub-point.  Here it indicates what's happening at the
 *	center of swath.
 */
  smmr.ascending = (ASC == this->orbit_dir);

/*
 *	write smmr record
 */
  if (fseek(fp, scan*sizeof(smmr), SEEK_SET) != 0) { return FALSE; }
  nbytes = fwrite(&smmr, 1, sizeof(smmr), fp);
  if (feof(fp) || ferror(fp) || sizeof(smmr) != nbytes) { return FALSE; }

  return TRUE;
}

/*------------------------------------------------------------------------
 * read_ssmi_full - read SSM/I .full data into generic swath
 *
 *------------------------------------------------------------------------*/
static bool read_ssmi_full(GSwath *this, int scan, FILE *fp, 
			   int nbands, int channel[])
{ register int band, sample;
  byte4 nbytes, hhmmss, yyyymmdd;
  byte2 year, month, dom, doy;
  static struct SSMI_FULL_REC ssmi;

/*
 *	read next full record
 */
  if (fseek(fp, scan*sizeof(ssmi), SEEK_SET) != 0) { return FALSE; }
  nbytes = fread(&ssmi, 1, sizeof(ssmi), fp);
  if (feof(fp) || ferror(fp) || sizeof(ssmi) != nbytes) { return FALSE; }

/*
 *	decode full record
 */
  this->status = ssmi.tht;

  this->timecode = ssmi.time;

  if (0 != ssmi.time) {
    timecode_to_gregorian_pmtime(ssmi.time, &yyyymmdd, &hhmmss, GS_TYPE_SSMI_FULL);
    parse_value_pmtime(yyyymmdd, &year, &month, &dom);
    ymd_to_ydoy_pmtime(year, month, dom, &doy);
    this->gmt_year = year;
    this->gmt_doy = doy;
    this->gmt_seconds = hhmmss_to_hours_pmtime(hhmmss)*60*60;
  }
  else
  { this->gmt_seconds = -1;
  }

  for (sample = 0; sample < this->info->num_samples; sample++)
  {
    this->pos[sample].lat 
      = ssmi.full[sample].lat * SSMI_LAT_SCALE - SSMI_LAT_BIAS;
    this->pos[sample].lon 
      = ssmi.full[sample].lon * SSMI_LON_SCALE - SSMI_LON_BIAS;

    for (band = 0; band < nbands; band++)
    { this->pos[sample].temp[band] 
	= ssmi.full[sample].temp[channel[band]] * SSMI_TEMP_SCALE;
    }
  }

  this->orbit_dir = orbit_direction(this);

  return TRUE;
}

/*------------------------------------------------------------------------
 * write_ssmi_full - write generic swath to SSM/I .full file
 *
 *------------------------------------------------------------------------*/
static bool write_ssmi_full(GSwath *this, int scan, FILE *fp, 
			    int nbands, int channel[])
{ register int band, sample;
  long int nbytes;
  static struct SSMI_FULL_REC ssmi;

/*
 *	encode full record
 */
  ssmi.tht = this->status;

  ssmi.time = this->timecode;

  for (sample = 0; sample < this->info->num_samples; sample++)
  {
    ssmi.full[sample].lat 
      = nint((this->pos[sample].lat + SSMI_LAT_BIAS) / SSMI_LAT_SCALE);
    ssmi.full[sample].lon
      = nint((this->pos[sample].lon + SSMI_LON_BIAS) / SSMI_LON_SCALE); 

    for (band = 0; band < nbands; band++)
    { ssmi.full[sample].temp[channel[band]]
	= nint(this->pos[sample].temp[band] / SSMI_TEMP_SCALE);
    }
  }

/*
 *	write full record
 */
  if (fseek(fp, scan*sizeof(ssmi), SEEK_SET) != 0) { return FALSE; }
  nbytes = fwrite(&ssmi, 1, sizeof(ssmi), fp);
  if (feof(fp) || ferror(fp) || sizeof(ssmi) != nbytes) { return FALSE; }

  return TRUE;
}

/*------------------------------------------------------------------------
 * read_ssmi_long - read SSM/I .long data into generic swath
 *
 *------------------------------------------------------------------------*/
static bool read_ssmi_long(GSwath *this, int scan, FILE *fp, 
			   int nbands, int channel[])
{ register int band, sample;
  byte4 nbytes, hhmmss, yyyymmdd;
  byte2 year, month, dom, doy;
  static struct SSMI_LONG_REC ssmi;

/*
 *	read next long record
 */
  if (fseek(fp, scan*sizeof(ssmi), SEEK_SET) != 0) { return FALSE; }
  nbytes = fread(&ssmi, 1, sizeof(ssmi), fp);
  if (feof(fp) || ferror(fp) || sizeof(ssmi) != nbytes) { return FALSE; }

/*
 *	decode long record
 */
  this->status = 0;

  this->timecode = ssmi.time;

  if (0 != ssmi.time) {
    timecode_to_gregorian_pmtime(ssmi.time, &yyyymmdd, &hhmmss, GS_TYPE_SSMI_LONG);
    parse_value_pmtime(yyyymmdd, &year, &month, &dom);
    ymd_to_ydoy_pmtime(year, month, dom, &doy);
    this->gmt_year = year;
    this->gmt_doy = doy;
    this->gmt_seconds = hhmmss_to_hours_pmtime(hhmmss)*60*60;
  }
  else
  { this->gmt_seconds = -1;
  }

  for (sample = 0; sample < this->info->num_samples; sample++)
  {
    this->pos[sample].lat 
      = ssmi.lon8[sample].lat * SSMI_LAT_SCALE - SSMI_LAT_BIAS;
    this->pos[sample].lon 
      = ssmi.lon8[sample].lon * SSMI_LON_SCALE - SSMI_LON_BIAS;

    for (band = 0; band < nbands; band++)
    { this->pos[sample].temp[band] 
	= ssmi.lon8[sample].temp[channel[band]] * SSMI_TEMP_SCALE;
    }
  }

  this->orbit_dir = orbit_direction(this);

  return TRUE;
}

/*------------------------------------------------------------------------
 * write_ssmi_long - write generic swath to SSM/I .long file
 *
 *------------------------------------------------------------------------*/
static bool write_ssmi_long(GSwath *this, int scan, FILE *fp, 
			    int nbands, int channel[])
{ register int band, sample;
  long int nbytes;
  static struct SSMI_LONG_REC ssmi;

/*
 *	decode long record
 */
  ssmi.time = this->timecode;

  for (sample = 0; sample < this->info->num_samples; sample++)
  {
    ssmi.lon8[sample].lat
      = nint((this->pos[sample].lat + SSMI_LAT_BIAS) / SSMI_LAT_SCALE);
    ssmi.lon8[sample].lon
      = nint((this->pos[sample].lon + SSMI_LON_BIAS) / SSMI_LON_SCALE);

    for (band = 0; band < nbands; band++)
    { ssmi.lon8[sample].temp[channel[band]] 
	= nint(this->pos[sample].temp[band] / SSMI_TEMP_SCALE);
    }
  }

/*
 *	write long record
 */
  if (fseek(fp, scan*sizeof(ssmi), SEEK_SET) != 0) return FALSE;
  nbytes = fwrite(&ssmi, 1, sizeof(ssmi), fp);
  if (feof(fp) || ferror(fp) || sizeof(ssmi) != nbytes) return FALSE;

  return TRUE;
}

/*------------------------------------------------------------------------
 * read_amsr_res1 - read AMSR .res1 data into generic swath
 *
 *------------------------------------------------------------------------*/
static bool read_amsr_res1(GSwath *this, int scan, FILE *fp, 
			   int nbands, int channel[])
{ register int band, sample;
  byte4 nbytes, hhmmss, yyyymmdd;
  byte2 year, month, dom, doy;
  static struct AMSR_RES1_SCAN amsr;

/*
 *	read next record
 */
  if (fseek(fp, scan*sizeof(amsr), SEEK_SET) != 0) { return FALSE; }
  nbytes = fread(&amsr, 1, sizeof(amsr), fp);
  if (feof(fp) || ferror(fp) || sizeof(amsr) != nbytes) { return FALSE; }

/*
 *	decode res1 record
 */
  this->status = amsr.qflag;

  this->timecode = amsr.time;

  if (0 != amsr.time) {
    timecode_to_gregorian_pmtime(amsr.time, &yyyymmdd, &hhmmss, GS_TYPE_AMSR_RES1);
    parse_value_pmtime(yyyymmdd, &year, &month, &dom);
    ymd_to_ydoy_pmtime(year, month, dom, &doy);
    this->gmt_year = year;
    this->gmt_doy = doy;
    this->gmt_seconds = hhmmss_to_hours_pmtime(hhmmss)*60*60;
  }
  else {
    this->gmt_seconds = -1;
  }

  for (sample = 0; sample < this->info->num_samples; sample++) {

    this->pos[sample].lat = amsr.pos[sample].lat;
    this->pos[sample].lon = amsr.pos[sample].lon;

    for (band = 0; band < nbands; band++) {
      this->pos[sample].temp[band] 
	= amsr.pos[sample].temp[channel[band]] * AMSR_TEMP_SCALE;
    }
  }

  this->orbit_dir = orbit_direction(this);

  return TRUE;
}

/*------------------------------------------------------------------------
 * write_amsr_res1 - write generic swath to AMSR .res1 file
 *
 *------------------------------------------------------------------------*/
static bool write_amsr_res1(GSwath *this, int scan, FILE *fp, 
			    int nbands, int channel[])
{ register int band, sample;
  long int nbytes;
  static struct AMSR_RES1_SCAN amsr;

/*
 *	decode record
 */
  amsr.qflag = this->status;

  amsr.time = this->timecode;

  for (sample = 0; sample < this->info->num_samples; sample++) {

    amsr.pos[sample].lat = this->pos[sample].lat;
    amsr.pos[sample].lon = this->pos[sample].lon;

    for (band = 0; band < nbands; band++) {
      amsr.pos[sample].temp[channel[band]] 
	= nint(this->pos[sample].temp[band] / AMSR_TEMP_SCALE);
    }
  }

/*
 *	write record
 */
  if (fseek(fp, scan*sizeof(amsr), SEEK_SET) != 0) return FALSE;
  nbytes = fwrite(&amsr, 1, sizeof(amsr), fp);
  if (feof(fp) || ferror(fp) || sizeof(amsr) != nbytes) return FALSE;

  return TRUE;
}

/*------------------------------------------------------------------------
 * read_amsr_res5 - read AMSR .res5 data into generic swath
 *
 *------------------------------------------------------------------------*/
static bool read_amsr_res5(GSwath *this, int scan, FILE *fp, 
			   int nbands, int channel[])
{ register int band, sample;
  byte4 nbytes, hhmmss, yyyymmdd;
  byte2 year, month, dom, doy;
  static struct AMSR_RES5_SCAN amsr;

/*
 *	read next record
 */
  if (fseek(fp, scan*sizeof(amsr), SEEK_SET) != 0) { return FALSE; }
  nbytes = fread(&amsr, 1, sizeof(amsr), fp);
  if (feof(fp) || ferror(fp) || sizeof(amsr) != nbytes) { return FALSE; }

/*
 *	decode res5 record
 */
  this->status = amsr.qflag;

  this->timecode = amsr.time;

  if (0 != amsr.time) {
    timecode_to_gregorian_pmtime(amsr.time, &yyyymmdd, &hhmmss, GS_TYPE_AMSR_RES5);
    parse_value_pmtime(yyyymmdd, &year, &month, &dom);
    ymd_to_ydoy_pmtime(year, month, dom, &doy);
    this->gmt_year = year;
    this->gmt_doy = doy;
    this->gmt_seconds = hhmmss_to_hours_pmtime(hhmmss)*60*60;
  }
  else {
    this->gmt_seconds = -1;
  }

  for (sample = 0; sample < this->info->num_samples; sample++) {

    this->pos[sample].lat = amsr.pos[sample].lat;
    this->pos[sample].lon = amsr.pos[sample].lon;

    for (band = 0; band < nbands; band++) {
      this->pos[sample].temp[band] 
	= amsr.pos[sample].temp[channel[band]] * AMSR_TEMP_SCALE;
    }
  }

  this->orbit_dir = orbit_direction(this);

  return TRUE;
}

/*------------------------------------------------------------------------
 * write_amsr_res5 - write generic swath to AMSR .res5 file
 *
 *------------------------------------------------------------------------*/
static bool write_amsr_res5(GSwath *this, int scan, FILE *fp, 
			    int nbands, int channel[])
{ register int band, sample;
  long int nbytes;
  static struct AMSR_RES5_SCAN amsr;

/*
 *	decode record
 */
  amsr.qflag = this->status;

  amsr.time = this->timecode;

  for (sample = 0; sample < this->info->num_samples; sample++) {

    amsr.pos[sample].lat = this->pos[sample].lat;
    amsr.pos[sample].lon = this->pos[sample].lon;

    for (band = 0; band < nbands; band++) {
      amsr.pos[sample].temp[channel[band]] 
	= nint(this->pos[sample].temp[band] / AMSR_TEMP_SCALE);
    }
  }

/*
 *	write record
 */
  if (fseek(fp, scan*sizeof(amsr), SEEK_SET) != 0) return FALSE;
  nbytes = fwrite(&amsr, 1, sizeof(amsr), fp);
  if (feof(fp) || ferror(fp) || sizeof(amsr) != nbytes) return FALSE;

  return TRUE;
}
