/*--------------------------------------------------------------------------
 * gswath - generic swath structure
 *
 * 30-Nov-1995 K.Knowles knowlesk@kryos.colorado.edu 303-492-0644
 * National Snow & Ice Data Center, University of Colorado, Boulder
 * Copyright (C) 1995-2004 University of Colorado
 *--------------------------------------------------------------------------*/
#ifndef gswath_h_
#define gswath_h_

#ifdef gswath_c_
const char gswath_h_rcsid[]="$Header: /NSIDC_CVS/FILES/generic_swath/src/gswath/gswath.h,v 1.13 2004/01/23 01:57:57 knowlesk Exp $";
#endif

#define GS_SECS_PER_DAY (24 * 60 * 60)
#define GS_MAX_CHANNELS (100)

typedef enum 
{ GS_TYPE_UNKNOWN = -1,
  GS_TYPE_SMMR_PF = 0, 	/* Pathfinder SMMR */
  GS_TYPE_SSMI_FULL, 	/* Decode SSM/I lower channels */
  GS_TYPE_SSMI_LONG, 	/* Decode SSM/I upper channels */
  GS_TYPE_AMSR_RES1,	/* AMSR-E LowRes (10km) */
  GS_TYPE_AMSR_RES5,	/* AMSR-E HiRes (5km) */
  NUM_GS_TYPE 
} GS_type;

static const char *GS_file_extension[] =
{ ".smmr", 
  ".full", 
  ".long",
  ".res1",
  ".res5"
};

static const char *GS_type_string[] = 
{ "SMMR pathfinder data",
  "SSM/I lower channels",
  "SSM/I upper channels",
  "AMSR-E low res (10km)",
  "AMSR-E high res (5km)"
};

static const int GS_epoch_year[] =
{ 1978,
  1987,
  1987,
  1993,
  1993
};

/* Time of day to begin default daily files */
static const byte4 GS_daily_begin_hhmmss[] =
  { 120000, /* noon */
    000000, /* midnight */
    000000, /* midnight */
    000000, /* midnight */
    000000 /* midnight */
  };

/* Length of default daily files */
static const int GS_daily_subset_length_sec[] =
  { 2 * GS_SECS_PER_DAY, /* 48 hours */
    GS_SECS_PER_DAY, /* 24 hours */
    GS_SECS_PER_DAY, /* 24 hours */
    GS_SECS_PER_DAY, /* 24 hours */
    GS_SECS_PER_DAY /* 24 hours */
  };

typedef enum 
{ UNKNOWN_DIR, 
  ASC, 
  DES, 
  NUM_DIR 
} GS_orbit_dir;

static const char *GS_orbit_dir_string[] = 
{ "Unknown",
  "Ascending",
  "Descending"
};

/*
 *	structure for invariant information about the data
 */
typedef struct
{ int num_samples;		/* number of samples in one scan */
  int num_channels;		/* number of channels in each sample */
  int center_pos;		/* sample position of nadir */
  float one_scan_sec;		/* duration of a single scan (seconds) */
  float one_scan_deg;		/* distance travelled in one scan (degrees) */
  float orbit_period_sec;	/* duration of a full orbit in seconds */
  float orbit_declination;	/* orbit declination (down from 90) */
  float nominal_res_km;		/* nominal resolution in kilometers */
  float *Tb_min;		/* minimum data value by channel */
  float *Tb_max;		/* maximum data value by channel */
  const char **channel_name;	/* channel name strings */
  bool (*read_swath)();
  bool (*write_swath)();
} GS_info;

/*
 *	structure for navigation information
 */
typedef struct 
{ GS_orbit_dir dir;
  float lat;
  byte4 timecode;
} GS_nav;

/*
 *	structure for a single sample position, all channels
 */
typedef struct
{ float lat;
  float lon;
  float *temp;
} GSample;

/*
 *	structure for a single scan
 */
typedef struct
{ GS_type type;
  GS_info *info;
  int gmt_year;
  int gmt_doy;
  float gmt_seconds;
  byte4 timecode;
  byte2 status;
  GS_orbit_dir orbit_dir;
  GSample *pos;
  GS_nav *nav;
} GSwath;

/*
 * once a new structure is allocated the
 * type, num_samples and num_channels fields
 * should be considered invariant (read-only)
 *
 * the timecode and status fields have
 * different definitions depending on the
 * sensor type, however the timecode is always
 * some linearly increasing value that can
 * be used in relational comparisons, 
 * further, valid timecodes are greater than zero
 * zero timecode indicates an invalid value
 * gmt_seconds == -1 also indicates bad time value
 *
 * here's an example of a reference to a
 * single sample position within a swath
 *
 *	GSwath *gs = new_gswath(type);
 *	latitude of sample number N = gs->pos[N].lat
 * 	Tb for sample N, channel 0 = gs->pos[N].temp[0]
 *
 * the nav pointer is for private navigational data
 *
 */

GSwath *new_gswath(GS_type type);

void close_gswath(GSwath *this);

GS_type get_type_from_name_gswath(char *orbit_filename, bool verbose);

bool read_gswath(GSwath *this, int scan, FILE *fp, int nbands, int channel[]);

bool write_gswath(GSwath *this, int scan, FILE *fp, int nbands, int channel[]);

bool copy_gswath(GSwath *to, GSwath *from);

#endif
