/*========================================================================
 * pmtime - time functions specific to passive microwave instruments
 *          (SMMR, SSM/I, AMSR-E)
 *
 * All functions are Y2K compliant.
 *
 * 05/04/1998 brodzik brodzik@kryos.colorado.edu 303-492-8263
 * National Snow & Ice Data Center, University of Colorado, Boulder
 * Copyright (C) 1998-2004 University of Colorado
 *========================================================================*/
#include "define.h"
#define pmtime_c_
#include "pmtime.h"
#include "days.h"

static const char pmtime_c_rcsid[]="$Header: /NSIDC_CVS/FILES/generic_swath/src/gswath/pmtime.c,v 1.8 2004/01/23 01:57:57 knowlesk Exp $";

const char *id_pmtime(void) {
  return pmtime_c_rcsid;
}

/*------------------------------------------------------------------------
 * parse_value_pmtime - parse an integer into 3 integer components, first is 
 *               unspecified number of digits, second and third are each 
 *               assumed to be 2-digit fields.  Suitable for parsing 
 *               dates and times of the form yyyymmdd or hhmmss.
 *
 *	input : value
 *    
 *      output : first, second, third
 *
 *------------------------------------------------------------------------*/
void parse_value_pmtime (byte4 value, byte2 *first, byte2 *second, 
			 byte2 *third)
{
  *third = (byte2) (value % 100);
  value /= 100;
  *second = (byte2) (value % 100);
  *first = (byte2) (value/100);

  return;
}

/*------------------------------------------------------------------------
 * join_values_pmtime - combine the 3 integer components into a single
 *               integer.  Use to combine components yyyy, mm, dd into
 *               an ingeger yyyymmdd.
 *
 *	input : first, second, third
 *    
 *      output : combined value
 *
 *------------------------------------------------------------------------*/
void join_values_pmtime (byte2 first, byte2 second, byte2 third,
			 byte4 *value)
{
  *value = ((byte4)first * 10000) + ((byte4)second * 100) + (byte4)third;

  return;
}

/*----------------------------------------------------------------------
 * ydoy_to_ymd_pmtime - convert year, day of year to month and day
 *
 *	input : yyyy - 4-digit year
 *		doy - 3-digit day of year
 *
 *	output: month - int month
 *              day - int day (of month)
 *
 *----------------------------------------------------------------------*/
void ydoy_to_ymd_pmtime (byte2 yyyy, byte2 doy, byte2 *month, byte2 *day)
{
  unsigned int mm=*month;
  unsigned int dd=*day;
  
  jdn_to_ymdg(ymdg_to_jdn((long)yyyy,0,0) + doy - 1,NULL,&mm,&dd,NULL,NULL);
  *month = mm+1;
  *day = dd+1;
  return;

}

/*----------------------------------------------------------------------
 * ymd_to_ydoy_pmtime - convert year, month and day to day of year
 *
 *	input : yyyy - 4-digit year
 *	        month - int month
 *              day - int day (of month)
 *
 *	output: doy - 3-digit day of year
 *
 *----------------------------------------------------------------------*/
void ymd_to_ydoy_pmtime (byte2 yyyy, byte2 month, byte2 day, byte2 *doy)
{
  assert(1900 < yyyy && 2100 > yyyy);
  assert(0 < month && 13 > month);
  assert(0 < day && 32 > day);

  *doy = (byte2) (ymdg_to_jdn((long)yyyy,month-1,day-1) 
    - ymdg_to_jdn((long)yyyy,0,0) + 1);

  return;

}

/*------------------------------------------------------------------------
 * hhmmss_to_hours_pmtime - convert hhmmss to decimal hours
 *
 *	input : integer hour,minute,second 2 digits each
 *
 *	result: t - decimal time in hours
 *
 *------------------------------------------------------------------------*/
double hhmmss_to_hours_pmtime (byte4 hhmmss)
{
  byte2 hh, mm, ss;

  parse_value_pmtime(hhmmss,&hh,&mm,&ss);

  return (((double)ss/PMTIME_SECS_PER_MINUTE + (double)mm)
	  /PMTIME_MINUTES_PER_HOUR 
	  + (double)hh);
}

/*------------------------------------------------------------------------
 * hhmmss_to_hours_pmtime - convert decimal hours to hours,minute,seconds
 *
 *	input : t - decimal time in hours (t < 24.00)
 *
 *	result: integer hour,minute,second 2 digits each
 *
 *      note: throws an assertions if t_hours < 0.0
 *
 *------------------------------------------------------------------------*/
byte4 hours_to_hhmmss_pmtime (double t_hours)
{
  byte2 hh, mm, ss;
  double t=t_hours;
  byte4 hhmmss;

  assert(0 <= t_hours);

  hh = (byte2) t;
  t = (t - hh) * PMTIME_MINUTES_PER_HOUR;
  mm = (byte2) t;
  t = (t - mm) * PMTIME_SECS_PER_MINUTE;
  ss = (byte2) (t+0.5);

  join_values_pmtime(hh,mm,ss,&hhmmss);
  return (hhmmss);

}

/*----------------------------------------------------------------------
 *  gregorian_to_timecode_pmtime - convert gregorian date to gswath 
 *                                 timecode value
 * 
 *    input : yyyymmdd - gregorian year, month, day
 *            hhmmss - hour, minute, second
 *            swath_type - swath type (used to determine timecode epoch) 
 * 
 *    result: timecode corresponding to input gregorian date
 *----------------------------------------------------------------------*/
byte4 gregorian_to_timecode_pmtime (byte4 yyyymmdd, byte4 hhmmss, 
				    GS_type swath_type) {

  long days_since_epoch;
  byte2 year, month, day, hour, minute, second;

  parse_value_pmtime(yyyymmdd,&year,&month,&day);
  parse_value_pmtime(hhmmss,&hour,&minute,&second);

  /* 
   * Use jdn (Julian Day Number, a Julian Date rounded to the nearest 
   * integer) to determine days since epoch.
   */
  days_since_epoch = ymdg_to_jdn(year,month-1,day-1) 
    - ymdg_to_jdn(GS_epoch_year[swath_type],0,0);

  return (byte4)((days_since_epoch * PMTIME_SECS_PER_DAY) + 
		 (hour * PMTIME_SECS_PER_HOUR) +
		 (minute * PMTIME_SECS_PER_MINUTE) + 
		 second);

}

/*----------------------------------------------------------------------
 *  timecode_to_gregorian_pmtime - convert gswath timecode value to 
 *                                   gregorian
 * 
 *    input : timecode - seconds since epoch
 *            swath_type - swath type (used to determine timecode epoch) 
 *
 *    output: yyyymmdd - gregorian year, month, day
 *            hhmmss - hour, minute, second
 * 
 *----------------------------------------------------------------------*/
void timecode_to_gregorian_pmtime (byte4 timecode, byte4 *yyyymmdd, 
				   byte4 *hhmmss, GS_type swath_type) {

  byte4 secs_of_day;
  long days_since_epoch;
  long year;
  unsigned int month, day, leap, doy;

  /* Split timecode into two parts, the nearest day, and the remainder */
  secs_of_day = timecode % PMTIME_SECS_PER_DAY;
  days_since_epoch = ((long)timecode)/PMTIME_SECS_PER_DAY;
  
  /* Convert the nearest day to jdn and then to gregorian */
  jdn_to_ymdg(days_since_epoch + ymdg_to_jdn(GS_epoch_year[swath_type],0,0),
	      &year,&month,&day,&leap,&doy);
  month = month + 1;
  day = day + 1;
  join_values_pmtime((byte2)year,(byte2)month,(byte2)day,yyyymmdd);

  /* Convert seconds of the day into hhmmss */
  *hhmmss = hours_to_hhmmss_pmtime(((double)(secs_of_day))
				   /((double)(PMTIME_SECS_PER_HOUR)));

  return;

}

/*----------------------------------------------------------------------
 *  day_offset_pmtime - add or subtract the specified number of days
 *                    to the input gregorian date
 * 
 *     input : yyyymmdd - gregorian date
 *             numdays - signed days to offset date by
 *
 *     output : 
 *             yyyymmdd - set to yyyymmdd + numdays
 * 
 *
 *----------------------------------------------------------------------*/
void day_offset_pmtime(byte4 *yyyymmdd, int4 numdays) {
  
  byte2 year, month, day;
  long yy;
  unsigned int mm,dd;

  parse_value_pmtime(*yyyymmdd,&year,&month,&day);
  jdn_to_ymdg(ymdg_to_jdn((long)year,month-1,day-1) + numdays,
	      &yy,&mm,&dd,NULL,NULL);
  mm = mm + 1; 
  dd = dd + 1;
  *yyyymmdd = ((yy * 10000) + (mm * 100) +  dd);

  return;

}

/*------------------------------------------------------------------------
 * standalone test driver
 *------------------------------------------------------------------------*/
#ifdef TEST_DRIVER
int main(int argc, char*argv[]) {
  char readln[MAX_STRING];
  GS_type swath_type;
  byte4 yyyymmdd, hhmmss, timecode;

  if (argc != 2 || sscanf(argv[1], "%d", &swath_type) != 1) {
    fprintf(stderr,"usage: pmtime swath_type\n");
    for (swath_type = 0; swath_type < NUM_GS_TYPE; swath_type++) {
      fprintf(stderr,"      %d = %s\n", swath_type, GS_type_string[swath_type]);
    }
    exit(ABORT);
  }

  for(;;) {
    printf("enter yyyymmdd hhmmss or timecode: ");
    fgets(readln, MAX_STRING, stdin);
    if (sscanf(readln, "%d %d", &yyyymmdd, &hhmmss) == 2) {
      timecode = gregorian_to_timecode_pmtime(yyyymmdd, hhmmss, swath_type);
      printf("timecode = %d\n", timecode);
    }
    else if (sscanf(readln, "%d", &timecode) == 1) {
      timecode_to_gregorian_pmtime(timecode, &yyyymmdd, &hhmmss, swath_type);
      printf("date time = %7d %6.6d\n", yyyymmdd, hhmmss);
    }
    else {
      if (feof(stdin)) { printf("\n"); exit(0);}
      if (*readln == '\n') break;
    }
  }

  exit(0);
}
#endif
