/*--------------------------------------------------------------------------
 * struct_ssmi.h - contains the structures used in SSM/I gridding programs
 *
 * 05/29/1998  M. J. Brodzik  brodzik@kryos.colorado.edu  303-492-8263
 * National Snow & Ice Data Center, University of Colorado, Boulder
 * Copyright (C) 1998 University of Colorado
 *--------------------------------------------------------------------------*/
#ifndef struct_ssmi_h_
#define struct_ssmi_h_

#ifdef gswath_c_
const char struct_ssmi_h_rcsid[]="$Header: /NSIDC_CVS/FILES/generic_swath/src/gswath/struct_ssmi.h,v 1.8 2004/01/23 01:57:57 knowlesk Exp $";
#endif

#include "define.h"

/* 
 * SSM/I satellite platforms 
 */
typedef enum {
  DMSP_F08, DMSP_F10, DMSP_F11, DMSP_F13, SSMI_NUM_SATS
} ssmi_sat_designator;

static const int ssmi_sat_number[] = { 8, 10, 11, 13 };
static const char *ssmi_sat_name[] = { "DMSP-F08","DMSP-F10",
				       "DMSP-F11","DMSP-F13"};

/*
 *  Local time of first ascending node, in decimal hours
 *  References: 
 *   F08: Wentz, User's Manual, SSM/I Antenna Temperature Tapes, 
 *        Revision 1, p. 3.
 *   others: Passive Microwave workshop notes from Marie Colton,
 *        AMS Meeting, Long Beach, CA, 2/2/97
 */
static const float ssmi_asc_node_hour[] = {  6.20, 20.75, 17.17, 17.58 };
static const float ssmi_des_node_hour[] = { 18.20,  8.75,  5.17,  5.58 };

/* SSM/I channel indices into "full" structure arrays */
typedef enum {
  TA19V, TA19H, TA22V, TA37V, TA37H, SSMI_FULL_CHANNELS
} ssmi_full_channel_designator;

static const char *ssmi_full_channel_name[] = {"19V","19H","22V","37V","37H"};
static float ssmi_full_Tb_min_Kelvins[] = {  55,  55,  55,  55,  55 };
static float ssmi_full_Tb_max_Kelvins[] = { 320, 320, 320, 320, 320 };

/* SSM/I channel indices into "long" structure arrays*/
typedef enum {
  TA85V, TA85H, SSMI_LONG_CHANNELS
} ssmi_long_channel_designator;

static const char *ssmi_long_channel_name[] = {"85V","85H"};
static float ssmi_long_Tb_min_Kelvins[] = {  55,  55 };
static float ssmi_long_Tb_max_Kelvins[] = { 320, 320 };

/*
 *  SSM/I 25-km orbit scans contain "full" channels data (19H+V, 22V, 37H+V GHz only)
 *  SSM/I 12.5-km  orbit scans contain "long" channels data (85H+V GHz only)
 */
#define SSMI_RECS_PER_FULL_SCAN 64
#define SSMI_RECS_PER_LONG_SCAN 128
#define SSMI_SEC_PER_SCAN 1.9
#define SSMI_SEC_PER_ORBIT 6200
#define SSMI_DECLINATION 9.0
#define SSMI_FULL_RES_km 25.0
#define SSMI_LONG_RES_km 12.5
#define SSMI_LAT_BIAS 90.00
#define SSMI_LAT_SCALE 0.01
#define SSMI_LON_BIAS  0.00
#define SSMI_LON_SCALE 0.01
#define SSMI_TEMP_SCALE 0.01

/*
 *  Full SSMI record, without time
 *  Stores latitude, longitude and an array of temperatures for
 *  the five lower channels
 */
struct SSMI_FULL_REC_NT {
  byte2 lat;
  byte2 lon;
  byte2 temp[SSMI_FULL_CHANNELS];
};

/*
 *  One SSM/I 25-km orbit scan
 *  tht is the incidence angle for this scan.
 */
struct SSMI_FULL_REC {
  byte4 time;
  byte2 tht;
  struct SSMI_FULL_REC_NT full[SSMI_RECS_PER_FULL_SCAN];
};

/*
 *  Long SSMI record, without time
 *  Stores latitude, longitude and an array of temperatures for
 *  the two upper channels
 */
struct SSMI_LONG_REC_NT {
  byte2 lat;
  byte2 lon;
  byte2 temp[SSMI_LONG_CHANNELS];
};

/*
 *  One SSM/I 12.5-km orbit scan
 */
struct SSMI_LONG_REC {
  byte4 time;
  struct SSMI_LONG_REC_NT lon8[SSMI_RECS_PER_LONG_SCAN];
};

/*
 * Constants used during quality control by EASE-Grid (Backus-Gilber)
 * interpolation routines:
 * 
 * QC on interpolation has 2 steps:
 * 1) The interpolation coefficient of any individual missing TB sample
 *    must be smaller in absolute value than SSMI_MISSING_COEF_THRESHOLD
 * 2) The sum of interpolation coefficients for all missing TBs must
 *    be within some SSMI_COEF_SUM_DELTA of 1.0, to use the interpolated 
 *    TB for the given pixel.
 * By coincidence, these values happen to agree now.  They may not in the
 * future, so are being implemented as different constants.
 */
#define SSMI_COEF_SUM_DELTA 0.0005
#define SSMI_MISSING_COEF_THRESHOLD 0.0005

/* Filler values used by EASE-Grid (Backus-Gilbert) interpolation routines */
static const byte2 ssmi_missing_Tb = 0;
static const byte1 ssmi_missing_tim = UCHAR_MAX;

/* Scaling factors for EASE-Grid output TB and time data */
#define SSMI_EASE_TB_SCALE 10.0
#define SSMI_EASE_TIME_SCALE 10.0
#define ssmi_ease_scale_Tb(x) ((byte2)((x)<=0 ? 0 : (((x)*SSMI_EASE_TB_SCALE)+0.5)))
#define ssmi_ease_scale_time(x) ((byte1)(((x)*SSMI_EASE_TIME_SCALE)+0.5))

#endif /* end of struct_ssmi.h */




