/*========================================================================
 * hdftogs - convert SSM/I Tb files from HDF to generic swath
 *
 * MSFC SSM/I Tb files in HDF are described in README.ssmi_tb
 * the swath structure is described in struct_ease.h
 * and the generic swath interface is described in gswath.c
 *
 * 28-Aug-1997 K.Knowles knowlesk@kryos.colorado.edu 303-492-0644
 * National Snow & Ice Data Center, University of Colorado, Boulder
 * Copyright (C) 1997-2004 University of Colorado
 *========================================================================*/
#include "define.h"
#include "hdf.h"
#include "mfhdf.h"
#include "struct_ssmi.h"
#include "pmtime.h"

static const char hdftogs_c_rcsid[] = "$Header: /NSIDC_CVS/FILES/generic_swath/src/hdftogs/hdftogs.c,v 1.12 2004/01/23 01:06:10 knowlesk Exp $";

#define usage								  \
" usage: hdftogs [-d -k -l -i -v] [filename...]\n"			  \
"  input : MSFC SSM/I Tb files in HDF\n"				  \
"          n.b. location files (*_hn_*.hdf) must be in same\n"		  \
"               directory as corresponding *_Tb_*.hdf files\n"		  \
"  output: SSM/I \"orbit\" files (*_Tb_*.orb.full and *_Tb_*.orb.long)\n" \
"\n"									  \
"  option: d - don't filter out bad data\n"				  \
"          k - keep scans with questionable geolocation\n"		  \
"              (useful for regression testing)\n"			  \
"          l - output lower channels only\n"				  \
"          i - get file list from stdin as opposed to command line\n"	  \
"          v - verbose\n"

#define MSFC_MAX_SCANS 1620
#define MSFC_LAT_SCALE 0.01
#define MSFC_LON_SCALE 0.01
#define MSFC_TEMP_SCALE 0.01
#define MSFC_MIN_TB_QC 100

static int verbose;

#define check_hdf_error(io_status, description)	\
if (FAIL == io_status) {			\
  fprintf(stderr,"error: %s\n", description);	\
  HEprint(stderr, 0);				\
}

/*
 * test for orbit which goes beyond midnight on last day of year
 */
#define next_year_test(doy) 			\
  (doy_from_filename >= 365 && 1 == (doy))

/*------------------------------------------------------------------------
 * get_sds_by_name - return SDS info of variable which matches given string
 *
 *      input : sd_id - SD interface identifier returned by SDstart
 *              var_name - string to match in long_name label
 *
 *      output: sds_id - SDS identifier from SDselect
 *              rank, dims, ntype, nattrs - from SDgetinfo
 *
 *      result: FAIL = -1 otherwise 0 for success
 *              if more than one match, returns first one found
 *
 *------------------------------------------------------------------------*/
static int32 get_sds_by_name(int32 sd_id, char *var_name,
			     int32 *sds_id, int32 *rank, int32 *dims, 
			     int32 *ntype, int32 *nattrs) {
  int32 ios, n_datasets, n_file_attrs, sds_index;
  char label[65], unit[65], form[65], csys[65], sds_name[65]; 

/*
 * get the total number of SDs
 */
  ios = SDfileinfo(sd_id, &n_datasets, &n_file_attrs);
  check_hdf_error(ios, "get_sds_id_by_name: fileinfo");
  if (FAIL == ios) return FAIL;

/*
 * search by index for label matching var_name
 */
  for (sds_index = 0; sds_index < n_datasets; sds_index++) {

    *sds_id = SDselect(sd_id, sds_index);
    check_hdf_error(*sds_id, "get_sds_by_name: select");

    ios = SDgetdatastrs(*sds_id, label, unit, form, csys, 64);
    if (FAIL == ios || '\0' == label[0]) continue;

    if (NULL != strstr(label, var_name)) break; /* found it */
  }

  if (sds_index >= n_datasets) {
    fprintf(stderr, "Can't find variable %s in HDF file\n", var_name);
    *sds_id = -1;
    return FAIL;
  }

  ios = SDgetinfo(*sds_id, sds_name, rank, dims, ntype, nattrs);
  check_hdf_error(ios, "get_sds_by)name: getinfo");

  return 0;
}

/*------------------------------------------------------------------------
 * ydt_to_timecode - convert year, dayofyear, timeofday to ssmi time
 *
 *	input : full integer year (eg. 1988)
 *              integer day of year
 *              decimal time of day
 *
 *	result: date/time in minutes since 1987
 *
 *------------------------------------------------------------------------*/
static byte4 ydt_to_timecode(int year, int doy, double tod) {
  byte2 month, day;
  byte4 yyyymmdd, hhmmss;

  assert(tod >= 0);

  ydoy_to_ymd_pmtime(year, doy, &month, &day);
  join_values_pmtime(year, month, day, &yyyymmdd);
  hhmmss = hours_to_hhmmss_pmtime(tod);
  return gregorian_to_timecode_pmtime(yyyymmdd, hhmmss, GS_TYPE_SSMI_FULL);
}


int main (int argc, char *argv[]) {
 
  bool do_stdin, do_qcheck, do_upper, do_abs_time;
  int ios, sample, scan;
  int year, year_from_filename, doy_from_filename;
  int num_bad_full, num_bad_long;
  char *option, *suffix, *needle, *basename;
  char Tb_filename[FILENAME_MAX];
  char hn_filename[FILENAME_MAX];
  char full_filename[FILENAME_MAX];
  char long_filename[FILENAME_MAX];
  float latitude, longitude;
  int32 Tb_file=0, hn_file=0;
  FILE *full_file=NULL, *long_file=NULL;
  int32 sds_tod=0, sds_doy=0, sds_19v=0, sds_19h=0, sds_22v=0;
  int32 sds_37v=0, sds_37h=0, sds_85v=0, sds_85h=0, sds_lat=0, sds_lon=0;
  struct SSMI_FULL_REC scan_low;
  struct SSMI_LONG_REC scan_high;
  int16 doy[MSFC_MAX_SCANS]; /* day of year */
  float32 tod[MSFC_MAX_SCANS]; /* time of day */
  int16 Tb_19v[MSFC_MAX_SCANS/2][SSMI_RECS_PER_FULL_SCAN];
  int16 Tb_19h[MSFC_MAX_SCANS/2][SSMI_RECS_PER_FULL_SCAN];
  int16 Tb_22v[MSFC_MAX_SCANS/2][SSMI_RECS_PER_FULL_SCAN];
  int16 Tb_37v[MSFC_MAX_SCANS/2][SSMI_RECS_PER_FULL_SCAN];
  int16 Tb_37h[MSFC_MAX_SCANS/2][SSMI_RECS_PER_FULL_SCAN];
  int16 Tb_85v[MSFC_MAX_SCANS][SSMI_RECS_PER_LONG_SCAN];
  int16 Tb_85h[MSFC_MAX_SCANS][SSMI_RECS_PER_LONG_SCAN];
  int16 lat[MSFC_MAX_SCANS][SSMI_RECS_PER_LONG_SCAN];
  int16 lon[MSFC_MAX_SCANS][SSMI_RECS_PER_LONG_SCAN];
  int32 rank, ntype, nattrs;
  int32 dims[2], dims_high[2], dims_low[2], start[2]={0,0};
  double tod_hours;

  /*
   * set defaults
   */
  verbose = 0;
  do_stdin = FALSE;
  do_qcheck = TRUE;
  do_upper = TRUE;
  do_abs_time = FALSE;

  /*
   * get command line options
   */
  while (--argc > 0 && (*++argv)[0] == '-')
  { for (option = argv[0]+1; *option != '\0'; option++)
    { switch (*option)
      { case 'v':
	  ++verbose;
	  break;
	case 'V':
	  fprintf(stderr, "%s\n", hdftogs_c_rcsid);
	  break;
	case 'd':
	  do_qcheck = FALSE;
	  break;
	case 'k':
	  do_abs_time = TRUE;
	  break;
	case 'l':
	  do_upper = FALSE;
	  break;
	case 'i':
	  do_stdin = TRUE;
	  break;
	default:
	  fprintf(stderr, "hdftogs: invalid option %c\n", *option);
	  error_exit(usage);
	}
    }
  }

  /*
   * check command line args
   */
  if (!do_stdin && argc < 1) error_exit(usage);



  /*
   * process files
   */
  for (;;)
  {
    /*
     * clean up
     */
    if (sds_doy > 0) { SDendaccess(sds_doy); sds_doy = 0; }
    if (sds_tod > 0) { SDendaccess(sds_tod); sds_tod = 0; }

    if (sds_19v > 0) { SDendaccess(sds_19v); sds_19v = 0; }
    if (sds_19h > 0) { SDendaccess(sds_19h); sds_19h = 0; }
    if (sds_22v > 0) { SDendaccess(sds_22v); sds_22v = 0; }
    if (sds_37v > 0) { SDendaccess(sds_37v); sds_37v = 0; }
    if (sds_37h > 0) { SDendaccess(sds_37h); sds_37h = 0; }

    if (sds_85v > 0) { SDendaccess(sds_85v); sds_85v = 0; }
    if (sds_85h > 0) { SDendaccess(sds_85h); sds_85h = 0; }

    if (sds_lat > 0) { SDendaccess(sds_lat); sds_lat = 0; }
    if (sds_lon > 0) { SDendaccess(sds_lon); sds_lon = 0; }

    if (long_file) { fclose(long_file); long_file = NULL; }
    if (full_file) { fclose(full_file); full_file = NULL; }
    if (hn_file) { SDend(hn_file); hn_file = 0; }
    if (Tb_file) { SDend(Tb_file); Tb_file = 0; }

    /*
     * open brightness temperature (Tb) file
     */
    if (do_stdin)
    { fgets(Tb_filename, FILENAME_MAX, stdin);
      if (feof(stdin)) break;
    }
    else
    { if (argc < 1) break;
      strcpy(Tb_filename, *argv);
      --argc; ++argv;
    }

    Tb_file = SDstart(Tb_filename, DFACC_RDONLY);
    check_hdf_error(Tb_file, "looking for Tb data file");
    if (FAIL == Tb_file) { perror(Tb_filename); continue; }

    /*
     * high resolution geolocation (hn) file
     */
    strcpy(hn_filename, Tb_filename);

    basename = strrchr(hn_filename, '/');
    if (!basename) basename = hn_filename;
    needle = strstr(basename, "_Tb_");
    if (!needle)
    { fprintf(stderr, "%s is a Tb file?\n", Tb_filename); continue; }
    strncpy(needle, "_hn_", 4);

    sscanf(needle+4, "%2d%3d", &year_from_filename, &doy_from_filename);
    if (year_from_filename >= 87) year_from_filename += 1900;
    else year_from_filename += 2000;

    hn_file = SDstart(hn_filename, DFACC_RDONLY);
    check_hdf_error(hn_file, "looking for location data file");
    if (FAIL == hn_file) { perror(hn_filename); continue; }


    /*
     * "full" orbit file
     */
    strcpy(full_filename, Tb_filename);
    basename = strrchr(full_filename, '/');
    if (!basename) basename = full_filename;

    suffix = strstr(basename, ".hdf");
    if (!suffix)
    { fprintf(stderr, "%s is a HDF file?\n", Tb_filename); continue; }
    strcpy(suffix, ".orb.full");

    full_file = fopen(full_filename, "w");
    if (!full_file) { perror(full_filename); continue; }

    /*
     * "long" orbit file
     */
    if (do_upper)
    {
      strcpy(long_filename, full_filename);
      suffix = strstr(long_filename, ".full");
      strncpy(suffix, ".long", 5);

      long_file = fopen(long_filename, "w");
      if (!long_file) { perror(long_filename); continue; }
    }

    if (verbose) 
    { fprintf(stderr, "> reading Tb  : %s\n", Tb_filename);
      fprintf(stderr, ">         hn  : %s\n", hn_filename);
      fprintf(stderr, "> writing full: %s\n", full_filename);
      fprintf(stderr, ">         long: %s\n", 
	      do_upper ? long_filename : "N/A");
    }

/*
 * read day of year (aka "Julian" day) and time of day
 */
    ios = get_sds_by_name(Tb_file, "Julian Day", &sds_doy, 
			  &rank, dims, &ntype, &nattrs);
    ios = SDreaddata(sds_doy, start, NULL, dims, (VOIDP)doy);
    check_hdf_error(ios, "reading day of year data");

    ios = get_sds_by_name(Tb_file, "Time of Day", &sds_tod, 
			  &rank, dims, &ntype, &nattrs);
    ios = SDreaddata(sds_tod, start, NULL, dims, (VOIDP)tod);
    check_hdf_error(ios, "reading time of day data");

    if (verbose >= 2) 
      fprintf(stderr,">> date/time: %d x %d\n", dims[0], dims[1]);

/*
 * read lower channel Tbs
 */
    ios = get_sds_by_name(Tb_file, "V19", &sds_19v, 
			  &rank, dims_low, &ntype, &nattrs);
    ios = SDreaddata(sds_19v, start, NULL, dims_low, (VOIDP)Tb_19v);
    check_hdf_error(ios, "reading Tb 19v data");

    ios = get_sds_by_name(Tb_file, "H19", &sds_19h, 
			  &rank, dims_low, &ntype, &nattrs);
    ios = SDreaddata(sds_19h, start, NULL, dims_low, (VOIDP)Tb_19h);
    check_hdf_error(ios, "reading Tb 19h data");

    ios = get_sds_by_name(Tb_file, "V22", &sds_22v, 
			  &rank, dims_low, &ntype, &nattrs);
    ios = SDreaddata(sds_22v, start, NULL, dims_low, (VOIDP)Tb_22v);
    check_hdf_error(ios, "reading Tb 22v data");

    ios = get_sds_by_name(Tb_file, "V37", &sds_37v, 
			  &rank, dims_low, &ntype, &nattrs);
    ios = SDreaddata(sds_37v, start, NULL, dims_low, (VOIDP)Tb_37v);
    check_hdf_error(ios, "reading Tb 37v data");

    ios = get_sds_by_name(Tb_file, "H37", &sds_37h, 
			  &rank, dims_low, &ntype, &nattrs);
    ios = SDreaddata(sds_37h, start, NULL, dims_low, (VOIDP)Tb_37h);
    check_hdf_error(ios, "reading Tb 37h data");

    if (verbose >= 2) 
      fprintf(stderr,">>  low freq: %d x %d\n", dims_low[0], dims_low[1]);

/*
 * read upper channel Tbs
 */
    if (do_upper) { 
      
      ios = get_sds_by_name(Tb_file, "V85", &sds_85v, 
			  &rank, dims_high, &ntype, &nattrs);
      ios = SDreaddata(sds_85v, start, NULL, dims_high, (VOIDP)Tb_85v);
      check_hdf_error(ios, "reading Tb 85v data");

      ios = get_sds_by_name(Tb_file, "H85", &sds_85h, 
			  &rank, dims_high, &ntype, &nattrs);
      ios = SDreaddata(sds_85h, start, NULL, dims_high, (VOIDP)Tb_85h);
      check_hdf_error(ios, "reading Tb 85h data");

      if (verbose >= 2) 
	fprintf(stderr,">> high freq: %d x %d\n", dims_high[0], dims_high[1]);
    }

/*
 * read location data
 */
    ios = get_sds_by_name(hn_file, "Latitude", &sds_lat, 
			  &rank, dims, &ntype, &nattrs);
    ios = SDreaddata(sds_lat, start, NULL, dims, (VOIDP)lat);
    check_hdf_error(ios, "reading lat data");

    ios = get_sds_by_name(hn_file, "Longitude", &sds_lon, 
			  &rank, dims, &ntype, &nattrs);
    ios = SDreaddata(sds_lon, start, NULL, dims, (VOIDP)lon);
    check_hdf_error(ios, "reading lon data");

    if (verbose >= 2) 
      fprintf(stderr,">>   lat/lon: %d x %d\n", dims[0], dims[1]);

    /*
     * write "full" records
     */
    num_bad_full = 0;
    for (scan = 0; scan < dims_low[0]; scan++)
    { 
      year = year_from_filename;
      if (next_year_test(doy[2*scan]))
      { year += 1;
#ifdef DEBUG
        if (verbose >= 2) 
	  fprintf(stderr,">> rollover from %d doy=%d to %d doy=%d\n",
		  year_from_filename, doy_from_filename, year, doy[2*scan]);
#endif
      }
      assert(doy[2*scan] == doy_from_filename ||
	     doy[2*scan] == doy_from_filename + 1 || 
	     year == year_from_filename + 1);

      tod_hours = (double)tod[2*scan]/60/60;

      if (do_abs_time) { tod_hours = fabs(tod_hours); }

      if (tod_hours < 0) { ++num_bad_full; continue; }

      scan_low.time = 
	ydt_to_timecode(year, doy[2*scan], tod_hours);

      for (sample = 0; sample < dims_low[1]; sample++)
      { 
	if (do_qcheck)
	{ if (Tb_19v[scan][sample] < MSFC_MIN_TB_QC ||
	      Tb_19h[scan][sample] < MSFC_MIN_TB_QC ||
	      Tb_22v[scan][sample] < MSFC_MIN_TB_QC ||
	      Tb_37v[scan][sample] < MSFC_MIN_TB_QC ||
	      Tb_37h[scan][sample] < MSFC_MIN_TB_QC)
	  {
	    scan_low.time = 0;
	    break;
	  }
	}

	latitude = lat[2*scan][2*sample]*MSFC_LAT_SCALE;
	longitude = lon[2*scan][2*sample]*MSFC_LON_SCALE;
	if (longitude < 0) longitude += 360;

	scan_low.full[sample].lat = 
	  nint((latitude + SSMI_LAT_BIAS)/SSMI_LAT_SCALE);

        scan_low.full[sample].lon = 
	  nint((longitude + SSMI_LON_BIAS)/SSMI_LON_SCALE);

	scan_low.full[sample].temp[TA19V] = 
	  nint(Tb_19v[scan][sample]*MSFC_TEMP_SCALE/SSMI_TEMP_SCALE);

	scan_low.full[sample].temp[TA19H] = 
	  nint(Tb_19h[scan][sample]*MSFC_TEMP_SCALE/SSMI_TEMP_SCALE);

	scan_low.full[sample].temp[TA22V] = 
	  nint(Tb_22v[scan][sample]*MSFC_TEMP_SCALE/SSMI_TEMP_SCALE);

	scan_low.full[sample].temp[TA37V] = 
	  nint(Tb_37v[scan][sample]*MSFC_TEMP_SCALE/SSMI_TEMP_SCALE);

	scan_low.full[sample].temp[TA37H] = 
	  nint(Tb_37h[scan][sample]*MSFC_TEMP_SCALE/SSMI_TEMP_SCALE);

      }

      if (scan_low.time <= 0) { ++num_bad_full; continue; }

      fwrite(&scan_low, 1, sizeof(scan_low), full_file);

    }

    if (verbose) fprintf(stderr,"> %d full scans %d bad\n", 
			 dims_low[0], num_bad_full);

    /*
     * write "long" records
     */
    if (do_upper)
    { num_bad_long = 0;
      for (scan = 0; scan < dims_high[0]; scan++)
      { 
	year = year_from_filename;
	if (next_year_test(doy[scan])) 
	{ year += 1;
#ifdef DEBUG
	  if (verbose >= 2) 
	    fprintf(stderr,">> rollover from %d doy=%d to %d doy=%d\n",
		    year_from_filename, doy_from_filename, year, doy[scan]);
#endif
	}
	assert(doy[scan] == doy_from_filename ||
	       doy[scan] == doy_from_filename + 1 || 
	       year == year_from_filename + 1);

	tod_hours = (double)tod[scan]/60/60;

	if (do_abs_time) { tod_hours = fabs(tod_hours); }

	if (tod_hours < 0) { ++num_bad_long; continue; }

	scan_high.time = 
	  ydt_to_timecode(year, doy[scan], tod_hours);

	for (sample = 0; sample < dims_high[1]; sample++)
	{ 
	  if (do_qcheck)
	  { if (Tb_85v[scan][sample] < MSFC_MIN_TB_QC ||
		Tb_85h[scan][sample] < MSFC_MIN_TB_QC)
	    {
	      scan_high.time = 0;
	      break;
	    }
	  }

	  latitude = lat[scan][sample]*MSFC_LAT_SCALE;
	  longitude = lon[scan][sample]*MSFC_LON_SCALE;
	  if (longitude < 0) longitude += 360;

	  scan_high.lon8[sample].lat = 
	    nint((latitude + SSMI_LAT_BIAS)/SSMI_LAT_SCALE);

	  scan_high.lon8[sample].lon = 
	    nint((longitude + SSMI_LON_BIAS)/SSMI_LON_SCALE);

	  scan_high.lon8[sample].temp[TA85V] = 
	    nint(Tb_85v[scan][sample]*MSFC_TEMP_SCALE/SSMI_TEMP_SCALE);

	  scan_high.lon8[sample].temp[TA85H] = 
	    nint(Tb_85h[scan][sample]*MSFC_TEMP_SCALE/SSMI_TEMP_SCALE);

	}

	if (scan_high.time <= 0) { ++num_bad_long; continue; }

	fwrite(&scan_high, 1, sizeof(scan_high), long_file);

      }
      if (verbose) fprintf(stderr,"> %d long scans %d bad\n", 
			   dims_high[0], num_bad_long);

    }

  }
}
