/***********************************************************************\
| Copyright  1997, by Philip David Howard - all rights reserved	|
|									|
| author	Philip D. Howard					|
| email		phil@txics.com, phil@milepost.com			|
| postal	Suite 129						|
|		17023 Preston Road					|
|		Dallas, Texas  75230					|
\***********************************************************************/

/***********************************************************************\
| program	compose							|
|									|
| purpose	To make a file based on a template file which can	|
|		select certain parts of certain files.			|
| usage		compose < template > output				|
\***********************************************************************/

/***********************************************************************\
| template file syntax							|
|									|
| The following template statement indicates that a file is to be	|
| inserted in the output at this point.					|
|									|
| @insert filename [ name ... ]						|
|									|
| The optional section names cause only portions of the file to be	|
| inserted.  Those sections to be inserted are found between lines	|
| that begin and end named sections.  The line that begins a section	|
| looks like:								|
|									|
| #pragma section begin [ name ... ]					|
|									|
| The line that ends a section looks like:				|
|									|
| #pragma section end [ name ... ]					|
|									|
| All other template input is copied to the output unchanged.		|
| If the insert statement contains no names at all, then the whole	|
| file is copied including all #pragma section begin or end lines.	|
| To suppress copying them, specify an arbitrary name on the insert.	|
|									|
\***********************************************************************/

#define LINESIZE 1024

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static int status = 0;

/***********************************************************************\
| function	pdh_nextword						|
|									|
| purpose	get the next word in a string				|
|									|
| arguments	1 (char * *) pointer to pointer to current character	|
|			position in the string.				|
|									|
| returns	(char *) pointer to next word, null terminated		|
|									|
| multi-thread  OK							|
|									|
| note		This code presumes an ASCII character set.		|
\***********************************************************************/
#define nextword pdh_nextword
char *
pdh_nextword (
	char * * base_p
)

{
    char * ch_p;
    char * word_p;
    char ch;

    /* check for NULL base pointer */
    if ( base_p == NULL )		return NULL;

    /* check for NULL string pointer in base */
    if ( (ch_p = * base_p) == NULL )	return NULL;

    /* check for empty string */
    if ( (ch = * ch_p) == '\0' )	return NULL;

    /* find beginning of a word */
    while ( ch <= ' ' ) {
	if ( ch == '\0' )		return NULL;
	ch = * ++ ch_p;
    }
    word_p = ch_p;

    /* find the end of the word */
    while ( ch > ' ' ) {
	ch = * ++ ch_p;
    }

    /* terminate the word and skip ahead if there is more */
    if ( ch != '\0' ) {
	* ch_p = '\0';
	++ ch_p;
    }

    /* save the base for next call */
    * base_p = ch_p;

    /* return the found word */
    return word_p;
}


/***********************************************************************\
| function	pdh_strdup						|
|									|
| purpose	Return a duplicate string filled in with the contents	|
|		of the argument string up to and including the null	|
|		termination character.					|
|									|
| arguments	1 (const void *) pointer to source string		|
|									|
| returns	(char *) NULL: duplicate could not be created		|
|		(char *) pointer to duplicate binary string		|
|									|
| multi-thread  OK							|
\***********************************************************************/
#define strdup pdh_strdup
char *
pdh_strdup (
	const void * str
)
{
    void * dup;

    dup = malloc( strlen( str ) );
    if ( ! dup ) return NULL;
    strcpy( dup , str );

    return dup;
}


/***********************************************************************\
| function	match_stmt						|
|									|
| purpose	Check to see if a line in a buffer is a statement	|
|		that has the specified prefix and symbol.		|
\***********************************************************************/
char *
match_stmt (
	char *		line,
	int		prefix,
	const char *	name
)
{
    char *	ptr		;
    char *	word		;

    /* find the first non-blank character */
    ptr = line;
    for (;;) {
	int c;
	if ( ( c = * ( ptr ++ ) ) == '\0' ) {
	    return NULL;
	}
	if ( c == prefix ) break;
	if ( ! isspace( c ) ) {
	    return NULL;
	}
    }

    /* check if this is the matching statement */
    if ( ! ( word = nextword( & ptr ) ) ) {
	return NULL;
    }
    if ( strcmp( word , name ) == 0 ) {
	return ptr;
    }
    return NULL;
}


/***********************************************************************\
| function	insert_stmt						|
|									|
| purpose	Check to see if a line from the template file is an	|
|		insert statement or not.  If it is, then do the insert.	|
\***********************************************************************/
int
insert_stmt (
	char *		t_line
)
{
    struct node {
	struct node *		next	;
	char *		name	;
	int			usage	;
    };

    struct node		sec_first	;
    struct node *	sec_curr	;

    FILE *		insin		;
    char *	fname		;
    char *	ptr		;
    char *	sname		;
    char *	word		;
    int			copying		;

    char	i_line		[ LINESIZE ];
    char	x_line		[ LINESIZE ];


    /**************************************\
    | check for an insert statement or not |
    ***************************************/
    if ( ! ( ptr = match_stmt( t_line , '@' , "insert" ) ) ) return 0;

    /**************************************\
    | check for a file name to be inserted |
    \**************************************/
    if ( ! ( fname = nextword( & ptr ) ) ) return 0;
    if ( ! ( insin = fopen( fname , "r" ) ) ) {
	fprintf(stderr,"Error %d opening file \"%s\"\n",errno,fname);
	return status = 1;
    }

    /*****************************************************************\
    | if no sections are specified then copy the whole file unchanged |
    \*****************************************************************/
    if ( ! ( sname = nextword( & ptr ) ) ) {
	for (;;) {
	    if ( ! fgets( i_line , LINESIZE , insin ) ) break;
	    fputs( i_line , stdout );
	}

	fclose( insin );
	return 1;
    }

    /**************************************************\
    | gather up the section names from the insert line |
    \**************************************************/
    sec_curr = & sec_first;
    while ( sname ) {
	if ( ! ( sec_curr -> next = malloc( sizeof ( struct node ) ) ) ) {
	    fprintf(stderr,"Error %d allocating section node\n",errno);
	    return status = 1;
	}
	sec_curr = ( sec_curr -> next );
	sec_curr -> next = NULL;
	sec_curr -> usage = 0;
	if ( ! ( sec_curr -> name = strdup( sname ) ) ) {
	    fprintf(stderr,"Error %d allocating section name\n",errno);
	    return status = 1;
	}	    
	sname = nextword( & ptr );
    }

    /****************************************************************\
    | copy portions of the insert file based on the defined sections |
    \****************************************************************/
    copying = 0;
    for (;;) {
	int be;
	int copyone;

	if ( ! fgets( i_line , LINESIZE , insin ) ) break;
	strcpy( x_line , i_line );
	copyone = copying > 0;

	/* check for start or end of a section */
	if ( ( ptr = match_stmt( i_line , '#' , "pragma" ) )
	  && ( word = nextword( & ptr ) )
	  && ( strcmp( word , "section" ) == 0 )
	  && ( word = nextword( & ptr ) )
	  && ( ( strcmp(word,"begin")==0 && ( be = 1 ) )
	    || ( strcmp(word,"end"  )==0 && ( be = 2 ) ) ) ) {
	    copyone = 0; /* do not copy this pragma section line */

	    /* go through all the section names */
	    while ( ( word = nextword( & ptr ) ) ) {

		/* see if this one matches any we are looking for */
		sec_curr = sec_first . next;
		while ( sec_curr ) {
		    if ( strcmp( word , sec_curr -> name ) == 0 ) break;
		    sec_curr = sec_curr -> next;
		}

		/* if this is a matching section name */
		if ( sec_curr ) {
		    /* increment for begin */
		    if ( be == 1 && ++ ( sec_curr -> usage ) == 1 ) ++ copying;
		    /* decrement for end */
		    if ( be == 2 && -- ( sec_curr -> usage ) == 0 ) -- copying;
		}

	    }
	}

	if ( copyone ) fputs( x_line , stdout );
    }

    /********************\
    | clean up and leave |
    \********************/
    fclose( insin );
    sec_curr = sec_first . next;
    while ( sec_curr ) {
	struct node * sec_next;
	sec_next = sec_curr -> next;
	free( sec_curr );
	sec_curr = sec_next;
    }

    return 1;
}

/***********************************************************************\
| function	main							|
|									|
| purpose	The usual main thing.					|
\***********************************************************************/
int
main (
	int			argc,
	char * *	argv
)
{
    char	t_line		[ LINESIZE ];

    /*************************************************\
    | main loop to read and process the template file |
    \*************************************************/
    for (;;) {
	if ( ! fgets( t_line , LINESIZE , stdin ) ) break;
	if ( ! insert_stmt( t_line ) ) fputs( t_line , stdout );
    }

    return 0;
}
