/***********************************************************************\
| Copyright  1996, 1997 - Philip David Howard - all right reserved	|
|									|
| author	Philip D. Howard					|
| email		phil@txics.com, phil@milepost.com			|
| postal	Suite 129						|
|		17023 Preston Road					|
|		Dallas, Texas  75230					|
\***********************************************************************/

/***********************************************************************\
| header	days.h							|
|									|
| purpose	Header for the DAYS library.				|
\***********************************************************************/

#ifndef DAYS_H
#define DAYS_H

/***********************************************************************\
| standard headers							|
\***********************************************************************/
#include <stdlib.h>
#include <time.h>

/***********************************************************************\
| macros								|
\***********************************************************************/
/* none */

/***********************************************************************\
| typedefs, structures, and unions					|
\***********************************************************************/
/* none */

/***********************************************************************\
| functions								|
\***********************************************************************/

/***********************************************************************\
| function	jdn_to_time						|
|									|
| purpose	Convert a Julian Day Number to ANSI Standard C time_t	|
|		value for midnight at the beginning of the day.		|
|									|
| argument	1 (long) the Julian Day Number				|
|									|
| returns	(time_t) the time value at the beginning of the day.	|
|									|
| note		A Julian Day Number (JDN) is a Julian Date rounded	|
|		to an integer.  The Julian Date is commonly used in	|
|		astronomy to identify dates in ephemeral data and	|
|		formulas.  The JDN is a number that is the same		|
|		through the ordinary civil day and can be used for	|
|		calendar references.					|
|									|
| ref		Explanatory Supplement to the Astronomical Almanac,	|
|		P. Kenneth Seidelmann, editor  [ISBN 0-935702-68-7]	|
|		http://www.amazon.com/exec/obidos/ISBN=0935702687	|
\***********************************************************************/
time_t
jdn_to_time (
	long	day
);


/***********************************************************************\
| function	jdn_to_tm						|
|									|
| purpose	Convert a Julian Day Number to ANSI Standard C		|
|		struct tm filled in (and allocated if needed).		|
|									|
| arguments	1 (long) the Julian Day Number				|
|		2 (struct tm *) struct to fill in or NULL to allocate	|
|									|
| returns	(struct tm *) given or allocated (struct tm) pointer	|
|		(struct tm *) NULL if error				|
|									|
| note		A Julian Day Number (JDN) is a Julian Date rounded	|
|		to an integer.  The Julian Date is commonly used in	|
|		astronomy to identify dates in ephemeral data and	|
|		formulas.  The JDN is a number that is the same		|
|		through the ordinary civil day and can be used for	|
|		calendar references.					|
|									|
| ref		Explanatory Supplement to the Astronomical Almanac,	|
|		P. Kenneth Seidelmann, editor  [ISBN 0-935702-68-7]	|
|		http://www.amazon.com/exec/obidos/ISBN=0935702687	|
\***********************************************************************/
struct tm *
jdn_to_tm (
	long	day
,
	struct tm *	tm_p
);


/***********************************************************************\
| function	jdn_to_ymdg						|
|									|
| purpose	Convert a Julian Day Number to a calendar Year, Month,	|
|		and Date in the Gregorian calendar.			|
|									|
| arguments	1 (long) the Julian Day Number				|
|		2 (long *) where to store the year (-1 = 2 BC)		|
|		3 (unsigned int *) where to store the month (0..11)	|
|		4 (unsigned int *) where to store the date (0..30)	|
|		5 (unsigned int *) where to store the leap year flag	|
|		6 (unsigned int *) where to store the day of year	|
|									|
| returns	(void)							|
|									|
| note		A Julian Day Number (JDN) is a Julian Date rounded	|
|		to an integer.  The Julian Date is commonly used in	|
|		astronomy to identify dates in ephemeral data and	|
|		formulas.  The JDN is a number that is the same		|
|		through the ordinary civil day and can be used for	|
|		calendar references.					|
|									|
| ref		Explanatory Supplement to the Astronomical Almanac,	|
|		P. Kenneth Seidelmann, editor  [ISBN 0-935702-68-7]	|
|		http://www.amazon.com/exec/obidos/ISBN=0935702687	|
\***********************************************************************/
void
jdn_to_ymdg (
	long		day
,
	long *		year_p
,
	unsigned int *	month_p
,
	unsigned int *	date_p
,
	unsigned int *	leap_p
,
	unsigned int *	doy_p
);


/***********************************************************************\
| function	jdn_to_ymdj						|
|									|
| purpose	Convert a Julian Day Number to a calendar Year, Month	|
|		and Date in the Julian calendary.			|
|									|
| arguments	1 (long) the Julian Day Number				|
|		2 (long *) where to store the year (-1 = 2 BC)		|
|		3 (unsigned int *) where to store the month (0..11)	|
|		4 (unsigned int *) where to store the date (0..30)	|
|		5 (unsigned int *) where to store the leap year flag	|
|		6 (unsigned int *) where to store the day of year	|
|									|
| returns	(void)							|
|									|
| note		A Julian Day Number (JDN) is a Julian Date rounded	|
|		to an integer.  The Julian Date is commonly used in	|
|		astronomy to identify dates in ephemeral data and	|
|		formulas.  The JDN is a number that is the same		|
|		through the ordinary civil day and can be used for	|
|		calendar references.					|
|									|
| ref		Explanatory Supplement to the Astronomical Almanac,	|
|		P. Kenneth Seidelmann, editor  [ISBN 0-935702-68-7]	|
|		http://www.amazon.com/exec/obidos/ISBN=0935702687	|
\***********************************************************************/
void
jdn_to_ymdj (
	long		day
,
	long *		year_p
,
	unsigned int *	month_p
,
	unsigned int *	date_p
,
	unsigned int *	leap_p
,
	unsigned int *	doy_p
);


/***********************************************************************\
| function	time_to_jdn						|
|									|
| purpose	Convert ANSI Standard C time_t value to a Julian Day	|
|		Number.							|
|									|
| argument	1 (time_t) time to be converted				|
|		2 (int) 0:local time, 1:UTC time			|
|									|
| returns	(long) Julian Day Number				|
|									|
| multi-thread	UNSAFE (due to unsafe ANSI functions that must be used)	|
|									|
| note		A Julian Day Number (JDN) is a Julian Date rounded	|
|		to an integer.  The Julian Date is commonly used in	|
|		astronomy to identify dates in ephemeral data and	|
|		formulas.  The JDN is a number that is the same		|
|		through the ordinary civil day and can be used for	|
|		calendar references.					|
|									|
| ref		Explanatory Supplement to the Astronomical Almanac,	|
|		P. Kenneth Seidelmann, editor  [ISBN 0-935702-68-7]	|
|		http://www.amazon.com/exec/obidos/ISBN=0935702687	|
\***********************************************************************/
long
time_to_jdn (
	time_t	time
,
	int	utc
);


/***********************************************************************\
| function	tm_to_jdn						|
|									|
| purpose	Convert ANSI Standard C struct tm to a Julian Day	|
|		Number.							|
|									|
| argument	1 (struct tm *) time structure to be converted		|
|									|
| returns	(long) Julian Day Number				|
|									|
| note		A Julian Day Number (JDN) is a Julian Date rounded	|
|		to an integer.  The Julian Date is commonly used in	|
|		astronomy to identify dates in ephemeral data and	|
|		formulas.  The JDN is a number that is the same		|
|		through the ordinary civil day and can be used for	|
|		calendar references.					|
|									|
| ref		Explanatory Supplement to the Astronomical Almanac,	|
|		P. Kenneth Seidelmann, editor  [ISBN 0-935702-68-7]	|
|		http://www.amazon.com/exec/obidos/ISBN=0935702687	|
\***********************************************************************/
long
tm_to_jdn (
	struct tm *	tm_p
);


/***********************************************************************\
| function	ymdg_to_jdn						|
|									|
| purpose	convert a Gregorian Year, Month, and Date to a Julian	|
|		Day Number.						|
|									|
| arguments	1 (long) year (-1 = 2 BC)				|
|		2 (unsigned int) month (0..11)				|
|		3 (unsigned int) date (0..30)				|
|									|
| returns	(long) Julian Day Number				|
|									|
| note		A Julian Day Number (JDN) is a Julian Date rounded	|
|		to an integer.  The Julian Date is commonly used in	|
|		astronomy to identify dates in ephemeral data and	|
|		formulas.  The JDN is a number that is the same		|
|		through the ordinary civil day and can be used for	|
|		calendar references.					|
|									|
| ref		Explanatory Supplement to the Astronomical Almanac,	|
|		P. Kenneth Seidelmann, editor  [ISBN 0-935702-68-7]	|
|		http://www.amazon.com/exec/obidos/ISBN=0935702687	|
\***********************************************************************/
long
ymdg_to_jdn (
	long		year
,
	unsigned int	month
,
	unsigned int	date
);


/***********************************************************************\
| function	ymdj_to_jdn						|
|									|
| purpose	convert a Julian Year, Month, and Date to a Julian Day	|
|		Number.							|
|									|
| arguments	1 (long) year (-1 = 2 BC)				|
|		2 (unsigned int) month (0..11)				|
|		3 (unsigned int) date (0..30)				|
|									|
| returns	(long) Julian Day Number				|
|									|
| note		A Julian Day Number (JDN) is a Julian Date rounded	|
|		to an integer.  The Julian Date is commonly used in	|
|		astronomy to identify dates in ephemeral data and	|
|		formulas.  The JDN is a number that is the same		|
|		through the ordinary civil day and can be used for	|
|		calendar references.					|
|									|
| ref		Explanatory Supplement to the Astronomical Almanac,	|
|		P. Kenneth Seidelmann, editor  [ISBN 0-935702-68-7]	|
|		http://www.amazon.com/exec/obidos/ISBN=0935702687	|
\***********************************************************************/
long
ymdj_to_jdn (
	long		year
,
	unsigned int	month
,
	unsigned int	date
);



#endif /* DAYS_H */
