/***********************************************************************\
| Copyright  1997, by Philip David Howard - all rights reserved	|
|									|
| author	Philip D. Howard					|
| email		phil@txics.com, phil@milepost.com			|
| postal	Suite 129						|
|		17023 Preston Road					|
|		Dallas, Texas  75230					|
\***********************************************************************/

#include <stddef.h>

#include "days.h"

#pragma section begin SYNOPSIS
/***********************************************************************\
| function	ymdg_to_jdn						|
|									|
| purpose	convert a Gregorian Year, Month, and Date to a Julian	|
|		Day Number.						|
|									|
| arguments	1 (long) year (-1 = 2 BC)				|
|		2 (unsigned int) month (0..11)				|
|		3 (unsigned int) date (0..30)				|
|									|
| returns	(long) Julian Day Number				|
|									|
| note		A Julian Day Number (JDN) is a Julian Date rounded	|
|		to an integer.  The Julian Date is commonly used in	|
|		astronomy to identify dates in ephemeral data and	|
|		formulas.  The JDN is a number that is the same		|
|		through the ordinary civil day and can be used for	|
|		calendar references.					|
|									|
| ref		Explanatory Supplement to the Astronomical Almanac,	|
|		P. Kenneth Seidelmann, editor  [ISBN 0-935702-68-7]	|
|		http://www.amazon.com/exec/obidos/ISBN=0935702687	|
\***********************************************************************/
#pragma section end SYNOPSIS

#define FUNCTION_NAME "ymdg_to_jdn"

#pragma section begin FUNCTION
long
ymdg_to_jdn (
	long		year
,
	unsigned int	month
,
	unsigned int	date
)
#pragma section end FUNCTION

#pragma section begin CODE
{
    /* day of year for first day of each month */
    static unsigned int my[ 12 ] = {
	0,31,59,90,120,151,181,212,243,273,304,334 };
    static unsigned int ml[ 12 ] = {
	0,31,60,91,121,152,182,213,244,274,305,335 };

    unsigned long	adjust		;
    unsigned long	days		;
    unsigned long	quadcent	;
    unsigned long	cent		;
    unsigned long	quadyear	;
    int			leap		;

    /* adjust for before epoch */
    adjust = 0;
    if ( year < 0 ) {
	adjust = ( 399 - year ) / 400;
	year += adjust * 400;
    }

    /* figure out what quad-century this is */
    quadcent = year / 400;

    /* and how many years within the quad-century we are at */
    year %= 400;

    /* figure out what century this is */
    cent = year / 100;

    /* and how many years within the century we are at */
    year %= 100;

    /* figure out what quad-year this is */
    quadyear = year / 4;

    /* and how many years within the quad-year we are at */
    year %= 4;

    /* figure out if this is a leap year */
    leap = ( ( cent == 0 || quadyear > 0 ) && ( year == 0 ) ) ? 1 : 0;

    /* start with day of month */
    days = date;

    /* add month offset is days for day of year */
    days += ( leap ? ml : my )[ month ];

    /* add the preceeding years in the quad-year */
    days += year * 365;

    /* compensate if the first year of the quad-year is leap */
    if ( year > 0 && ( quadyear > 0 || cent == 0 ) ) ++ days;

    /* add the preceeding quad-years in the century */
    days += quadyear * 1461;

    /* compensate if the first quad-year of the century is not leap */
    if ( quadyear > 0 && cent > 0 ) -- days;

    /* add the preceeding centuries of the quad-century */
    days += cent * 36524;

    /* compensate if the first century of the quad-century is leap */
    if ( cent > 0 ) ++ days;

    /* add the preceeding quad-centuries */
    /* adjust for before epoch */
    days += ( quadcent - adjust ) * 146097;

    /* adjust for Rounded Julian Day Number */
    days += 1721060;

    /* return final Rounded Julian Day Number */
    return days;
}
#pragma section end CODE

#undef FUNCTION_NAME
