/***********************************************************************\
| Copyright  1997, by Philip David Howard - all rights reserved	|
|									|
| author	Philip D. Howard					|
| email		phil@txics.com, phil@milepost.com			|
| postal	Suite 129						|
|		17023 Preston Road					|
|		Dallas, Texas  75230					|
\***********************************************************************/

#include <stddef.h>

#include "days.h"

#pragma section begin SYNOPSIS
/***********************************************************************\
| function	ymdj_to_jdn						|
|									|
| purpose	convert a Julian Year, Month, and Date to a Julian Day	|
|		Number.							|
|									|
| arguments	1 (long) year (-1 = 2 BC)				|
|		2 (unsigned int) month (0..11)				|
|		3 (unsigned int) date (0..30)				|
|									|
| returns	(long) Julian Day Number				|
|									|
| note		A Julian Day Number (JDN) is a Julian Date rounded	|
|		to an integer.  The Julian Date is commonly used in	|
|		astronomy to identify dates in ephemeral data and	|
|		formulas.  The JDN is a number that is the same		|
|		through the ordinary civil day and can be used for	|
|		calendar references.					|
|									|
| ref		Explanatory Supplement to the Astronomical Almanac,	|
|		P. Kenneth Seidelmann, editor  [ISBN 0-935702-68-7]	|
|		http://www.amazon.com/exec/obidos/ISBN=0935702687	|
\***********************************************************************/
#pragma section end SYNOPSIS

#define FUNCTION_NAME "ymdj_to_jdn"

#pragma section begin FUNCTION
long
ymdj_to_jdn (
	long		year
,
	unsigned int	month
,
	unsigned int	date
)
#pragma section end FUNCTION

#pragma section begin CODE
{
    /* day of year for first day of each month */
    static unsigned int my[ 12 ] = {
	0,31,59,90,120,151,181,212,243,273,304,334 };
    static unsigned int ml[ 12 ] = {
	0,31,60,91,121,152,182,213,244,274,305,335 };

    unsigned long	adjust		;
    unsigned long	days		;
    unsigned long	quadyear	;
    int			leap		;

    /* adjust for before epoch */
    adjust = 0;
    if ( year < 0 ) {
	adjust = ( 3 - year ) / 4;
	year += adjust * 4;
    }

    /* figure out what quad-year this is */
    quadyear = year / 4;

    /* and how many years within the quad-year we are at */
    year %= 4;

    /* figure out if this is a leap year */
    leap = ( year == 0 ) ? 1 : 0;

    /* start with day of month */
    days = date;

    /* add month offset is days for day of year */
    days += ( leap ? ml : my )[ month ];

    /* add the preceeding years in the quad-year */
    days += year * 365;

    /* compensate if the leap year preceeds us */
    if ( year > 0 ) ++ days;

    /* add the preceeding quad-years */
    /* adjust for before epoch */
    days += ( quadyear - adjust ) * 1461;

    /* adjust for Rounded Julian Day Number */
    days += 1721058;

    /* return final Rounded Julian Day Number */
    return days;
}
#pragma section end CODE

#undef FUNCTION_NAME
